{ ====================================================================
Window interface

Copyright 2008 Roelf Toxopeus

Part of GUI stuff.
SwiftForth version.
A minimal window creating interface
Compatible where possible with MacForth and Mach2.
Last: 3 February 2020 at 14:24:41 CET     -rt
==================================================================== }

{ --------------------------------------------------------------------
About this window interface.

The OS does the housekeeping. Forth keeps track of things in case of a
turnkey, so windows can be restored with last settings.

Forth creates a named window record with NEW.WINDOW.
This record is initialy used to create the new window. The first entrees
will be set with the all important windowref and possible CGRef during adding
the window or during CoreGraphic initialisation. The other fields are initialy
filled with default values. The bounds field is used for scratch later on.
The record is also used after a turnkey/snapshot to re-add the window.
All the W.xxx and +W.xxx words affect this record (only) and don't change
the window (directly).
For real changes use the other words like the SET.Wxxx and xxx.WINDOW words.
Note: there is a variant on NEW.WINDOW, (NEW.WINDOW) which allows creating
nameless windows.

Windows are added to the environment with ADD.WINDOW.
If an existing window is re-added, it will be closed first.

To update the current window settings in the window record, use SAVE.WINDOW.
Useful before a snapshot or turnkey.

When closing a window, the window structure in Forth is not forgotten,
it's just pointing to 0: no window
You can reuse the window structure for another window, with ADD.WINDOW.

Most words defined are in MacForth style:

/WINDOW.RECORD -- size in bytes for a default window record.
Field pointers in to the record:
+W.REF -- points to windowref field, contains NSWindow ref.
+W.GPORT -- points to windowgport field, contains gport ref.
+W.BOUNDS -- points to window Content bounds field.
+W.BACKING -- points to windowbacking field, do we need this?
+W.DEFER -- points to windowdefer field, do we need this?
+W.STYLE -- points to windowstyle field
+W.TITLE -- points to windowtitle field, contains pointer to counted string.
+W.TRANSPARENCY -- points to windowtransparency field, contains transparency CGFloat.
+W.BACKGROUND -- points to windowbackground field, contains NSColor ref.

Affects the window record only, execute these before ADD.WINDOW.
W.BOUNDS -- set initial window Content bounds, without eventual titlebar 22e0 width 
W.STYLE -- set initial window style
W.TITLE -- set initial window title
W.TRANSPARENCY--  set initial window transparency
W.BACKGROUND -- set initial window background colour

Change window after ADD.WINDOW, window record is used for scratch.
RESIZE.WINDOW -- resize window Content size, with given width and height.
MOVE.WINDOW -- move window to given x and y.
FIX.WINDOW -- place window with given CGRect.
SHOW.WINDOW -- show window.
HIDE.WINDOW -- hide window, doesn't close it.
CLOSE.WINDOW -- close window, reset window record.
SET.WTITLE -- set/change the window title.
SET.WTRANSPARENCY -- set/change the window transparency.
SET.WBACKGROUND -- set/change the window background colour.
SET.WSTYLE -- set/change the window stylemask.

Get window properties
WINDOW.TITLE -- retrieve window title
WINDOW.TRANSPARENCY -- get window transparency
WINDOW.STYLE -- return window style
WINDOW.BACKGROUND - return background color as rgba
WINDOW.FRAME -- get window frame size CGRect, including titlebar 22e0 width
SAVE.WINDOW -- save window info. Content bounds is saved. Use before doing a turnkey/snapshot.

/WINDOW -- initialises a window record to default values.
(NEW.WINDOW) -- allocates an initialised window record.
NEW.WINDOW -- creates a new named window record, initialised to defaults.
WINDOW.ARGS -- fetches the necessary parameters for creating a NSWindow
instance, from a window record.
CREATE.WINDOW -- creates new NSWindow instance. Values stored in window
record can be used as parameters.
ADD.WINDOW -- create NSWindow instance and show it on screen. Uses values
from given window record.

Some color utillities:
RGBA>NSCOLOR -- create a NSColorRef from r g b a CGFloats.
quartz.framework -- the Quartz framework, needed for:
@RGBA -- the reverse of RGBA>NSCOLOR, returns the r g b a components/values
from a NSColorRef. Uses CGColor functions. NSColor itself can't do this,
with all colors!

-------------------------------------------------------------------- }

/FORTH
DECIMAL

\ --------------------------------------------------------------------
\ NSWindow class

COCOACLASS NSWindow

COCOA: @initWithContentRect:styleMask:backing:defer: ( cgfloat:x cgfloat:y cgfloat:w cgfloat:h stylemask backing defer -- id ) \ 4 CGFloats int int BOOL -- NSWindow object
COCOA: @setOpaque: ( flag -- ret )                        \ BOOL flag
COCOA: @orderFront: ( sender -- ret )                     \ id sender
COCOA: @orderOut: ( sender -- ret )                       \ id sender
COCOA: @close ( -- ret )
COCOA: @setTitle: ( title -- ret )                        \ NSString object
COCOA: @title ( -- NSStringRef )
COCOA: @setAlphaValue: ( viewAlpha -- ret )               \ CGFloat is sfloat
COCOA-FPRET: @alphaValue ( -- ) ( F: -- alpha )
COCOA: @setBackgroundColor: ( backgroundColor -- ret )    \ NSColor object
COCOA: @backgroundColor ( -- ref )
\ COCOA: @display ( -- ret )
COCOA: @setStyleMask: ( nsuinteger:stylemask -- ret )
COCOA: @styleMask ( -- stylemask )
COCOA: @setContentSize: ( cgfloat:w cgfloat:h -- ret )
COCOA: @setFrameOrigin: ( cgfloat:x cgfloat:y -- ret )
COCOA: @setFrame:display: ( cgfloat:x cgfloat:y cgfloat:w cgfloat:h flag -- ret )
COCOA-STRET: @frame ( *rect -- ret )
COCOA-STRET: @contentRectForFrameRect: ( x y w h *rect -- *rect )

\ --------------------------------------------------------------------
\ NSWindow parameters

COCOACLASS NSColor

COCOA: @CGColor ( -- CGColorRef )
COCOA: @colorWithCalibratedRed:green:blue:alpha: ( red green blue alpha -- NSColorRef )

: RGBA>NSCOLOR ( -- NSColorRef ) ( F: r g b a -- )   4FPUSHS NSColor @colorWithCalibratedRed:green:blue:alpha: ;

0 CONSTANT NSBorderlessWindowMask
1 CONSTANT NSTitledWindowMask
2 CONSTANT NSClosableWindowMask
4 CONSTANT NSMiniaturizableWindowMask
8 CONSTANT NSResizableWindowMask
256 CONSTANT NSTexturedBackgroundWindowMask

NSTitledWindowMask NSClosableWindowMask OR NSMiniaturizableWindowMask OR
NSResizableWindowMask OR CONSTANT ALLWIDGETS

NSTexturedBackgroundWindowMask CONSTANT NOWIDGETS

0 CONSTANT NSBackingStoreRetained
1 CONSTANT NSBackingStoreNonretained
2 CONSTANT NSBackingStoreBuffered

\ --------------------------------------------------------------------
\ necessary CoreGraphics color handling

FRAMEWORK Quartz.framework
FUNCTION: CGColorGetNumberOfComponents ( color -- #components )
FUNCTION: CGColorGetComponents ( color -- *components )

: @RGBA ( NSColorRef -- ) ( F: -- r g b a )
   @cgcolor DUP
   CGColorGetComponents
   SWAP CGColorGetNumberOfComponents 2 > IF
      CG4@   ( r g b a )
   ELSE
      DUP SF@ FDUP FDUP  ( r g b )
      SFLOAT+ SF@    ( a )
   THEN ;

\ --------------------------------------------------------------------
\ Forth NSWindow record and parameters

6 CELLS 9 SFLOATS + CONSTANT /WINDOW.RECORD

: +W.REF ( wptr4 -- 'windowref )  ;

: +W.GPORT ( wptr4 -- 'windowgport )  CELL + ;

: +W.BOUNDS ( wptr4 -- 'windowbounds ) [ 0 +W.GPORT CELL+ ] LITERAL + ;

: +W.BACKING ( wptr4 -- 'windowbacking ) [ 0 +W.BOUNDS 4 SFLOATS + ] LITERAL + ;

: +W.DEFER ( wptr4 -- 'windowbacking ) [ 0 +W.BACKING CELL+ ] LITERAL +  ;

: +W.STYLE ( wptr4 -- 'windowstyle ) [ 0 +W.DEFER CELL+ ] LITERAL + ;

: +W.TITLE ( wptr4 -- 'windowtitle ) [ 0 +W.STYLE CELL+ ] LITERAL + ;

: +W.TRANSPARENCY ( wptr4 -- 'windowtransparency ) [ 0 +W.TITLE CELL+ ] LITERAL + ;

: +W.BACKGROUND ( wptr4 -- 'windowbackground ) [ 0 +W.TRANSPARENCY SFLOAT+ ] LITERAL + ;

\ --------------------------------------------------------------------
\ set/change stuff before window is added with ADD.WINDOW

: W.BOUNDS ( wptr4 -- ) ( F: x y width height -- )   +W.BOUNDS CG4! ;

: W.STYLE ( n wptr4 -- )   +W.STYLE ! ;

: W.TITLE ( a n wptr4 -- )   -ROT HERE >R STRING, R> SWAP +W.TITLE ! ;

: W.TRANSPARENCY ( wptr4 -- ) ( F: alpha -- )  +W.TRANSPARENCY SF! ;

: W.BACKGROUND ( wptr4 -- ) ( F: r g b a -- )   +W.BACKGROUND CG4! ;

\ --------------------------------------------------------------------
\ User interface words, interact with Cocoa
\ -- move, size etc.

: RESIZE.WINDOW ( wptr4 -- ) ( F: w h -- )
   >R FPUSHS FPUSHS SWAP R> +W.REF @ @setContentSize: DROP ;

: MOVE.WINDOW ( wptr4 -- ) ( F: x y -- )
   >R FPUSHS FPUSHS SWAP R> +w.ref @ @setFrameOrigin: DROP  ;

: FIX.WINDOW   ( wptr4 -- ) ( F: x y w h -- )
   >R 4FPUSHS YES R> +W.REF @ @setFrame:display: DROP ;

: SHOW.WINDOW ( wptr4 -- )
   +W.REF @ 0 SWAP @orderFront: DROP ;

: HIDE.WINDOW ( wptr4 -- )
   +W.REF @ DUP @orderOut: DROP ;
   
: CLOSE.WINDOW ( wptr4 -- )
   DUP +W.REF @ DUP @close DROP @release DROP OFF ;

\ -- set window properties
: SET.WTITLE ( 0string wptr4 -- )
   SWAP >NSSTRING SWAP +W.REF @  @setTitle: DROP ;

: SET.WTRANSPARENCY ( wptr4 -- ) ( F: alpha -- )
   >R FPUSHS R> +W.REF @ @setAlphaValue: DROP ;

: SET.WSTYLE ( stylemask wptr4 -- )  \ crashes?
   +W.REF @ @setStyleMask: DROP ;

: SET.WBACKGROUND ( wptr4 -- ) ( F: red green blue alpha -- )
   RGBA>NSCOLOR SWAP +W.REF @ @setBackgroundColor: DROP ;

\ -- retrieve window properties
: WINDOW.FRAME ( wptr4 -- ) ( F: -- x y w h )
   DUP +W.BOUNDS DUP >R SWAP +W.REF @ @frame DROP R>  CG4@ ;

: CONTENT.FRAME ( wptr4 -- ) ( F: -- x y w h )
   DUP >R WINDOW.FRAME  4FPUSHS R> DUP +W.BOUNDS SWAP +W.REF @
   @contentRectForFrameRect: CG4@ ;
   
: WINDOW.TITLE ( wptr4 -- a n )  +W.REF @ @title >4THSTRING ;

: WINDOW.TRANSPARENCY ( wptr4 -- ) ( F: -- alpha )  +W.REF @ @alphaValue ;

: WINDOW.STYLE ( wptr4 -- stylemask )   +W.REF @ @styleMask ;
      
: WINDOW.BACKGROUND ( wptr4 -- ) ( F: -- red green blue alpha )   +W.REF @ @backgroundColor @RGBA ;

\ -- save window properties
: SAVE.WINDOW ( wptr4 -- )
   \ DUP WINDOW.FRAME 22e0 F- DUP W.BOUNDS   \ better is CONTENT.FRAME
   DUP CONTENT.FRAME FDROP FDROP FDROP FDROP
   DUP DUP WINDOW.TITLE ROT W.TITLE
   DUP WINDOW.TRANSPARENCY DUP W.TRANSPARENCY
   DUP WINDOW.STYLE OVER W.STYLE
   DUP WINDOW.BACKGROUND W.BACKGROUND ;

\ --------------------------------------------------------------------
\ Window creating words

\ Note: was z" for s" !!
: /WINDOW ( a -- )
   0                         OVER +W.REF !
   0                         OVER +W.GPORT !
   100e0 575e0 340e0 180e0   DUP  W.BOUNDS     \ was 600, 575 leaves room for titlebar and menubar
   ALLWIDGETS                OVER W.STYLE
   NSBackingStoreBuffered    OVER +W.BACKING !
   NO                        OVER +W.DEFER !
   DUP S" untitled"          ROT  W.TITLE         \ 'saves' as a counted string!
   1e0                       DUP  W.TRANSPARENCY
   1e0 fDUP F2DUP            W.BACKGROUND
;

: (NEW.WINDOW) (  -- addr )   HERE /WINDOW.RECORD ALLOT DUP /WINDOW ;

\ : NEW.WINDOW ( <name > -- )  (NEW.WINDOW) CONSTANT ;

: NEW.WINDOW ( <name > -- )  CREATE (NEW.WINDOW) DROP ;

: WINDOW.ARGS ( wptr4 -- n1 ... n7 )
   >R
   R@ +W.BOUNDS CGRECT@
   R@ +W.STYLE @
   R@ +W.BACKING @
   R> +W.DEFER @
;

: CREATE.WINDOW ( x y w h style backing defer -- instance )
   NSWindow @alloc @initWithContentRect:styleMask:backing:defer: ;

: ADD.WINDOW ( wptr4 -- )
   DUP +W.REF @ IF DUP CLOSE.WINDOW THEN
   DUP WINDOW.ARGS CREATE.WINDOW
\ default opaque setting is YES, so need to chage it to change transparency!
   NO OVER @setOpaque: DROP
   OVER +W.REF !
   DUP +W.TITLE @ COUNT DROP OVER SET.WTITLE         \ this is ugly because of counted vs zero string
   DUP +W.TRANSPARENCY SF@ DUP SET.WTRANSPARENCY
   DUP +W.BACKGROUND CG4@ DUP SET.WBACKGROUND
   SHOW.WINDOW ;

\\ ( eof )