{ ====================================================================
Transform SF in to GUI app

Copyright 2009 Roelf Toxopeus

Part of adding Cocoa event handling to Forth.
SwiftForth version.
Last: 10 December 2019 at 12:51:23 CET  -rt
==================================================================== }

{ --------------------------------------------------------------------
Make current background proces (SwiftForth) a GUI app, at least wrt
the GUI itself. Uses NSApplication ActivationPolicies.
macOS 10.9 and later allow an extra policy:
NSApplicationActivationPolicyAccessory
see below.

SF>COCOA -- transform sf in to coco-sf, following standard procedure:
allocate autoreleasepool, create shared application and set the NSApp
policy to NSApplicationActivationPolicyRegular.
*It becomes an ordinary app with an icon in the Dock and a user interface.
NSApplicationActivationPolicyRegular is the default for bundled apps,
unless overridden in the Info.plist.

Prior to SF>COCOA, sf activation policy is NSApplicationActivationPolicyProhibited.
Which is typical for processes running in a shell, like sf.
*It does not appear in the Dock and may not create windows or be activated.
This corresponds to the value of the LSBackgroundOnly key in the application’s
Info.plist being 1. This is also the default for unbundled executables
that do not have Info.plists.

* = learned from NSRunningApplication.h

0 CONSTANT NSApplicationActivationPolicyRegular
2 CONSTANT NSApplicationActivationPolicyProhibited

Test: redefine SF>COCOA and recomplie
COCOA: @activationPolicy ( -- policy )

VARIABLE SF-POLICY

: SF>COCOA ( -- )
   NSApp @ @activationPolicy SF-POLICY !
   SF>COCOA ;

macOS 10.9 and later:
Attempts to modify the application's activation policy.  In OS X 10.9, any
policy may be set; prior to 10.9, the activation policy may be changed to
NSApplicationActivationPolicyProhibited or
NSApplicationActivationPolicyRegular, but may not be changed to
NSApplicationActivationPolicyAccessory.  This returns YES if setting the
activation policy is successful, and NO if not.


*NSApplicationActivationPolicyAccessory, The application does not appear in the
Dock and does not have a menu bar, but it may be activated programmatically or
by clicking on one of its windows. This corresponds to LSUIElement=1 in the
Info.plist.

This gives a coco-sf without the NSApp part in the Dock, but still capable
to run GUI stuff, but no sfmenubar.

1 CONSTANT NSApplicationActivationPolicyAccessory

/APP -- make given NSAppRef a Regular app with NSAppRef in Dock and
a menubar.
/ACC -- make given NSApp an Accessory app without NSAppRef in Dock
and without menubar.

Notice you can change the policy at will from app to acc, back to app etc.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

COCOACLASS NSApplication

COCOA: @sharedApplication ( -- NSApplication:ref )
COCOA: @setActivationPolicy: ( policy -- bool )

VARIABLE MAINPOOL    \ cache main autoreleasepool

: /APP ( nsapp - )   0 SWAP @setActivationPolicy: DROP ;

: /ACC ( nsapp - )   1 SWAP @setActivationPolicy: DROP ;

: SF>COCOA ( -- )
   allocPool MAINPOOL !
   NSApplication @sharedApplication DUP 0= ABORT" Can't create a shared application !"
  /APP ;  \ NSApplicationActivationPolicyRegular

\\ ( eof )

\ test:
( launch coco-sf)

hello again   ok                      <- boot with regular policy, NSApp app in Dock, empty menubar
  ok
nsapp @ /acc  ok                      <- change to accessory policy, NSApp app out Dock, no menubar
new.window win win add.window  ok     <- put window on screen
win close.window  ok                  <- window disappears and released                       
nsapp @ /app  ok                      <- change to regular policy, NSApp app in Dock, empty menubar
mac include hotcoco/simplemenu.f
/SIMPLEMENU to add a menubar to coco-sf ok
/SIMPLEMENU  ok                       <- simplemenu menubar
win add.window  ok                    <- add new window
nsapp @ /acc  ok                      <- change to accessory policy, NSApp app out Dock, no menubar, window hidden, released?
win show.window  ok                   <- window shows, still there
nsapp @ /app  ok                      <- change to regular policy, NSApp app in Dock, simplemenu menubar back on                         
bye 
