{ ====================================================================
File and URL app launching.

Copyright 2013 Roelf Toxopeus

SwitForth version.
Use of NSWorkspace to open files/locations in their default applications.
Last: 30 October 2020 at 16:46:39 CET  -rt
==================================================================== }

{ --------------------------------------------------------------------
Note for 64b versions: nearly everything used here for NSWorkspace is
deprecated or removed in 10.15/10.16/11.0. LAUNCHURL seems ok.
See: https://developer.apple.com/documentation/appkit/nsworkspace/deprecated_symbols?language=objc
32b versions up to and including 10.14 Mojave are ok.

LAUNCHAPP -- launches new app instance. Example of toll-free bridging
between CFURL and NSURL. Here CFURL is used.

LAUNCHURL -- launches default browser with given url. NSURL is used.
GOTO" -- browser and go to quoted url, as example:
SFWEB -- go to FORTH, Inc web home web page.

LAUNCHFILE -- launches default opener with given file.
SFMANUAL -- open SwiftForth manual in default PDF viewer.

More examples of NSWorkspace usage:
\ launch new Terminal instance. Interesting...
\ as if it launched a copy of currently running Terminal.app
\ with coco-sf and whatever else running !
: >TERMINAL ( -- )
  Z" /Applications/Utilities/Terminal.app" LAUNCHAPP 2DROP ;

\ launch another coco-sf and quit current
: HELLO ( -- )
   THIS-EXE-NAME PAD ZPLACE  PAD LAUNCHFILE  BYE ;

\ hide all applications from sight
: HH ( -- )
  NSWorkspace @sharedWorkspace @hideOtherApplications DROP ;
-------------------------------------------------------------------- }

/FORTH
DECIMAL

COCOACLASS NSWorkspace
COCOA: @sharedWorkspace ( -- ref )
COCOA: @openURL: ( nsurl -- boolean )
COCOA: @openFile: ( nsstring -- boolean )
COCOA: @launchApplication: ( string -- ret )
COCOA: @launchApplicationAtURL:options:configuration:error: ( url options dict *error -- 0|ref )
COCOA: @hideOtherApplications ( -- ret )

COCOACLASS NSURL
COCOA: URLWithString: ( nsstring -- nsurl )

COCOACLASS NSDictionary
COCOA: @dictionary ( -- ref )

$00080000 CONSTANT NSWorkspaceLaunchNewInstance

: LAUNCHAPP ( zstring -- ref error)
   >CFURL DUP
   NSWorkspaceLaunchNewInstance
   NSDictionary @dictionary
   0 >R RP@
   NSWorkspace @sharedWorkspace
   @launchApplicationAtURL:options:configuration:error: R>
   ROT CFRelease DROP ;

: LAUNCHURL ( zstring -- )
   >NSSTRING DUP
   NSURL URLWithString:
   NSWorkspace @sharedWorkspace @openURL: DROP
   @release DROP ;

\ : URL: ( "<spaces>name" -- )   BL WORD COUNT DROP LaunchURL ;
   
: GOTO" ( -- )
   POSTPONE Z"
   STATE @
         IF POSTPONE LAUNCHURL
      ELSE LAUNCHURL
   THEN ; IMMEDIATE

: SFWEB ( -- )   GOTO" http://www.forth.com" ;

: LAUNCHFILE ( zstring -- )   
   >NSSTRING DUP
   NSWorkspace @sharedWorkspace @openFile: DROP
   @release DROP ;

: SFMANUAL ( -- )
   ROOTPATH COUNT PAD ZPLACE
   s" swiftforth/doc/swiftforth-linux-macos.pdf" PAD ZAPPEND
   PAD LAUNCHFILE ;

cr .( SFWEB     will bring you to Forth, Inc.)
cr .( SFMANUAL  will open the SF manual)

\\ ( eof )