{ ====================================================================
Show alert

Copyright 2010 Roelf Toxopeus

SwiftForth version.
Preferred alert usage: NSAlert
Last: 6 October 2020 at 12:18:25 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
This example shows how to bring up a NSAlert panel.
It shows how to add features like a help button, a style, messages and
buttons. It also shows how to override the default button action.
It deals with the modal responses, ie the values returned by the modal
runloop when quitting the panel. 

/ALERT -- returns an alert reference initialised with info and message
zstrings. Two buttons and the help button are added.

ALERT -- runs the alert panel from the mainthread. The id from the
button hit to dismiss the panel is returned.

WELCOME -- example, shows which button fired.
Does not release the alert, so panel can be examined.
It check for an existing panel, creates one if not there.

-WELCOME -- release an existing panel.

---
For the returned values after running the alert:
https://developer.apple.com/documentation/appkit/nsmodalresponse?language=objc

NSAlertFirstButtonReturn
The user clicked the first (rightmost) button on the dialog or sheet.

NSAlertSecondButtonReturn
The user clicked the second button from the right edge of the dialog or sheet.

NSAlertThirdButtonReturn
The user clicked the third button from the right edge of the dialog or sheet

Any additional buttons return:
NSAlertThirdButtonReturn + position from the right edge of the dialog or sheet

For their values see NSAlert.h  they're not in the normal documentation :(

---
myAlert @ @buttons                    \ returns an array of buttons
myAlert @ @buttons @count             \ returns #buttons
0 myAlert @ @buttons @objectAtIndex:  \ returns right most object

0 myAlert @ @buttons @objectAtIndex: okButton @ =
1 myAlert @ @buttons @objectAtIndex: extraButton @ =

TEST -- runs WELCOME and notifies which button was hit.
All buttons have a default action of quiting the modal runloop.
Depending on which button is hit, actions can be performed after
the modal runloop.

---
/XBUTTON -- will override the default behaviour of the extraButton.
Now the extraButton will not quit the modal runloop but performs
its action inside the modal runloop.

TEST2 -- runs TEST with /XBUTTON.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

COCOACLASS NSAlert

COCOA: @setAlertStyle: ( style -- ret )
COCOA: @setInformativeText: ( NSStringRef -- ret )
COCOA: @setMessageText: ( NSStringRef -- ret )
COCOA: @setShowsHelp: ( bool -- ret )
COCOA: @setDelegate: ( ref -- ret )
COCOA: @addButtonWithTitle: ( NSStringRef -- NSButtonRef )
COCOA: @buttons ( -- NSArrayRef )

VARIABLE myAlert

COCOACLASS NSButton
COCOA: @keyEquivalent ( -- NSStringRef )
COCOA: @setKeyEquivalent: ( NSStringRef -- ret )

VARIABLE okButton
VARIABLE extraButton

1 CONSTANT NSAlertStyleInformational

FORTHCALL: *alertShowHelp: ( rec sel alert -- bool )   IMPOSTOR'S CR ." Helping Hand" TRUE ;

: ALERTTYPE ( -- NSStringref )   Z" b@:@" ;

*alertShowHelp: ALERTTYPE Z" alertShowHelp:" NSAlert ADD.METHOD

: /ALERT ( info message -- NSAlertRef )
   myAlert @ ?DUP IF -ROT 2DROP EXIT THEN               \ already got an alert
   NSAlert @alloc @init >R
   NSAlertStyleInformational R@ @setAlertStyle: DROP
   >NSSTRING DUP R@ @setMessageText: DROP @release DROP
   >NSSTRING DUP R@ @setInformativeText: DROP @release DROP
   TRUE R@ @setShowsHelp: DROP
   R@ DUP @setDelegate: DROP     \ self delegate
   @" OK" R@ @addButtonWithTitle: okButton !
   @" Extra" R@ @addButtonWithTitle: extraButton !
   R> DUP myAlert ! ;

: (ALERT) ( NSAlertRef -- )   /ALERT @runModal POCKET ! ;
: ALERT ( za1 za2 -- button )   2 0 ['] (ALERT) PASS POCKET @ ;

: (-ALERT) ( -- )   myAlert @ ?DUP IF @release DROP myAlert OFF THEN ;
: -ALERT ( -- )   0 DUP ['] (-ALERT) PASS ;  

: /WELCOME ( -- za1 za2 )
   Z" Yes, again..."
   Z" Hi !" ;

: WELCOME  ( -- )    /WELCOME ALERT . ;

\ ---
\ check for returned button after quiting the modal runloop.
1000 CONSTANT NSAlertFirstButtonReturn
1001 CONSTANT NSAlertSecondButtonReturn
1002 CONSTANT NSAlertThirdButtonReturn

: TEST ( -- )
   /WELCOME ALERT
   DUP NSAlertFirstButtonReturn = IF DROP CR ." OK button hit " EXIT THEN
   DUP NSAlertSecondButtonReturn = IF DROP CR ." Extra button hit" EXIT THEN
   CR . S"   Not recognised !" ;

\ ---
\ take action after extra button is hit while in the modal runloop.
\ NSActionCell
COCOA: @setTarget: ( id -- ret )
COCOA: @setAction: ( sel -- ret )

COCOA: @window ( -- wref )
VARIABLE TOGGLE-BG

FORTHCALL: *klik ( rec sel sender -- void )
   IMPOSTOR'S ( CR ) INVERSE ." Taking Action on hitting Extra Button"
   NSColor
   TOGGLE-BG DUP @ IF OFF @windowBackgroundColor ELSE ON @redColor THEN
   myAlert @ @window  @setBackgroundColor: DROP
   CR ." Hit OK button when done" NORMAL ;

: CONTROLTYPES ( -- za )  Z" v@:@" ; 

*klik CONTROLTYPES Z" klik" NSButton ADD.METHOD

: /XBUTTON ( -- )    extraButton @ DUP DUP @setTarget: DROP    ( self is target )
   Z" klik" @selector SWAP @setAction: DROP ;
   
: TEST2 ( -- )
   myAlert @ 0= IF /WELCOME 2 0 ['] /ALERT PASS THEN
   /XBUTTON /WELCOME ALERT
   DUP NSAlertFirstButtonReturn = IF DROP CR ." OK button hit " EXIT THEN
   DUP NSAlertSecondButtonReturn = IF DROP CR ." Extra button hit" EXIT THEN
   CR . S"   Not recognised !" ;

cr .( try  TEST   reset with -ALERT  try   TEST2)

\\ ( eof )


