\ screensaver...
\ using Padbury Clock Quartz Composition:
\ https://www.screensaversplanet.com/screensavers/padbury-clock-1027/
\ It seems that in macOS 10.14 Mojave, the Quartz Composition screen
\ savers stopped working. But Quartz compositions can still be run as
\ is shown here. Both 32bit and 64bit (iForth).

\ Cobbled together from
\  quartzcomposer3.f
\  test-cocoa-eventhandler.f
\  cocoa-focus2.f
\  etc.
\ Last 30 January 2019 at 10:05:27 CEST

/forth
decimal

lacking quartz.framework  framework quartz.framework

quartz.framework

cocoaclass QCView

cocoa: @loadCompositionFromFile: ( NSString:path -- bool )
cocoa: @unloadComposition ( -- ret )
cocoa: @setAutostartsRendering: ( bool -- ret )
cocoa: @startRendering ( -- bool )
cocoa: @stopRendering ( -- ret )
cocoa: @setEventForwardingMask: ( NSUInteger:mask -- ret )
   
\ window
new.window mywin
nowidgets            mywin w.style
0e0 fdup fdup 1e0    mywin w.background
                     mywin full.frame  f2drop f2drop

\ quartzcomposerview
variable myqcview

\ add and glue window and view
: win+view ( -- )
   mywin dup add.window
   pause
   QCView @alloc @init dup myqcview !
   swap make.contentview ;

\ remove and release win and view
: -win-view ( -- )
   myqcview @ dup @unloadComposition drop
   @release drop
   mywin close.window ;
   
\ some settings
: /qcview ( -- )
   myqcview @ 
   yes over @setAutostartsRendering: drop
   0 swap @setEventForwardingMask: drop  \ mouse down kept for view !
;

\ turnkey app
: padbury-composition ( -- za )
   CFBundleGetMainBundle dup CFBundleCopyResourcesDirectoryURL
   swap cfrelease drop
   cfurl> >r s" /24_Hour_Clock.qtz" r@ zappend r> ;

(*
\ non turnkey app
: padbury-composition ( -- za )
   z" /Users/roelf/Library/Screen Savers/24_Hour_Clock.qtz" ;
*)

: /qc ( -- )
   -stderr  \ Mojave will show some information about lacking filters in 32bit
   padbury-composition
   >nsstring myqcview @ @loadCompositionFromFile: 0= abort" no clock composition !"
   +stderr ;

\ run it
: qc ( -- )   0" startRendering" @selector 0 no myqcview @ formain drop ;

\ stop it
: -qc ( -- )  myqcview @ @stopRendering drop ;

\ build and run
: hello-qc ( -- )   win+view  /qcview  /qc qc ;

\ stop and close
: bye-qc ( -- )   -qc -win-view ;


\ mouse interaction, short click/tap.
\ just terminate coco-sf to avoid problems

cocoa: @terminate: ( sender -- ret )

cocoaclass NSView

forthcall: *mouseDown: ( rec sel event -- ret )   bye-qc nsapp @ dup @terminate:  ;

: mousedowntypes ( -- za )  0" v@:@" ;

: /mousedown ( -- )   *mouseDown: mousedowntypes z" mouseDown:" NSView add.method ;

cocoa: @setPresentationOptions: ( options -- ret ) \ NSApplication

\ and there're many more to literaly lock yourself in.
1 1 lshift constant NSApplicationPresentationHideDock
1 3 lshift constant NSApplicationPresentationHideMenuBar
1 5 lshift constant NSApplicationPresentationDisableProcessSwitching

: weown ( -- )
   NSApplicationPresentationHideDock
   NSApplicationPresentationHideMenuBar or
   NSApplicationPresentationDisableProcessSwitching or
   NSApp @ @setPresentationOptions: drop ;

\ run screen saver with full running mainthread, coco-sf 2.4 %
: ss ( -- )   /mousedown pushme weown 100 ms hello-qc ;

\ run screen saver with minimal running mainthread, coco-sf 0.5 %
\ tell mainthread every second to sleep for 900 ms and thus work for 100 ms.
\ this seems to be enough to drive the clock. you could fine tune...
\ note: PASS waits for return, so sleeping on mainthread = sleeping in IMPOSTOR
: xx ( -- )   ss  begin 900 1 0 ['] ms pass  100 ms again ;

\\