{ ====================================================================
ShellView Console

Copyright 2014 Roelf Toxopeus

SwiftForth version.
Behaves (hopefully) as a Terminal.app console.
Last: 17 October 2020 at 23:33:08 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
/ACCEPT -- to avoid having to deal with the line editing stuff for now
simple accept is used.
-ACCEPT -- restore initial accept.

PIPE1 PIPE2 -- variables containing the NSPipeRefs.
Two pipes are needed: shellviewIn  <- forthOut
                      shellviewOut -> forthIn
/PIPES -- create NSPipe instances for I/O. Store refs in PIPE1, PIPE2.
-PIPES -- release NSPipe instances in PIPE1, PIPE2.
/IO -- create pipes for I/O, set the appropriate fd's in Forth's
FD-IN and FD-OUT, return the appropriate NSFileHandle refs.
These will be used to create a SFShellView instance.

MYSHELLWIN -- the window for the shellview

/VIEWIO -- init shellview in window and set Forth I/O accordingly.
This makes MYSHELLWIN the I/O console for Forth.

-VIEWIO -- restore to default Terminal console. Restore ACCEPT and
release used NSPipe refs.

/TERMINAL -- Warm boot in Terminal console.

/COCOA-CONSOLE -- Warm boot in new console window.

---

When running coco-sf in the cocoa console, you can disconnect from
the Terminal if you whish.
1. Bring the Terminal console to the front
2. type  CTRL-Z
3. type  disown -h; fg <ret>
4. Close the Terminal console window or just quit Terminal.app

Caveat: any graphics in another window, will have to run on separate
threads/tasks.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

\ --------------------------------------------------------------------
\ simple console

VARIABLE SAVED-ACCEPT
: /ACCEPT ( -- )   ['] (ACCEPT) 'ACCEPT DUP @ SAVED-ACCEPT ! ! ;
: -ACCEPT ( -- )   SAVED-ACCEPT @ ?DUP IF 'ACCEPT ! THEN ;

\ --------------------------------------------------------------------
\ the pipes for IPC between Forth and SFShellView

COCOACLASS NSPipe
COCOA: @pipe ( -- NSPipeRef )    \ class method
COCOA: @fileHandleForReading ( -- NSFileHandle )
COCOA: @fileHandleForWriting ( -- NSFileHandle )

VARIABLE PIPE1 VARIABLE PIPE2

: /PIPES ( -- pipe1 pipe2 )
   NSPipe @pipe DUP 0= ABORT" Could not create a NSPipe instance !"
   DUP PIPE1 !
   NSPipe @pipe DUP 0= IF
        PIPE1 DUP @ @release DROP OFF ." Could not create a NSPipe instance !"
        ABORT
   THEN
   DUP PIPE2 ! ;

: -PIPES ( -- )
   PIPE1 DUP @ @release DROP OFF
   PIPE2 DUP @ @release DROP OFF ;

\ COCOACLASS NSFileHandle
COCOA: @fileDescriptor ( -- Posixfd )

: /IO ( -- fhin fhout )
   /PIPES    \ before setting ACCEPT
   /ACCEPT   \ pipes ok, simple ACCEPT
   >R DUP @fileHandleForReading SWAP @fileHandleForWriting        ( fhin1 fhout1 )
   @fileDescriptor FD-OUT !                                       ( fhin1 )
   R> DUP @fileHandleForReading SWAP @fileHandleForWriting        ( fhin1 fhin2 fhout2 )
   SWAP @fileDescriptor FD-IN !                                   ( fhin1 fhout2 )
;

\ --------------------------------------------------------------------
\ the window and the SFShellView contentview

NEW.WINDOW MYSHELLWIN
100e0 100e0 500e0 200e0 MYSHELLWIN W.BOUNDS
S" Coco-SF I/O View"    MYSHELLWIN W.TITLE

VARIABLE MYSHELLVIEW

: /VIEWIO ( -- )
   /IO                          \ do this first, if no pipe, don't bother to continue
   MYSHELLWIN ADD.WINDOW
   MYSHELLWIN WINDOWFORSHELL DUP MYSHELLVIEW !
   0" Monaco" 14 ROT VIEWFONT ;

: -VIEWIO ( -- )
   -PIPES
   MYSHELLVIEW DUP @ -SHELLVIEW OFF
   >STDIN >STDOUT
   -ACCEPT ;

\ --------------------------------------------------------------------
\ the topwords

: /TERMINAL ( -- )
   -VIEWIO
   /CONSOLE
   CR ." I/O redirected to and from Terminal.app"
   CR PROMPT
   MYSHELLWIN 1 0 ['] CLOSE.WINDOW PASS
   QUIT ;

: /COCOA-CONSOLE ( -- )
   /VIEWIO
   .ABOUT
   CR CR ." hello again"  CR
   QUIT ;

CR .( shellview console loaded)
CR .( run:      /COCOA-CONSOLE)
CR .( to stop:  /TERMINAL)

\\ ( eof )