{ ====================================================================
cblock utillity

Copyright 2011 Roelf Toxopeus

SwiftForth version
C and Objectve C blocks as arguments and functions in Forth.
Last: 4 November 2020 at 22:02:44 CET -rt
==================================================================== }

{ --------------------------------------------------------------------
Passing a block parameter to foreign functions and Cocoa methods.

|CBLOCK| -- Block literal structure size, 7 cells.
cell0 *isa
cell1 flags
cell2 reserved 
cell3 *invoke
cell4 *descriptor
cell5 descriptor -> reserved
cell6 descriptor -> size

/CBLOCK -- initializes a block literal with Forth callback and default
values for block usage as argument to C and Objective C functions
and methods. In effect you wrap a Forth callback in a structure, the
so called 'block literal'.

Usage:
|CBLOCK| BUFFER: <name>
'callback' <name> /CBLOCK
Pass <name> to functions and methods accepting blocks as arguments.
An alternative is to use a temporary buffer on the return stack, this
only works in definitions: |CBLOCK| R-ALLOC

When the block, actually the Forth callback, runs, it is passed the
block literal pointer as the left most parameter. Any other arguments
are passed above it.

Example:
\ the specific cblock
|CBLOCK| BUFFER: KLOK

\ the stuff to do in the block
VARIABLE SECONDS
FORTHCALL: *TIKKER ( block -- )   BEGIN 1000 MS  1 SECONDS +! AGAIN ;

\ GCD call to dispatch a block to a queue
FUNCTION: dispatch_async ( queue block -- ret )

\ run the klok, queue it
: TIKT ( -- )
   KLOK *TIKKER OVER /CBLOCK
   DEFAULTQUEUE SWAP dispatch_async DROP ;

\ Note the difference with the function taking version:
: TIKT2 ( -- )   DEFAULTQUEUE KLOK *TIKKER dispatch_async_f DROP ;

\ Running a block as a foreign function:
FORTHCALL: *1TIK ( block -- )   1000 MS  1 SECONDS +! ;

: TIK ( -- )
   *1TIK KLOK /CBLOCK                        
   KLOK 1 *1TIK EXTERN-CALL DROP ;
-------------------------------------------------------------------- }

/FORTH
DECIMAL

\ isa pointer
GLOBAL: _NSConcreteStackBlock

1 28 LSHIFT CONSTANT BLOCKFLAGS

7 CELLS CONSTANT |CBLOCK|

: /CBLOCk ( cb buf -- )                  \ the numbering is for 32b
    ( 0) _NSConcreteStackBlock OVER !    \  0 -- isa  <= experiment with uninitialized
    CELL+ 0 ( BLOCKFLAGS)      OVER !    \  4 -- flags  <= experiment with uninitialized
    CELL+ 0                    OVER !    \  8 -- reserved
    CELL+ SWAP                 OVER !    \ 12 -- callback pointer
    CELL+ DUP CELL+            OVER !    \ 16 -- pointer to Block_descriptor is actually next field
    CELL+ 0                    OVER !    \ 20 -- Block_descriptor: reserved
    CELL+ |CBLOCK|             SWAP !    \ 24 -- Block_descriptor: Block literal structure size
\ no copy and dispose functions, no extra data
;

\\ ( eof)