{ ====================================================================
NDP exception handling

Copyright 2020 Roelf Toxopeus

SwiftForth version.
Set NDP exceptions handling with exception mask options
Last: 1 August 2020 at 10:44:59 CEST   -rt
==================================================================== }

{ --------------------------------------------------------------------
With new exceptionhanlers for sf in 3.9.0, the ObjC Runtime -> NSApp @run
on the main thread hangs the moment you bring it to the front. 
Crashreports show it's a numeric exception. It's an Apple internal code
issue, no Forth code is involved.
Replacing the default sf mask with a default Intel/Apple ignore fp
exceptions mask, fixes the  issue.

>FPEXP -- set NDP with given mask. See below for some masks.

-FPEXP -- set NDP to ignore exceptions, they will not be catched by
sf. The mask used for the NDP controlword is $37F, ignore all.
-FPEXP will be executed for the main thread in RUNCOCOA.

FPEXP -- set NDP to throw numeric exceptions. They will be catched
by sf. The mask used for the NDP controlword is the default SwiftForth
$332 mask as found in fpconfig.f
The Forth thread can catch fp exceptions. FPEXP runs in GOES.FORTH.

Note: /NDP -> (/NDP) -> FINIT
fpmath.f   E3DB INH2 FINIT
But E3DB is FNINIT according:
From IA-32 Intel® Architecture Software Developer’s Manual
Volume 2: Instruction Set Reference
Page Vol. 2A  3-361
Opcode          Instruction       Description
9B DB E3        FINIT             Initialize FPU after checking for pending
                                  unmasked floating-point exceptions.
   DB E3        FNINIT            Initialize FPU without checking for pending
                                  unmasked floating-point exceptions.

The FNINIT instruction does not check for and handle any pending unmasked
floating-point exceptions before performing the initialization.
The real FINIT does FWAIT FNINIT.

FCLEX -- clear pending fp exceptions.
Assembler: E2DB INH2 FCLEX   -> E2DB is actually FNCLEX ?!
From IA-32 Intel® Architecture Software Developer’s Manual
Volume 2: Instruction Set Reference
Page 3-215
Opcode      Instruction    Description
9B DB E2    FCLEX          Clear floating-point exception flags after
                           checking for pending unmasked floating-point
                           exceptions.
   DB E2    FNCLEX         Clear floating-point exception flags without
                           checking for pending unmasked floating-point
                           exceptions.

If one or more exception flags are set in the FPU status word prior to
loading a new FPU control word and the new control word unmasks one or
more of those exceptions, a floating-point exception will be generated
upon execution of the next floating-point instruction (except for the
no-wait floating-point instructions, see the section titled “Software
Exception Handling” in Chapter 8 of the IA-32 Intel Architecture
Software Developer’s Manual, Volume 1). To avoid raising exceptions
when changing FPU operating modes, clear any pending exceptions (using
the FCLEX or FNCLEX instruction) before loading the new control word.

Don't know if FCLEX is needed when we use FINIT.

---
NDP Controlword
7 low bits, right to left:
invalid operation
denormal operand
zero divide
underflow
overflow
precission
reserved

The default for SwiftForth is $332 -> %1100110010
The default for Apple is the FINIT $37F with bit6 reserved set -> %1101111111
The FPU control word is set to 037FH (round to nearest, all exceptions masked, 64-bit precision).

What happens if we set bit6 in the SwiftForth default as well?
%1101110010 >FPEXP            still numeric exception in RUNCOCOA
Mask invalid operation
%1100110011 >FPEXP            ok!!! $333

But see:
https://stackoverflow.com/questions/3378696/iphone-try-end-try/3379240#3379240
We should not recover from an exception in the ObjC Runtime.
Nice for debugging, but actually treat it as terminated/crashed.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

: >FPEXP ( n -- )   'FPOPT 2+ W! /NDP ;

: FPEXP ( -- )   %1100110010 >FPEXP ;      \ $332

: -FPEXP ( -- )  %1101111111 >FPEXP ;      \ $37F

\\ ( eof)