{ ====================================================================
OSX Frameworks

Copyright 2010 Roelf Toxopeus

SwiftForth version.
MacForth Framework (like) words, the OSX way for access to foreign
function libraries.
Simple, crude version making use of LIBRARY.
Last: 24 October 2020 at 15:27:34 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
FRAMEWORK -- create word named after an existing framework. When executed
the framework will be set as the current searched  library: 'LIB
In addition the actual framework or library will be added to SwiftForth's
internal library list. FRAMEWORK uses LIBRARY.

This version differs from previous, by adding LIBRARY semantics, but it
keeps the old way as well:
FRAMEWORK Carbon.framework -> execute Carbon.framework when existing <- new
                              else add to dictionary and execute it  <- old
Carbon.framework           -> execute it                             <- old

So the following is not necessary anymore:

LACKING System.framework  FRAMEWORK System.framework
System.framework

Just executing  FRAMEWORK System.framework  will cover both the undefined 
and defined cases. This is similar to LIBRARY System.framework/System

Note: FRAMEWORK (actualy dlopen) will search for frameworks in all the
default Frameworks directories:
/Library/Frameworks
/System/Library/Frameworks
~/Library/Frameworks

Problem with EVALUATE and LIBRARY ?
: FRAMEWORK ( SPACES<NAME> -- )
  >IN @ >R CREATE R> >IN !
  PARSE-WORD FRAMEWORK>LIBRARY COUNT 2DUP STRING, EVALUATE
  DOES> COUNT EVALUATE ;      \ this line is never executed...

Solving this with first compiling the DOES> part and then doing the
EVALUATE, see (FRAMEWORK)

Example FRAMEWORK usage:
framework Cocoa.framework  ok
.libs 
00900240 [REQUIRED]  /System/Library/Frameworks/Cocoa.framework/Cocoa
8FE46700 [REQUIRED]  libpthread.dylib
8FE46700 [REQUIRED]  libc.dylib ok
framework Carbon.framework  ok                                            
.libs 
0090CB50 [REQUIRED]  /System/Library/Frameworks/Carbon.framework/Carbon
00900240 [REQUIRED]  /System/Library/Frameworks/Cocoa.framework/Cocoa
8FE46700 [REQUIRED]  libpthread.dylib
8FE46700 [REQUIRED]  libc.dylib ok
function: NSRunAlertPanel ( ns" ns" ns" n n -- ret )  not in current library
cocoa.framework  ok
.libs 
0090CB50 [REQUIRED]  /System/Library/Frameworks/Carbon.framework/Carbon
00900240 [REQUIRED]  /System/Library/Frameworks/Cocoa.framework/Cocoa
8FE46700 [REQUIRED]  libpthread.dylib
8FE46700 [REQUIRED]  libc.dylib ok
function: NSRunAlertPanel ( ns" ns" ns" n n -- ret )  ok
-------------------------------------------------------------------- }

/FORTH
DECIMAL

CREATE FRAME.PAD 256 ALLOT
S" .framework" NIP CONSTANT #EXTCHARS
   
: FRAMEWORK>LIBRARY ( a1 n1 -- a2 )
   S" LIBRARY " FRAME.PAD DUP >R PLACE
   2DUP R@ APPEND                         \ Library xxx.framework
   S" /" R@ APPEND                        \ Library xxx.framework/
   #EXTCHARS - R@ APPEND                  \ Library xxx.framework/xxx
   R> ;

: (FRAMEWORK) ( a n spaces<name> -- )
   CREATE STRING,
   DOES> COUNT EVALUATE ;

: FRAMEWORK  ( spaces<name> -- )
   >IN @ >R  BL WORD FIND
   IF    R> DROP     EXECUTE
   ELSE  R> >IN !    COUNT FRAMEWORK>LIBRARY
         COUNT 2DUP (FRAMEWORK)   EVALUATE
   THEN ;


CR .( Framework stuff loaded)

\\ ( eof )

\ test

LIB-INTERFACE +ORDER
: LIB? ( -- )   BASE @ HEX 'LIB @REL ? BASE ! ;   \ print handle current lib
PREVIOUS

.libs 
01057ED8 [REQUIRED]  System.framework/System
01100FE0 [REQUIRED]  quartz.framework/quartz
01101060 [REQUIRED]  Cocoa.framework/Cocoa
01200340 [REQUIRED]  CoreFoundation.framework/CoreFoundation
01057ED8 [REQUIRED]  libm.dylib
0111A2D0 [REQUIRED]  libstdc++.6.dylib
01057ED8 [REQUIRED]  libpthread.dylib
01057ED8 [REQUIRED]  libc.dylib ok
  ok                                                                                               
lib? 1057ED8  ok    <- points to System.framework/System, libc.dylib, libpthread.dylib, libm.dylib
  ok
framework Carbon.framework  ok
  ok
' carbon.framework . 307103  ok   <- creates new dict. entree
  ok
.libs 
012042B0 [REQUIRED]  Carbon.framework/Carbon  <- handle added to libs list
01057ED8 [REQUIRED]  System.framework/System
01100FE0 [REQUIRED]  quartz.framework/quartz
01101060 [REQUIRED]  Cocoa.framework/Cocoa
01200340 [REQUIRED]  CoreFoundation.framework/CoreFoundation
01057ED8 [REQUIRED]  libm.dylib
0111A2D0 [REQUIRED]  libstdc++.6.dylib
01057ED8 [REQUIRED]  libpthread.dylib
01057ED8 [REQUIRED]  libc.dylib ok
  ok
lib? 12042B0  ok    <- points to Carbon.framework/Carbon
  ok
framework system.framework  ok <- no 'redefined' message, similar to LIBRARY
  ok
.libs 
012042B0 [REQUIRED]  Carbon.framework/Carbon
01057ED8 [REQUIRED]  System.framework/System   <- not added agin to lib list
01100FE0 [REQUIRED]  quartz.framework/quartz
01101060 [REQUIRED]  Cocoa.framework/Cocoa
01200340 [REQUIRED]  CoreFoundation.framework/CoreFoundation
01057ED8 [REQUIRED]  libm.dylib
0111A2D0 [REQUIRED]  libstdc++.6.dylib
01057ED8 [REQUIRED]  libpthread.dylib
01057ED8 [REQUIRED]  libc.dylib ok
  ok
lib? 1057ED8  ok    <- points to System.framework/System, libc.dylib, libpthread.dylib, libm.dylib
  ok
framework Carbon.framework  ok   <-  does not create new dict. entree, no 'redefine' message
  ok
lib? 12042B0  ok           <- points to Carbon.framework/Carbon
  ok
cocoa.framework  ok        <- execute framework, the current way
  ok
lib? 1101060  ok           <- points to Cocoa.framework/Cocoa,
  ok
carbon.framework  ok       <- execute framework, the current way
  ok
lib? 12042B0  ok           <- points to Carbon.framework/Carbon