{ ====================================================================
Mach kernel task and thread stuff

Copyright 2011 Roelf Toxopeus

SwiftForth version.
Last: 8 September 2020 at 14:20:44 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
Utillities for Mach tasks and threads.

THREAD-INFO -- return thread info block for given 4thtask.
See http://web.mit.edu/darwin/src/modules/xnu/osfmk/man/thread_info.html
and thread_info.h

These are all 32b integers or time structures of 2 32b integers
See coco-i4-64 version how to do it for 64b Forth
structure thread_basic_info
   time_value_t    user_time      \ user run time
   time_value_t    system_time    \ system run time
   integer_t       cpu_usage      \ scaled cpu usage percentage
   policy_t        policy         \ scheduling policy in effect
   integer_t       run_state      \ run state (see below)
   integer_t       flags          \ various flags (see below)
   integer_t       suspend_count  \ suspend count for thread
   integer_t       sleep_time     \ number of seconds that thread has been sleeping

THREAD? -- return thread running and scheduling state for given 4thtask.
Thread run states (state field).
1 CONSTANT TH_STATE_RUNNING	      \ thread is running normally
2 CONSTANT TH_STATE_STOPPED	      \ thread is stopped
3 CONSTANT TH_STATE_WAITING	      \ thread is waiting normally
4 CONSTANT TH_STATE_UNINTERRUPTIBLE \ thread is in an uninterruptible wait
5 CONSTANT TH_STATE_HALTED		      \ thread is halted at a clean point

Thread flags (flags field).
1 CONSTANT TH_FLAGS_SWAPPED	            \ thread is swapped out
2 CONSTANT TH_FLAGS_IDLE		            \ thread is an idle thread
4 CONSTANT TH_FLAGS_GLOBAL_FORCED_IDLE	\ thread performs global forced idle

#THREADS -- return threadlist and current number of threads in our
task/process. This is an ever changing number!
THREAD# -- return port number id from calling thread. Same number as
used by GDB for THREAD n.

I'M -- set current forth-task's mach thread name. While including this
file, it's assumed I'm Impostor...
-------------------------------------------------------------------- }

/FORTH
DECIMAL

LACKING SYSTEM.FRAMEWORK   FRAMEWORK System.framework

SYSTEM.FRAMEWORK

FUNCTION: thread_info ( target_thread flavor *info *count -- ret )
FUNCTION: mach_task_self ( -- mach_port_t )
FUNCTION: task_threads ( task *thread_list *thread_count -- ret )
FUNCTION: mach_thread_self ( -- thread_port )
FUNCTION: pthread_getname_np ( thread *name size -- ret )  
FUNCTION: pthread_setname_np ( name -- ret )      

10 CELLS CONSTANT /TINFO
/TINFO BUFFER: TINFO

3 CONSTANT THREAD_BASIC_INFO

: THREAD-INFO ( 4THtask -- *info count )
   CELL+ @ pthread_mach_thread_np 
   THREAD_BASIC_INFO 
   TINFO DUP /TINFO ERASE
   /TINFO >R RP@
   thread_info
   IF R> DROP 0 DUP
   ELSE TINFO R> THEN ;

: THREAD? ( 4THtask -- state flag )
   THREAD-INFO IF
      DUP [ 6 CELLS ] LITERAL + @
      SWAP [ 7 CELLS ] LITERAL + @
   ELSE DUP THEN ;

: #THREADS ( -- addr n )
   mach_task_self 0 >R RP@ 0 >R RP@
   task_threads DROP R> R> SWAP ;

: THREAD# ( -- n )
   mach_thread_self
   #THREADS 0 DO
      TUCK @ OVER =
      IF 2DROP I 1+  UNLOOP EXIT THEN
      SWAP CELL+
   LOOP DROP 0 ;

: I'M ( 0$name -- )   pthread_setname_np DROP ;

Z" impostor" I'M 

CR .( mach task and thread info loaded)

\\ ( eof )

\ example
task pipo
variable accu
: go pipo activate begin 1 accu +! pause again ;
go
pipo thread? swap . . \ expect: 1 0
pipo sleep
pipo thread? swap . . \ expect: 3 1
pipo wake
pipo thread? swap . . \ expect: 1 0
pipo done
pipo thread? swap . . \ expect: 0 0

