{ ====================================================================
fp precission conversion

Copyright 2002 Roelf Toxopeus

SwiftForth version.
A zoo of conversion words.
Comes from MacForth version, which uses PPC instructions.
Converting between single precision (SF) and native format precision (F) fp.
Many OS calls need and/or return single precision fp. These 32bit SF's
are passed via the data stack.
Also contains the integer <=> fp formats conversions.
Last: 8 April 2018 at 09:30:25 CEST  -rt
==================================================================== }

/FORTH
DECIMAL

{ --------------------------------------------------------------------
conversion between fp formats

These are the faster assembler versions. The Intel XCHG opcode is replaced
by a faster sequence of opcodes.
Also the stock SF S>F F>S D>F F>D versions are redefined: 3 times faster!

SF>F converts 32b SFLOAT on data stack to native FLOAT on fp stack.
F>SF does the reverse.
Note: F is not necessarely a DF. In SwiftForth it is, but not in iForth.

LOCATE S>F D>F F>D F>S o.s.s. for info on those.
-------------------------------------------------------------------- }

CODE SF>F ( sfloat -- ) ( F: -- r )
   EBX EAX MOV   0 [EBP] EBX MOV   EAX 0 [EBP] MOV   \ replacement for    0 [EBP] EBX XCHG
   0 [EBP] DWORD FLD  4 # EBP ADD  f>
   FNEXT

CODE F>SF ( -- sfloat ) ( F: r -- )
   >f  4 # EBP SUB  0 [EBP] DWORD FSTP
   EBX EAX MOV   0 [EBP] EBX MOV   EAX 0 [EBP] MOV   \ replacement for    0 [EBP] EBX XCHG
   FNEXT

\ --------------------------------------------------------------------
\ redefined SwiftForth versions:

CODE S>F ( n -- ) ( -- r )
   EBX EAX MOV   0 [EBP] EBX MOV   EAX 0 [EBP] MOV   \ replacement for   0 [EBP] EBX XCHG
   0 [EBP] DWORD FILD  4 # EBP ADD  f>
   FNEXT

CODE D>F ( d -- ) ( -- r )
   EBX EAX MOV   4 [EBP] EBX MOV   EAX 4 [EBP] MOV   \ replacement for   4 [EBP] EBX XCHG
   0 [EBP] QWORD FILD  8 # EBP ADD  f>
   FNEXT

CODE (F>S) ( -- n ) ( r -- )
   >f  4 # EBP SUB  0 [EBP] DWORD FISTP
   EBX EAX MOV   0 [EBP] EBX MOV   EAX 0 [EBP] MOV   \ replacement for   0 [EBP] EBX XCHG
   FNEXT

CODE (F>D) ( -- d ) ( r -- )
   >f  8 # EBP SUB  0 [EBP] QWORD FISTP
   EBX EAX MOV   4 [EBP] EBX MOV   EAX 4 [EBP] MOV   \ replacement for   4 [EBP] EBX XCHG
   FNEXT

: F>D ( -- d ) ( r -- )   TRUNCATE (F>D) ;
: F>S ( -- n ) ( r -- )   TRUNCATE (F>S) ;

{ --------------------------------------------------------------------
integer <=> (s)fp conversions

S>SF -- convert integer on data stack to 32b fp format on data stack.
SF>S -- convert 32b fp format on data stack to integer on stack.
-------------------------------------------------------------------- }

ICODE S>SF ( n -- sfloat )
   0 [EBP] EAX MOV      \ save second on stack
   EBX 0 [EBP] MOV      \ need EBP for addresing mode
   0 [EBP] DWORD FILD   \ convert integer to float
   0 [EBP] DWORD FSTP   \ return as sloat
   0 [EBP] EBX MOV      \ to top
   EAX 0 [EBP] MOV      \ restore second on stack
   FNEXT

ICODE SF>S ( sfloat -- n )
   0 [EBP] EAX MOV      \ save second on stack
   EBX 0 [EBP] MOV      \ need EBP for addressing mode
   0 [EBP] DWORD FLD    \ load fp
   0 [EBP] DWORD FISTP  \ for conversion to integer
   0 [EBP] EBX MOV      \ is top stack
   EAX 0 [EBP] MOV      \ restore second on stack
   FNEXT

{ --------------------------------------------------------------------
passing sfloat and dfloats to calls

Many could be just aliasses for existing words, but might involve special
handling wrt call ABI on other architectures: 32b SF and VFX - 64b iForth
FPUSHS 2FPUSHS 4FPUSHS pass fp numbers as sf's on data stack.
The names are Marcel Hendrix's from iForth, the original MacForth names
were less attractive.
The fast assembler 4FPUSHS learned from RickVanNorman in the Faq-o-Matic
OpenGL Spin Cube. 2FPUSHS and FPUSHS are derived from that version.

FPUSHD 2FPUSHD 4FPUSHD and DF>F are aliasses for George Kozlowski
fp stack passing words found in fp-passing.f
Each DF consumes 2 cells on the data stack, take care when defining
functions and methods!
-------------------------------------------------------------------- }

\ --------------------------------------------------------------------
\ fp to sfloat on data stack conversion

(* skip hi level versions
\ Here alias for F>SF hilevel version which is faster than assembler version?!?!?
: FPUSHS ( -- sfloat ) ( F: r -- )   0 >r rp@ sf! r>  ;

: 2FPUSHS ( -- sf1 sf2 ) ( F: r1 r2 -- )
   fswap fpushs fpushs ;
   
: 4FPUSHS ( -- sf1 sf2 sf3 sf4 ) ( F: r1 r2 r3 r4 -- )
   f2swap 2fpushs 2fpushs ;
*)

\ much faster code definition.
CODE 4FPUSHS ( -- sf1 sf2 sf3 sf4 ) ( F: r1 r2 r3 r4 -- )
   4 >fs                \ make sure data on hardware stack
   16 # EBP SUB         \ room for 4 integers and tos
   12 [EBP] DWORD FSTP  \ convert t
    0 [EBP] DWORD FSTP  \ convert z
    4 [EBP] DWORD FSTP  \ convert y
    8 [EBP] DWORD FSTP  \ convert x
   12 [EBP] EBX XCHG    \ swap t and old tos
   FNEXT

\ Note: the XCHG replacement pays off!
CODE 2FPUSHS ( -- sf1 sf2 ) ( F: r1 r2 -- )
   2 >fs               \ make sure data on hardware stack
   8 # EBP SUB         \ room for 2 integers and tos
   4 [EBP] DWORD FSTP  \ convert t
   0 [EBP] DWORD FSTP  \ convert z
   EBX EAX MOV   4 [EBP] EBX MOV   EAX 4 [EBP] MOV   \ replacement for   4 [EBP] EBX XCHG    \ swap t and old tos
   FNEXT

\ this one is as fast as the hilevel version
AKA F>SF  FPUSHS  ( -- sfloat ) ( F: r -- )

\ --------------------------------------------------------------------
\ fp to dfloat on data stack conversion

AKA FSTACK>STACK    FPUSHD  ( -- dfloat )  ( F: r -- )
AKA 2FSTACK>STACK   2FPUSHD ( -- dfloat1 dfloat2 )  ( F: r1 r1 -- )
AKA 4FSTACK>STACK   4FPUSHD ( -- dfloat1 dfloat2 dfloat3 dfloat4 )  ( F: r1 r2 r3 r4 -- ) 
AKA STACK>FSTACK    DF>F ( dfloat -- )  ( F: -- r )
AKA FSTACK>STACK    F>DF ( -- dfloat )  ( F: r -- )

CR .( faster fp precision loaded)

\\ ( eof )