{ ====================================================================
Multi line string

SwiftForth version

Copyright 2018 Roelf Toxopeus

Last: 5 April 2018 at 23:12:50 CEST
==================================================================== }

{ --------------------------------------------------------------------
PARSE-LINES -- parses text using passed delimiter over multiple lines.
               Returns char delimited string pair at transient HERE.
            
LINES, -- compiles parsed delimited text as a zero terminated counted
          string. Returns string pair ( -- a n ) at runtime.
          
S` -- parses ` delimited string over multiple lines. Depending on STATE,
      compiles it or returns it as a string pair in transient QPAD.

Easy to create other delimiter versions. Whatever is appropriate:
: S~ ( "ccc"<~> -- )   [CHAR] ~  POSTPONE LINES, ; IMMEDIATE

: S| ( "ccc"<|> -- )   [CHAR] |  POSTPONE LINES, ; IMMEDIATE

: S" ( "ccc"<"> -- )   [CHAR] "  POSTPONE LINES, ; IMMEDIATE

etc.

Useful for Applescript where linefeeds are crucial.

: "test" ( -- za )
   S` tell application "Terminal"
   activate
   do script with command "ls;sleep 4;exit"
   end tell`
   DROP ;

: TT ( -- )   "test" DOSCRIPT ;

Mach2 and MacForth had S" C" and 0" implemented for multiple lines.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

: PARSE-LINES ( delimiter "ccc"<char> -- a n )
   HERE >R                                            \ construct counted string here
   S" " R@ PLACE                                      \ start with empty string
   BEGIN 
      DUP DUP PARSE                                   \ parse till delimiter or end of line
      2DUP R@ APPEND                                  \ whatever we got, append to string
      + C@ <>                                         \ if the delimiter is parsed we're done
   WHILE
      <EOL> COUNT R@ APPEND                           \ not a delimiter so a linefeed is added to string
      SOURCE-ID 0= IF CR THEN                         \ at the keyboard, do a cr
      REFILL 0= ABORT" Unexpected End Of String !"    \ refill next line or abort 
   REPEAT
   DROP                                               \ drop delimiter
   R> COUNT ;                                         \ return string pair

: LINES, ( delimiter "ccc"<char> -- )
   POSTPONE (S")                                      \ runtime for string
   PARSE-LINES                                        \ parse delimiter over lines
   NIP 1+ ALLOT                                       \ allot memory for string
   0 C, ; IMMEDIATE                                   \ zero terminated

: S` ( "ccc"<`> -- )
   [CHAR] ` STATE @ IF POSTPONE LINES, EXIT THEN
   PARSE-LINES >QPAD COUNT ; IMMEDIATE

\\
   
\ compiled:
: ll ( -- a n )
   s` this is a
string, which has
more lines` ;

ll type this is a
string, which has
more lines ok

\ interpreted:
s` this is                                       
   a formatted 
   string over 
   five lines                                    
ending here` cr type                             
this is
   a formatted
   string over
   five lines
ending here ok
