{ ====================================================================
waypoints

Copyright 2010 Roelf Toxopeus

SwiftForth version.
Useful words to set current working directory, works like setting
a context vocabulary. Also working directory utilities.
Last: 12 December 2019 at 21:48:01 CET  rt
==================================================================== }

/FORTH
DECIMAL

{ --------------------------------------------------------------------
MAX-PATH -- returned value inherited from PATH_MAX found in
usr/include/sys/syslimits.h
However, most words defined here and elsewhere assume path is 255 chars max.
PATHPAD -- buffer for creating and receiving path strings.

CHWD -- changes working directory, takes care of saving cwd before changing it.
Like CD but takes path from the stack. Name is a joke of course, chew on it ;-)

CWD -- get current working directory in PATHPAD (1024 max), zero terminated.
WHAMI -- returns current working directory as string pair.
Note: both using the PADPATH instead of PAD. Could be a factor of PWD,
which prints the string.
-------------------------------------------------------------------- }

1024 CONSTANT MAX-PATH
MAX-PATH BUFFER: PATHPAD

: CHWD ( a n -- )   (CD) !OLDPWD  chdir ABORT" Invalid Directory !" ;

: (CWD) ( -- zstring )   PATHPAD MAX-PATH getcwd ;

: CWD ( -- zstring )   (CWD) >R S" /" R@ ZAPPEND R> ;

: WHAMI ( -- a n )   (CWD) ZCOUNT ;

{ --------------------------------------------------------------------
WAYPOINT -- used to create named 'waypoints' which when executed will
change the working directory to the path given at initialisation.
Originally a navigational aid used at sea and in the air.
Example usage:  WHAMI  WAYPOINT  MyPivot

MAC is the pivotal directory for the Cocoa package. It uses MACPATH,
set during compilation of coco-sf and relative to ROOTPATH. It is
assumed you keep the SwiftForth and mac folders in the same relation
as during compilation of coco-sf. Otherwise MACPATH is invalid.
SWIFTFORTH, ROOT, HOME and DESKTOP allow for quick and easy access to
these default directories.
Note: ROOTPATH contains full path up till the SwiftForth folder.
So as alternative   : ROOT ( -- )   ROOTPATH COUNT CHWD ;

From MacForth: MYFOLDER -- set directory from file being included as
the current working directory. Further including can commence from this
directory as parent. Only from within INCLUDED/LOADED file!!!
INCLUDING ought to return a zero string when nothing is INCLUDED? It
appears it returns a 1 length zero string, so check for anything bigger
than 1.
Uses copy of returned path by INCLUDED, so original isn't touched!

Avoid getting lost, use with care.
-------------------------------------------------------------------- }

: WAYPOINT ( a n <name> -- )
   CREATE STRING,
   DOES> COUNT CHWD ;

MACPATH COUNT +ROOT   WAYPOINT MAC
S" %SwiftForth" +ROOT WAYPOINT SWIFTFORTH
S" %" +ROOT           WAYPOINT ROOT
$HOME ZCOUNT          WAYPOINT HOME
$HOME ZCOUNT PAD PLACE
S" /Desktop" PAD APPEND
PAD COUNT             WAYPOINT DESKTOP

: MYFOLDER ( -- )
   INCLUDING DUP 2 < ABORT" Only use MYFOLDER while loading a file"
   PATHPAD PLACE PATHPAD COUNT
   -NAME CHWD ;

CR .( Waypoints loaded)

\\ ( eof )
