{ ====================================================================
Better Maps

Copyright 2011 Roelf Toxopeus

SwiftForth version.
Turnkey Cocoa extended Darwin Forth, here SwiftForth OSX. The new app
is running a webwindow with OpenStreepMaps: MAPS.
Last: 17 September 2020 at 08:20:56 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
Example of creating an application.

Using vanilla sf, include this file and in 6 steps a native macOS
application bundle is created: MAPS. Which is coco-sf running maps
with its own webview window and menubar independently from Terminal.app,
which is also not used for launching MAPS.app.

Portability:
Each Darwin Forth has its own words to create a turnkey, so re-code
here for specific Forth system.

Directory independant:
You can place the mac folder and this file wherever you want.
Put them in  %swiftforth/lib/samples/macos/  or  %swiftforth/../
Important: keep this file and the mac folder together. This file must
be in the parent folder of the mac folder, to work out of the box.
Of course you can change things...

Application bundle notes:
The interpreter is not used. 'MAIN is overwritten directly. We never
get out of 'MAIN except back to the desktop. Our main routine doesn't
call STARTER or any of the interpreter related words. So no TTY,
no console, no QUIT loop. The IMPOSTOR thread is not used either.
The initialization GOES.MAPS is done on the main thread OPERATOR by
queueing it on the main event runloop. The ObjC Runtime takes care,
when it starts running. There is no more Forth to execute.
(Compare with the STARTER in new-toy.ldr)

See info on 'MAIN and STARTER in:
SwiftForth/src/kernel/macos/start.f   executes 'MAIN in START
SwiftForth/src/kernel/start.f         defines 'MAIN
SwiftForth/src/ide/ul/startup.f       defines 'STARTER and
                                      init 'STARTER and 'MAIN

Usage: run maps.app
-------------------------------------------------------------------- }

CR .( -------------------------------------------------------------- )
CR .( 1. Including the Forth, OS and Cocoa extensions and interfaces)

/FORTH
DECIMAL

INCLUDING -NAME PAD FULLNAME
S" /mac/" PAD ZAPPEND
PAD chdir DROP                       \ change directory to /wherever/mac

INCLUDE mac-sf.f                     \ the regular coco-sf stuff

CR .( -------------------------------------------------------------- )
CR .( 2. The added GUI items and extras on top of coco-sf)

PUSHPATH MAC
INCLUDE hotcoco/simplemenu.f            \ main menubar
INCLUDE hotcoco/cocoawebview.f          \ webview and MAPS !
POPPATH

cr .( --------------------------------------------------------------------)
cr .( 3. Necessary redefinitions)

\ hitting close box should terminate app using default self delegate
COCOA: @terminate: ( sender -- )   \ won't return

FORTHCALL: *windowShouldClose: ( rec sel sender:button -- bool )   _PARAM_2 NSApp @ @terminate: YES ; 

: DELEGATETYPES   0" c@:@" ;

*windowShouldClose: DELEGATETYPES 0" windowShouldClose:"  NSWindow ADD.METHOD

\ bigger webwin
: /WEBWIN ( -- )   100e0 600e0 1057e0 700e0 WEBWIN  W.BOUNDS ;

\ like PASS, queue on main runloop, but don't wait
: -PASS ( #ints #floats xt -- )
   'PASS GET
   IS FWORD !PARAMETERS
   Z" doForth:" @selector 0 NO ForthClass FORMAIN DROP
   'PASS RELEASE ;

CR .( -------------------------------------------------------------- )
CR .( 4. Runner)

\ two definitions to make it work out of the box:
\ 1. like GOES.FORTH, but for MAPS only. NONAP is not used.
: GOES.MAPS ( -- )   \ modelled after GOES.FORTH
   /SIMPLEMENU  \ toy app inits: add simple menu
   /WEBWIN      \ toy app inits: set webwin size
   MAPS         \ toy app inits: open webview with map
;
 
\ 2. like COCO.RUN, but not using impostor. Queue GOES.MAPS on main runloop.
: MAPS.RUN ( -- )   INIT.COCOA 0 DUP ['] GOES.MAPS -PASS RUNCOCOA ;

\ our main
' MAPS.RUN 'MAIN !

GILD

CR .( -------------------------------------------------------------- )
CR .( 5. Turnkey app)

PUSHPATH
MAC
INCLUDE bundling/bundling.f          \ necessary turnkey tools
POPPATH

\ Create our app on the desktop
DESKTOP

TURNKEY maps

CR .( -------------------------------------------------------------- )
CR .( 6. Add resources and corrections to bundle)

\ in my mac/resources folder
S" OSM.icns"   S" maps"  COPY-RESOURCE

\ make corrections
s" maps.app/Contents/Resources/OSM.icns" PAD FULLNAME                                               
s" maps.app/Contents/Resources/app.icns" PATHPAD FULLNAME
PAD ZCOUNT PATHPAD ZCOUNT RENAME-FILE DROP

CR .( -------------------------------------------------------------- )
CR .( 7. Quit)

BYE

\\ ( eof )