{ ====================================================================
Padbury Clock screen saver

Copyright 2011 Roelf Toxopeus

SwiftForth version.
Turnkey Cocoa extended Darwin Forth, here SwiftForth OSX. The new app
is a screensaver. The Padbury Clock was a favourite screensaver which
stopped working in macOS 10.14 Mojave. It's revived here.
Last: 30 July 2020 at 16:00:37 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
Example of creating an application.

Using vanilla sf, include this file and in 6 steps a native macOS
application bundle is created: padbury-clock. Which is coco-sf running
a Quartz composition by Robert Padbury. It shows the time in hours,
minutes and seconds, independently from Terminal.app, which is also not
used for launching padbury-clock.app.
https://dribbble.com/padbury
https://www.screensaversplanet.com/screensavers/padbury-clock-1027/

Portability:
Each Darwin Forth has its own words to create a turnkey, so re-code
here for specific Forth system.

Directory independant:
You can place the mac folder and this file wherever you want.
Put them in  %swiftforth/lib/samples/macos/  or  %swiftforth/../
Important: keep this file and the mac folder together. This file must
be in the parent folder of the mac folder, to work out of the box.
Of course you can change things...

Application bundle note:
The interpreter is not used. SCREEN.SAVER is a variation on HOT.COCO 
but plugged in 'MAIN not in 'STARTER. It never gets out of 'MAIN except
back to the desktop. It doesn't call STARTER or any of the interpreter
related words. So no TTY, no console, no QUIT loop. IMPOSTOR will be
activated and after the necessary initializations starts off the
Quartz Composition and will schedule the ObjC Runtime activities.

See info on 'MAIN and STARTER in:
SwiftForth/src/kernel/osx/start.f   executes 'MAIN in START
SwiftForth/src/kernel/start.f       defines 'MAIN
SwiftForth/src/ide/ul/startup.f     defines 'STARTER and
                                    init 'STARTER and 'MAIN

Usage: run padbury-clock.app
mouse click, pad tap, cmd-q or force-quit ends screen saver
-------------------------------------------------------------------- }

CR .( -------------------------------------------------------------- )
CR .( 1. Including the Forth, OS and Cocoa extensions and interfaces)

/FORTH
DECIMAL

INCLUDING -NAME PAD FULLNAME
S" /mac/" PAD ZAPPEND
PAD chdir DROP                       \ change directory to /wherever/mac

INCLUDE mac-sf.f                     \ the regular coco-sf stuff

CR .( -------------------------------------------------------------- )
CR .( 2. The added GUI items and extras on top of coco-sf)

PUSHPATH MAC
INCLUDE hotcoco/simplemenu.f            \ main menubar
POPPATH

CR .( -------------------------------------------------------------- )
CR .( 3. The screensaver stuff on top of coco-sf)

PUSHPATH MAC
INCLUDE hotcoco/cursor.f
INCLUDE hotcoco/padbury-clock.f
POPPATH

CR .( -------------------------------------------------------------- )
CR .( 4. Runner)

\ two definitions to make it work out of the box:
\ 1. first thing to do in Forth thread

: RUN.SAVER ( -- )   \ modelled after GOES.FORTH
   PAUSE                             \ give time to OPERATOR, i.e. NSApp
   /RND                              \ coco-sf inits: prng
   NONAP                             \ coco-sf inits: appnap setting
   /SIMPLEMENU                       \ screensaver: simple menu
   HIDE-MOUSE                        \ screensaver: hide cursor
   0 DUP ['] NOP PASS                \ High Sierra fix
   FPEXP                             \ Forth thread can receive numeric exceptions
   PAUSE                             \ give time to OPERATOR, i.e. NSApp
   XX ;                              \ screensaver: init and runner
 
\ 2. starter a la HOT.COCO but put in 'MAIN
: SCREEN.SAVER ( -- )   ['] RUN.SAVER 'PRETENDING ! COCO.RUN ;

\ our main
' SCREEN.SAVER 'MAIN !

GILD

CR .( -------------------------------------------------------------- )
CR .( 5. Turnkey app)

PUSHPATH
MAC
INCLUDE bundling/bundling.f          \ necessary turnkey tools
POPPATH

\ Create our app in ROOTPATH:
ROOT

TURNKEY padbury-clock

CR .( -------------------------------------------------------------- )
CR .( 6. Add resources and corrections to bundle)

\ in my mac/resources folder
S" 24_Hour_Clock.qtz"  S" padbury-clock"  COPY-RESOURCE
S" padbury.icns"       S" padbury-clock"  COPY-RESOURCE

\ make corrections
s" padbury-clock.app/Contents/Resources/padbury.icns" pad fullname                                               
s" padbury-clock.app/Contents/Resources/app.icns" pathpad fullname
pad zcount pathpad zcount rename-file drop

CR .( -------------------------------------------------------------- )
CR .( 7. Quit)

BYE

\\ ( eof )