{ ====================================================================
toy application

Copyright 2011 Roelf Toxopeus

SwiftForth version.
Turnkey Cocoa extended Darwin Forth, here SwiftForth for macOS.
The new coco-sf is a bundled application. A double clickable icon: sf-app.
Last: 7 August 2020 at 15:36:06 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
Example of creating an application.
 
Using vanilla sf, include this file and in 6 steps a native macOS
application bundle is created: sf-app. Which is coco-sf with its own
I/O window and menubar independently from Terminal.app, which is not
used at all.

Portability:
Each Darwin Forth has its own words to create a turnkey, so re-code
here for specific Forth system.

Directory independant:
You can place the mac folder and this file wherever you want.
Put them in  %swiftforth/lib/samples/macos/  or  %swiftforth/../
Important: keep this file and the mac folder together. This file must
be in the parent folder of the mac folder, to work out of the box.
Of course you can change things...

Application bundle notes:
Like coco-sf, (redefined) GOES.FORTH initializes Forth on IMPOSTOR,
but the redefined HOT.COCO is put in 'MAIN not 'STARTER. The Terminal
console and the words needed to initialize it are not used.

macOS specifics: app bundled applications launched without Terminal
involved while launching, f.i. double clicked or ctrl-click-open in
Finder, have a different PATH environmental variable than those where
Terminal was involved.
My global PATH=/usr/local/bin:/usr/bin:/bin:/usr/sbin:/sbin is not
propagated! Instead PATH=/usr/bin:/bin:/usr/sbin:/sbin is used. This
means that the editor executable can't be found!
/PATHENV takes care of that. Supply your own path, see utils/app-PATH.f

Words defined in a file which executes PROGRAM can't be located.
Their sourcefile won't be registered, until INCLUDE-FILE in INCLUDED
is finished. This is remedied by executing  F# @ 'FNAME @  =FILENAME
inside the file executing PROGRAM.

When you put the bundled application in the same parent folder as your
SwiftForth folder, you enjoy the benefits of EDIT, LOCATE, G and the
Help menu's. The folder structures are similar, same depths.
Conveniently this is ROOTPATH.

WARNING: this version of coco-sf *is*a*toy* version. It uses a very
primitive I/O window compared with the Terminal window as used by sf
and the regular coco-sf. A lot more additions and debugging should be
applied. Use at your own risk.
-------------------------------------------------------------------- }

CR .( -------------------------------------------------------------- )
CR .( 1. Including the Forth, OS and Cocoa extensions and interfaces)

/FORTH
DECIMAL

INCLUDING -NAME PAD FULLNAME
S" /mac/" PAD ZAPPEND
PAD chdir DROP                       \ change directory to /wherever/mac

INCLUDE mac-sf.f                     \ the regular coco-sf stuff

CR .( -------------------------------------------------------------- )
CR .( 2. The added GUI items and extras on top of coco-sf)

PUSHPATH MAC
INCLUDE utils/app-PATH.f             \ set $PATH for app
INCLUDE hotcoco/sfmenu.f             \ main menubar
INCLUDE hotcoco/shellview.f          \ textview I/O console
INCLUDE utils/myok.f                 \ for fun the Mach2 prompt
INCLUDE hotcoco/included-menu.f      \ the included menu
POPPATH

CR .( -------------------------------------------------------------- )
CR .( 3. Necessary definitions and redefinitions)

\ add our PATH
: /PATHENV ( -- )   S" /usr/local/bin" PATHENV+ ;

\ can't return to terminal console
: /TERMINAL ( -- )   1 ABORT" Can't !" ;

\ need a new BYE, which stops the NSApplication coco-sf
\ not using terminal console, no tty etc.
: (APP-BYE) ( -- )   'ONSYSEXIT CALLS  NSApp @ DUP @terminate: ;

\ all words including/loading file should add that file to the Included menu.
: LOAD-FILE ( -- )    LOAD-FILE >INCLUDED ;
: LOAD-EDIT ( -- )    LOAD-EDIT >INCLUDED ;
: INCLUDE-FILE ( fid -- )   INCLUDE-FILE >INCLUDED ;
: INCLUDED ( c-addr u -- )   INCLUDED >INCLUDED ; 
: INCLUDE ( <spaces>filepath -- )   INCLUDE >INCLUDED ;

CR .( -------------------------------------------------------------- )
CR .( 4. Runner)

\ two redefinitions to make it work out of the box:
\ 1. first thing to do in Forth thread
: GOES.FORTH ( -- )
   PAUSE
   /PATHENV                          \ experimental toy app init: PATH correction
(* -- not using terminal console, no tty etc.  
   /CONSOLE                          \ regular sf inits
   /ONLOAD  'ONENVLOAD CALLS         \ regular sf inits
   /INTERPRETER  /CMDLINE            \ regular sf inits, QUIT executes /INTERPRETER as well -rt
*)
   /RND                              \ coco-sf inits: prng
   NONAP                             \ coco-sf inits: appnap setting
   SFMenuClass ADD.CLASS SFMENU      \ toy app inits: menu class
   ['] (APP-BYE) IS BYE              \ toy-app inits: quit app
   PUSHME                            \ toy app inits: make our app the front app
   /OK                               \ toy app inits: mach2 prompt
   INCLUDED>MENU                     \ toy app inits: included menu
   FPEXP                             \ Forth thread can receive numeric exceptions
   SFShellView ADD.CLASS             \ toy app inits: add shellview class   
   /COCOA-CONSOLE ;                  \ toy app inits: the cocoa console running Forth
 
\ 2. same as the original version, but uses new GOES.FORTH
: HOT.COCO ( -- )   ['] GOES.FORTH 'PRETENDING ! COCO.RUN ;

\ overwrite previous 'MAIN
' HOT.COCO 'MAIN !

\ nothing for STARTER
STARTER NOOP

\ keep current included filename for locate
F# @ 'FNAME @  =FILENAME

GILD

CR .( -------------------------------------------------------------- )
CR .( 5. Turnkey app)

PUSHPATH
MAC
INCLUDE bundling/bundling.f          \ necessary turnkey tools
POPPATH

\ Create our app in ROOTPATH:
ROOT

TURNKEY sf-app

CR .( -------------------------------------------------------------- )
CR .( 6. Add sfmenu.nib and app.icns to bundle)

\ assumes these 'files' exists in the default resources folder:
S" sfmenu.nib" S" sf-app"  COPY-RESOURCE
S" app.icns"   S" sf-app"  COPY-RESOURCE

\ any bundle corrections and additions can be done here:
CR .( Extra: add the nib files needed for the hotcoco examples)
S" MyButtons.nib"       S" sf-app"  COPY-RESOURCE
S" MyButtonsTagged.nib" S" sf-app"  COPY-RESOURCE
S" mycontroller.nib"    S" sf-app"  COPY-RESOURCE
S" runweb-sf.nib"       S" sf-app"  COPY-RESOURCE
S" testcontroller.nib"  S" sf-app"  COPY-RESOURCE

CR .( -------------------------------------------------------------- )
CR .( 7. Quit)

BYE

\\ ( eof )