{ ====================================================================
Program loader and file save

Copyright 2020 by FORTH, Inc.

This file is common to all the target compiler and turnkey image
generators. The loader file and its header field base class as well as
object file generation are all defined here.
==================================================================== }

PACKAGE PROGRAM-GENERATOR

{ --------------------------------------------------------------------
LOADER base class

'STRUCT is a pointer to the base of a header struct in a buffer.

FIELD: defines fields within the struct. Takes offset u1, size (in
bytes) u2, and returns next offset u3.
-------------------------------------------------------------------- }

CLASS LOADER

VARIABLE 'STRUCT

: SET ( addr -- )   'STRUCT ! ;

: FIELD: ( u1 u2 -- u3 )
   CREATE OVER , +
   DOES> ( -- addr )   @ 'STRUCT @ + ;

END-CLASS

{ ------------------------------------------------------------------------
File image

ALLOCATE-IMAGE allocates memory for the object file whose size is u
and leaves its address in >FILEIMAGE and the size in #FILEIMAGE.
Allocates enough memory for the img file plus SwiftForth program image.

SAVE-IMAGE parses the object filename, creates and writes the file.

READ-IMAGE opens the binary loader file whose name is in the string
addr len, determines its size, allocates a buffer in extended memory
for the loader image plus u bytes, and reads the image into the
buffer. Takes size u of SwiftForth program image, rounded to page size
and file name addr len.

+IMG returns an address in the object file memory image.
CHUNK rounds u1 up to u2 chunk size.
------------------------------------------------------------------------ }

VARIABLE 'FILEIMAGE     \ Address of new file image buffer
VARIABLE #FILEIMAGE     \ Total size of image buffer

: ALLOCATE-IMAGE ( u -- )
   DUP ALLOCATE THROW  'FILEIMAGE !  #FILEIMAGE ! ;

PUBLIC

: SAVE-IMAGE ( -- )
   BL WORD COUNT +ROOT R/W EXE CREATE-FILE
   ABORT" Can't create object file" >R
   'FILEIMAGE @ #FILEIMAGE @ R@ WRITE-FILE
   R> CLOSE-FILE DROP  THROW  'FILEIMAGE @ FREE DROP ;

VARIABLE #LOADER

: READ-IMAGE ( u addr len -- )
   +ROOT R/O OPEN-FILE THROW >R                 \ Open loader binary image
   R@ FILE-SIZE THROW  DROP  DUP #LOADER !      \ Loader size
   + ALLOCATE-IMAGE                             \ Allocate buffer for object file image
   'FILEIMAGE @ #LOADER @ R@ READ-FILE
   R> CLOSE-FILE DROP  THROW DROP ;

: +IMG ( n -- addr)   'FILEIMAGE @ + ;

: CHUNK ( u1 u2 -- u3 )   SWAP OVER /MOD SWAP 0<> - * ;

END-PACKAGE
