{ ====================================================================
Save executable turnkey image

Copyright 2011  FORTH, Inc.

This file supplies the utility for generating an executable turnkey
program image.  The SwiftForth program is written into a precompiled
loader image read in from the file %SwiftForth/bin/linux/sf-loader.bin.

==================================================================== }

?( Save turnkey)

PACKAGE PROGRAM-GENERATOR

{ ---------------------------------------------------------------------
Loader file header

These objects, pointers, and offsets are used to access the sf-loader
header image.

(FIND-HEADER) scans the loader header in memory for the given string.
Aborts if not found.

VALIDATE-HEADER checks that the image header starts with MH_MAGIC.
--------------------------------------------------------------------- }

HEX

464C457F CONSTANT ESIG  \ ELF file signature
444AAA CONSTANT FSIG    \ Loader p_memsz signaure for Forth segment
VARIABLE 'PM            \ Address of p_memsz field

: ?BADIMG ( flag -- )   ABORT" Bad loader image" ;

PUBLIC

VARIABLE 'SF            \ Address of Forth image

: p_filesz ( -- addr )   'PM @ CELL- ;          \ p_filesz field in Forth segment
: p_memsz ( -- addr )   'PM @ ;                 \ p_memsz field in Forth segment

: SCAN-HEADERS ( -- )
   0 +IMG @ ESIG <> ?BADIMG
   -1  #LOADER @ +IMG 10 +IMG DO  I @ FSIG = IF
   I 'PM !  1+  LEAVE  THEN  4 +LOOP  ?BADIMG
   0 +IMG #LOADER @ S" FORTHIMG" SEARCH 0= ?BADIMG
   DROP  'SF ! ;

{ --------------------------------------------------------------------
Save program

PROGRAM saves the turnkey.
-------------------------------------------------------------------- }

: PROGRAM ( -- )
   HERE DP0 !REL  HERE SAVE-XREF                \ xref table
   HERE -ORIGIN OVER +
   S" %SwiftForth/bin/linux/sf-loader.bin"
   READ-IMAGE  SCAN-HEADERS                     \ Read the binary loader program image into memory, scan headers
   ORIGIN 'SF @ HERE -ORIGIN 2DUP 2>R CMOVE     \ Copy SwiftForth into file image
   HERE -ORIGIN 8 - OVER + p_filesz +!          \ File size of SwiftForth image (less 8 for "FORTHIMG")
   2R> + SWAP CMOVE                             \ Copy xref table to image
   $4000000 p_memsz !  SAVE-IMAGE ;             \ Virtual memory 64MB, save image

DECIMAL

END-PACKAGE
