{ ====================================================================
Save executable turnkey image

Copyright 2011  FORTH, Inc.

This file supplies the utility for generating an executable turnkey
program image.  The SwiftForth program is written into a precompiled
loader image read in from the file %SwiftForth/bin/macos/sf-loader.bin.

==================================================================== }

?( Save program )

PACKAGE PROGRAM-GENERATOR

{ ---------------------------------------------------------------------
Loader file header

These objects, pointers, and offsets are used to access the sf-loader
header image.

(FIND-HEADER) scans the loader header in memory for the given string.
Aborts if not found.

VALIDATE-HEADER checks that the image header starts with MH_MAGIC.
--------------------------------------------------------------------- }

PUBLIC

SEGMENT BUILDS __DATA           \ __DATA segment
SECTION BUILDS __FORTH          \ __data section (container for Forth image)
SEGMENT BUILDS __LINKEDIT       \ __LINKEDIT section
SECTION BUILDS __text           \ __text section

VARIABLE 'SF                    \ Address of Forth image
VARIABLE >LE                    \ Offset to __LINKEDIT section in loader image

PRIVATE

: (FIND-HEADER) ( addr1 len -- addr2 )
   'FILEIMAGE @ #LOADER @ 2SWAP SEARCH 0= ABORT" Loader section not found" DROP ;

PUBLIC

: SCAN-HEADER ( -- )
   0 +IMG L@ MH_MAGIC <> ABORT" Bad loader image"
   S" __DATA" (FIND-HEADER) __DATA SET
   S" __data" (FIND-HEADER) __FORTH SET
   S" __LINKEDIT" (FIND-HEADER)  __LINKEDIT SET
   S" __text" (FIND-HEADER) __text SET
   S" FORTHIMG" (FIND-HEADER) 'SF !
   __LINKEDIT fileoff L@  >LE !
;

: FIX-DYLD ( n -- )
   __text offset @ +IMG  __LINKEDIT vmaddr DO
      I L@  >LE @ #LOADER @ WITHIN IF  DUP I L+!
   THEN  4 +LOOP DROP ;

{ --------------------------------------------------------------------
Save program

PROGRAM saves the turnkey.
-------------------------------------------------------------------- }

HEX

: PROGRAM ( -- )
   HERE -ORIGIN #XREF + 1000 CHUNK  DUP >R      \ Extra size to insert SwiftForth image
   S" %SwiftForth/bin/macos/sf-loader.bin"      \ Loader image file
   READ-IMAGE  SCAN-HEADER                      \ Allocate image, read loader, scan header for markers
   >LE @ +IMG  DUP R@ +  #LOADER @ >LE @ - CMOVE>  \ Move dyld tables to make room for SwiftForth image
   HERE DP0 !REL  HERE SAVE-XREF                \ ( addr u) xref table
   'SF @  HERE -ORIGIN 2>R                      \ Dest addr in program image and size of SwiftForth
   ORIGIN 2R@ CMOVE  2R> + SWAP CMOVE           \ Copy SwiftForth and xrefs into file image
   $1000000 __DATA vmsize !                     \ set vmsize of __DATA segment (16MB)
   R@ __DATA filesize +!                        \ increase filesize of __DATA segment
   HERE -ORIGIN #XREF + __FORTH size !          \ size of __data section (actual size in file)
   $1000000 __LINKEDIT vmaddr +!                \ vmaddr of __LINKEDIT segment
   R> FIX-DYLD                                  \ fix all the fileoff fields for dyld tables
   SAVE-IMAGE ;                                 \ Save binary image

DECIMAL

END-PACKAGE
