{ ====================================================================
User input

Copyright 2001  FORTH, Inc.
Rick VanNorman

ACCEPT and REFILL are defined here.
==================================================================== }

TARGET

{ ------------------------------------------------------------------------
Accept

ACCEPT  The built-in function is very simple -- it responds to only
the backspace, escape, and carriage return controls.

Backspace does the obvious, moving the cursor backwards until it
reaches the beginning of the line.

Escape kills all characters in the current line.

Carriage return ends the input, leaving the number of characters
accepted on the stack.

ACCEPT is deferred and replaced with a better version later.

TIB is the address of the current terminal input buffer.
------------------------------------------------------------------------ }

| : .BS ( lim addr n char -- lim addr n' )
   DROP  DUP IF  1-  8 EMIT BL EMIT 8 EMIT  THEN ;

| : .ESC ( lim addr n char -- lim addr 0 )
   DROP  BEGIN  DUP WHILE 0 .BS REPEAT ;

| : .RETURN ( lim addr n char -- n addr -1 )
   DROP -ROT NIP -1 ;

| : .ACCEPTED ( lim addr n char -- lim addr n' )
   FOURTH THIRD = IF  DROP  EXIT  THEN
   BL MAX  DUP EMIT  THIRD THIRD + C!  1+ ;

| : ASTROKE ( lim addr n char -- lim addr n )
   DUP 127 > IF  DROP EXIT  THEN
   DUP <BS> = IF  .BS EXIT  THEN
   DUP 13 = IF  .RETURN EXIT  THEN
   DUP 27 = IF  .ESC EXIT  THEN
   .ACCEPTED ;

| : SIMPLE-ACCEPT ( addr n -- n )
   SWAP 0 BEGIN ( lim addr n )
      DUP 0< NOT WHILE
      AKEY ASTROKE
   REPEAT 2DROP ;

XDEFER (ACCEPT)   ' SIMPLE-ACCEPT ,

: TIB ( -- addr )   'TIB @ ;

{ --------------------------------------------------------------------
Input stream source

SOURCE-ID returns the contents of 'SOURCE-ID:
  -1  Source is EVALUATE string
   0  Source is keyboard
other Source is file (BLK=0 means text file; BLK<>0 means blocks)

SOURCE=FILE returns true if SOURCE-ID and BLK indicate that the
current input stream is from a text file.

SOURCE  returns the address and length of the current input source.
This is normally in memory pointed to by the 'TIB variable.

/SOURCE  returns the unexamined part of the input source.
-------------------------------------------------------------------- }

: SOURCE-ID ( -- n )   'SOURCE-ID @ ;

: SOURCE=FILE ( -- flag )
   SOURCE-ID 1+ 1 U>  BLK @ 0= AND ;

: (SOURCE) ( -- c-addr u )   #TIB 2@ ;

XDEFER SOURCE   ' (SOURCE) ,

: /SOURCE ( -- c-addr u )
   SOURCE >IN @ /STRING ;

{ ------------------------------------------------------------------------
Refill

We refill the current input stream differently depending on the input
source.  For keyboard input, we ask for another line of input (ie
ACCEPT). For blocks or files, the behaviour is defined elsewhere.

The decision is made from SOURCE-ID and BLK.

REFILL during EVALUATE always return false.  REFILL during keyboard
input always returns true.  Files and blocks depend on the input
stream.

source-id @ case
  0 of keyboard
 -1 of evaluate
  n of blk if block else file then

REFILL has the side effect of resetting the warning flags if they have
been temporarily over-ridden.
------------------------------------------------------------------------ }

XDEFER REFILL-LINE      0 ,
XDEFER REFILL-BLOCK     0 ,

: REFILL-KEYBOARD ( -- flag )
   >IN OFF 'TIB @ |TIB| ACCEPT #TIB !  SPACE  -1 ;

: REFILL-DISK ( -- flag )
   >IN OFF  BLK @ IF  REFILL-BLOCK  ELSE  REFILL-LINE  THEN ;

: REFILL-STREAM ( -- flag )
   SOURCE-ID 1+ IF  REFILL-DISK  ELSE  ( refill evaluate) 0 THEN ;

: REFILL ( -- f )   /WARNING  >AS OFF
   SOURCE-ID IF  REFILL-STREAM  ELSE  REFILL-KEYBOARD  THEN ;
