{ ====================================================================
Compiler interface

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Dictionary

We compile data into the dictionary at HERE via the "comma" words.

HERE returns the current dictionary pointer.

ALLOT allocates n bytes in the dictionary.

/ALLOT allocates u bytes and erases them.

PAD is a temporary buffer above HERE. Numeric conversion is done below
PAD, while the space above it is reserved as a user scratch buffer.

C, allocates a byte and writes a value to it.
W,  allocates a 16-bit word and writes a value to it.
L,  allocates a 32-bit longword and writes a value to it.
,  allocates a cell and writes a value to it.

Note that W, and L, always compile the same size (word or longword) on
both 32- and 64-bit host systems.
-------------------------------------------------------------------- }

: HERE ( -- a )   H  @ ;

: ALLOT ( n -- )   ?ENUF  H +! ;

: /ALLOT ( u -- )   HERE SWAP DUP ALLOT ERASE ;

: PAD  ( -- a )   HERE |NUM| + ;

: C, ( c -- )   HERE 1 ALLOT  C! ;
: W, ( h -- )   HERE 2 ALLOT  W! ;
: L, ( h -- )   HERE 4 ALLOT  L! ;
: ,  ( x -- )   HERE CELL ALLOT  ! ;

{ ------------------------------------------------------------------------
Inline compiler

SwiftForth is a subroutine threaded system, which means that the
primary mechanism for referencing a Forth word is to call it. We
extend this concept somewhat, with the ability to INLINE short code
routines. In the dictionary header, immediately preceeding the
executable code of any executable entity, there is an INLINE byte,
which tells how many bytes to copy if non-zero.

/OPT will reset the optimizer, when the optimizer is loaded.

[/OPT] resets it immediately, even while compiling a word.

,CALL  compiles a subroutine call instruction to the address.
,JMP   compiles a JUMP instruction to the address.

EXPAND lays down a "string" into the dictionary.

(COMPILE,) is the default compiler behaviour, run thru a vector so the
debug code can intercept it and lay down debugging code.

COMPILE,  compiles an XT by either CALL-ing it or INLINE-ing it.
------------------------------------------------------------------------ }

XDEFER /OPT  0 ,

: [/OPT]   /OPT ; IMMEDIATE

: (,CALL) ( addr -- )   $E8 C,  HERE 4 + - L, ;
: (,JMP)  ( addr -- )   $E9 C,  HERE 4 + - L, ;

XDEFER ,CALL ( addr -- )   ' (,CALL) ,
XDEFER ,JMP  ( addr -- )   ' (,JMP)  ,

: EXPAND ( addr len -- )
   HERE SWAP DUP H +! CMOVE ;

: (COMPILE,) ( xt -- )
   +ORIGIN DUP 1- C@ ?DUP IF ( addr len)  EXPAND  ELSE  ,CALL  THEN ;

XDEFER COMPILE,   ' (COMPILE,) ,
