{ ====================================================================
Forth core wordset

Copyright 2011  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Memory Reference Operators

Memory reference operators  @ ,  ! , and  +!  operate on cell-wide
values.

The following are provided for portability with external interfaces (API
calls, network buffers, etc.):

W@ and W! are 16-bit ("word") operators.
L@ and L! are 32-bit ("longword") operators.

W@S and L@S provide signed versions of word and longword operators.
For 32-bit systems, L@ and L@S are synonyms for @.

-------------------------------------------------------------------- }

ICODE @ ( addr -- x )
   0 [EBX] EBX MOV   RET   END-CODE

ICODE L@ ( addr -- u )
   0 [EBX] EBX MOV   RET   END-CODE

ICODE L@S ( addr -- n )
   0 [EBX] EBX MOV   RET   END-CODE

ICODE W@ ( addr -- u )
   0 [EBX] EBX MOVZXW   RET   END-CODE

ICODE W@S ( addr -- n )
   0 [EBX] EBX MOVSXW   RET   END-CODE

ICODE C@ ( addr -- char )
   EAX EAX XOR   0 [EBX] AL MOV   EAX EBX MOV
   RET   END-CODE

ICODE ! ( x addr -- )
   0 [EBP] EAX MOV   EAX 0 [EBX] MOV
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE L! ( x addr -- )
   0 [EBP] EAX MOV   EAX 0 [EBX] MOV
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE W! ( x addr -- )
   EBX ECX MOV   0 [EBP] EAX MOV   CELL [EBP] EBX MOV
   AX 0 [ECX] MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE C! ( c addr -- )
   0 [EBP] EAX MOV   AL 0 [EBX] MOV   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   RET   END-CODE

ICODE 2C@ ( addr -- nh nl )
   -CELL [EBP] EBP LEA   1 [EBX] EAX MOVZX
   0 [EBX] EBX MOVZX   EAX 0 [EBP] MOV
   RET   END-CODE

ICODE 2C! ( nh nl addr -- )
   EBX ECX MOV
   0 [EBP] AL MOV   CELL [EBP] AH MOV
   2 CELLS [EBP] EBX MOV
   AL 0 [ECX] MOV   AH 1 [ECX] MOV
   3 CELLS [EBP] EBP LEA   RET   END-CODE

ICODE 2! ( x1 x2 a-addr -- )
   0 [EBP] EAX MOV   CELL [EBP] ECX MOV
   EAX 0 [EBX] MOV   ECX CELL [EBX] MOV
   2 CELLS [EBP] EBX MOV   3 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE 2@ ( a-addr -- x1 x2 )
   -CELL [EBP] EBP LEA   CELL [EBX] EAX MOV
   EAX 0 [EBP] MOV   0 [EBX] EBX MOV
   RET   END-CODE

{ ------------------------------------------------------------------------
More Memory Operators

These operators implement some common memory access words.

+! adds to the existing value at an address. (Variants supplied for
longword, word, byte.)

ON and OFF set the cell at an address to TRUE or FALSE.

++ increments a value.

@+ reads and increments the address.  !+ writes and increments the
address.  ~!+ also writes and increments the address, but accepts
the parameters in reverse order.
------------------------------------------------------------------------ }

ICODE +! ( n addr -- )
   0 [EBP] EAX MOV   EAX 0 [EBX] ADD
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE L+! ( n addr -- )
   0 [EBP] EAX MOV   EAX 0 [EBX] ADD
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE W+! ( n addr -- )
   0 [EBP] EAX MOV   AX 0 [EBX] ADD
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE C+! ( n addr -- )
   0 [EBP] EAX MOV   AL 0 [EBX] ADD
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE ON ( addr -- )
   -1 # 0 [EBX] MOV   POP(EBX)
   RET   END-CODE

ICODE OFF ( addr -- )
   0 # 0 [EBX] MOV   POP(EBX)
   RET   END-CODE

ICODE ++ ( addr -- )
   1 # 0 [EBX] ADD   POP(EBX)
   RET   END-CODE

ICODE @+ ( addr1 -- addr2 n )
   PUSH(EBX)   0 [EBX] EBX MOV   CELL # 0 [EBP] ADD
   RET   END-CODE

ICODE !+ ( addr1 n -- addr2 )
   EBX EAX MOV   POP(EBX)
   EAX 0 [EBX] MOV   CELL # EBX ADD
   RET   END-CODE

ICODE ~!+ ( n addr1 -- addr2 )
   0 [EBP] EAX MOV   EAX 0 [EBX] MOV
   CELL [EBP] EBP LEA   CELL # EBX ADD
   RET   END-CODE

{ ------------------------------------------------------------------------
Return Stack Operators

We often store temporary values on the return stack. These words move
data to and from the return stack.

Note that the double number operators leave the top of the return
stack with the same value that was on top of the data stack. This
means that, for instance, 2>R is equivalent to "SWAP >R >R"

All of the operators that move data to and from the return stack are
designed to be used in matching pairs, as  ">R ... R>" "2>R ... 2R>"
and "N>R ... NR>" .

The N>R operator takes a parameter and saves that many items on the
return stack plus the number of items saved.  NR> depends on the
number of items being top on the return stack.
------------------------------------------------------------------------ }

ICODE R> ( -- x )   ( R: x -- )
   PUSH(EBX)   EBX POP   RET   END-CODE

ICODE >R ( x -- )   ( R: -- x )
   EBX PUSH   POP(EBX)   RET   END-CODE

ICODE R@ ( -- x )   ( R: x -- x )
   PUSH(EBX)   0 [ESP] EBX MOV   RET   END-CODE

ICODE DUP>R  ( x -- x )  ( R: -- x )
   EBX PUSH   RET   END-CODE

ICODE R>DROP ( -- )      ( R: x -- )
   EAX POP   RET   END-CODE

ICODE R>2DROP ( x -- )   ( R: x -- )
   EAX POP   POP(EBX)   RET   END-CODE

ICODE DROPR> ( n -- x )   ( R: x -- )
   EBX POP   RET   END-CODE

ICODE 2>R ( x1 x2 -- )  ( R: -- x1 x2 )
   0 [EBP] PUSH   EBX PUSH   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   RET   END-CODE

ICODE 2R> ( -- x1 x2 )   ( R: x1 x2 -- )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   EBX POP   0 [EBP] POP   RET   END-CODE

ICODE 2R@ ( -- x1 x2 )   ( R: x1 x2 -- x1 x2 )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   0 [ESP] EBX MOV   CELL [ESP] EAX MOV
   EAX 0 [EBP] MOV   RET   END-CODE

CODE N>R ( x1 x2 .. xn n -- )  ( R: -- x1 x2 .. xn n )
   EDX POP   EBX ECX MOV        \ ret addr, n
   ECX ECX OR  0<> IF
      BEGIN   0 [EBP] PUSH   CELL [EBP] EBP LEA   LOOP
   THEN
   EBX PUSH   EDX PUSH          \ push n, ret addr onto R
   POP(EBX)   RET   END-CODE

ICODE NR> ( -- x1 x2 .. xn n )  ( R: x1 x2 .. xn n -- )
   -CELL [EBP] EBP LEA   EBX 0 [EBP] MOV
   EBX POP   EBX ECX MOV
   ECX ECX OR  0<> IF
      BEGIN   -CELL [EBP] EBP LEA   0 [EBP] POP   LOOP
   THEN   RET   END-CODE

{ ------------------------------------------------------------------------
Return stack workspace allocation

A provision to allocate temporary data space on the return stack makes
many OS-related programming tasks easier. The space is allocated in
bytes by the R-ALLOC function which returns the address of the space
allocated. The space is automatically de-allocated when the calling
routine executes a return.

The trick is that R-ALLOC intercepts its return address before
completing, and by calling R-GO-ON to do its return, pushes a new
return address for the eventual return from its caller to come back to
and thereby restore the return stack.

Do not use this after a >R operation or inside a DO..LOOP !
------------------------------------------------------------------------ }

LABEL R-GO-ON
   EAX JMP   END-CODE

CODE R-ALLOC ( n -- addr )
   EAX POP                          \ get my return address into eax
   ESP ECX MOV                      \ save esp in ecx
   BEGIN                            \ make the allocation in 4k chunks
      4096 # EBX CMP                \ since this is the size of the OS
      > WHILE                       \ allocation units
      4096 # ESP SUB
      4096 # EBX SUB
      0 [ESP] EDX MOV
   REPEAT
   EBX ESP SUB
   $FFFFFFF0 # ESP AND              \ align ESP mod 16 (required by macOS)
   ESP EBX MOV                      \ get the new ESP onto TOS
   ECX PUSH                         \ push saved ESP onto return stack
   R-GO-ON CALL                     \ make return to R-ALLOC 's caller
                                    \ when R-ALLOC 's caller returns, it
HERE ( *)
   ECX POP                          \ comes here, reads the old esp
   ECX ESP MOV                      \ and write back to discard space
   RET   END-CODE

( *) CONSTANT 'R-ALLOC

LABEL R-GO-ON2
   EDX PUSH                         \ put the address on rp initially
   EAX JMP                          \ jump to address eax
   END-CODE

CODE R-BUF ( -- )                   \ create buffer and push address on rp
   EAX POP                          \ get my return address into eax
   ESP ECX MOV                      \ save esp in ecx
   MAX_PATH # ESP SUB               \ allocate 260 bytes below esp
   $FFFFFFF0 # ESP AND              \ align ESP mod 16 (required by macOS)
   ESP EDX MOV                      \ hold buffer address
   ECX PUSH                         \ push saved ESP onto return stack
   R-GO-ON2 CALL                    \ make return to R-ALLOC 's caller
                                    \ when R-ALLOC 's caller returns, it
HERE ( *)
   ECX POP                          \ comes here, reads the old esp
   ECX ESP MOV                      \ and write back to discard space
   RET   END-CODE

( *) CONSTANT 'R-BUF

ICODE +RP ( n -- )
   EBX ESP ADD                      \ adjust return stack pointer
   POP(EBX)
   RET   END-CODE

{ ------------------------------------------------------------------------
Data Stack Operations

The data stack operators manipulate the top items on the data stack.
------------------------------------------------------------------------ }

ICODE NIP ( x1 x2 -- x2 )
   CELL [EBP] EBP LEA   RET   END-CODE

ICODE DROP ( x -- )
   POP(EBX)   RET   END-CODE

ICODE DUP ( x -- x x )
   PUSH(EBX)   RET   END-CODE

ICODE SWAP ( x1 x2 -- x2 x1 )
   0 [EBP] EAX MOV   EBX 0 [EBP] MOV
   EAX EBX MOV   RET   END-CODE

ICODE OVER  ( x1 x2 -- x1 x2 x1 )
   PUSH(EBX)   CELL [EBP] EBX MOV
   RET   END-CODE

ICODE THIRD ( x1 x2 x3 -- x1 x2 x3 x1 )
   PUSH(EBX)   2 CELLS [EBP] EBX MOV
   RET   END-CODE

ICODE FOURTH ( x1 x2 x3 x4 -- x1 x2 x3 x4 x1 )
   PUSH(EBX)   3 CELLS [EBP] EBX MOV
   RET   END-CODE

ICODE ROT ( x1 x2 x3 -- x2 x3 x1 )
   EBX ECX MOV   CELL [EBP] EBX MOV
   0 [EBP] EAX MOV   EAX CELL [EBP] MOV
   ECX 0 [EBP] MOV   RET   END-CODE

ICODE -ROT ( x1 x2 x3 -- x3 x1 x2 )
   EBX ECX MOV   CELL [EBP] EAX MOV
   0 [EBP] EBX MOV   EAX 0 [EBP] MOV
   ECX CELL [EBP] MOV   RET   END-CODE

ICODE TUCK ( x1 x2 -- x2 x1 x2 )
   0 [EBP] EAX MOV   -CELL [EBP] EBP LEA
   EBX CELL [EBP] MOV   EAX 0 [EBP] MOV
   RET   END-CODE

ICODE ?DUP ( x -- 0 | x x )
   EBX EBX OR   0<> IF   PUSH(EBX)   THEN
   RET   END-CODE

ICODE NOOP ( -- )
   NOP   RET   END-CODE

{ ------------------------------------------------------------------------
Boolean Logic Operators

These boolean operators operate on the bits of the top number(s) on
the stack.  AND OR and XOR produce the bit-wise logical result of the
top two items, and INVERT changes all ones to zeros and visa-versa on
the top stack element.
------------------------------------------------------------------------ }

ICODE AND ( x1 x2 -- x3 )
   0 [EBP] EBX AND   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE OR  ( x1 x2 -- x3 )
   0 [EBP] EBX OR   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE XOR ( x1 x2 -- x3 )
   0 [EBP] EBX XOR   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE INVERT ( x1 -- x2 )
   EBX NOT   RET   END-CODE

{ ------------------------------------------------------------------------
Simple Math Operators

+ adds the top two numbers on the stack, - subtracts the top
item from the second item.  SWAP- subtracts the second item
from the top item.  NEGATE returns the two's compliment of the
top item on the stack.  ABS returns the absolute value.

>< swaps bits 0-7 with bits 8-15 of the top item, while >H< swaps
bits 0-15 with bits 16-31.
------------------------------------------------------------------------ }

ICODE + ( n1 n2 -- n3)
   0 [EBP] EBX ADD   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE SWAP- ( n1 n2 -- n3)
   0 [EBP] EBX SUB   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE - ( n1 n2 -- n3)
   EBX NEG   0 [EBP] EBX ADD
   CELL [EBP] EBP LEA   RET   END-CODE

ICODE NEGATE ( n -- -n )
   EBX NEG   RET   END-CODE

ICODE ABS ( n1 -- +n2 )
   EBX EBX OR   0< IF   EBX NEG   THEN
   RET   END-CODE

ICODE >< ( x1 -- x2 )   \
   BL BH XCHG   RET   END-CODE

ICODE >H< ( x1 -- x2 )
   16 # CL MOV   EBX CL ROL
   RET   END-CODE

{ ------------------------------------------------------------------------
Binary Comparison Operators

The binary comparison operators test the relationship between the
top two items on the data stack.  Their code is identical except
for the actual comparison operation used.  This lends itself to
an assembler macro to compile the code required to generate the
truth flag from the comparison.

The tests < > <= >= are signed tests, U< and U> are unsigned.

MIN leaves the smallest signed value on the stack, while MAX
leaves the largest signed value.

WITHIN tests for membership in a range of numbers.
------------------------------------------------------------------------ }

ASM

: BINARY-TEST ( flag -- )   [ ASSEMBLER ]
   EBX 0 [EBP] CMP   0 # EBX MOV
   ( cc) IF   EBX DEC   THEN   CELL [EBP] EBP LEA
   RET   END-CODE ;

TARGET

CODE <  ( u u -- flag )     < BINARY-TEST
CODE >  ( u u -- flag )     > BINARY-TEST
CODE =  ( n n -- flag )    0= BINARY-TEST
CODE >= ( n n -- flag )    >= BINARY-TEST
CODE <= ( n n -- flag )    <= BINARY-TEST
CODE <> ( n n -- flag )   0<> BINARY-TEST

CODE U< ( u1 u2 -- flag )
   EBX 0 [EBP] CMP   EBX EBX SBB   CELL [EBP] EBP LEA
   RET   END-CODE

CODE U> ( u1 u2 -- flag )
   0 [EBP] EBX CMP   EBX EBX SBB   CELL [EBP] EBP LEA
   RET   END-CODE

CODE MAX ( n1 n2 -- n3 )
   0 [EBP] EAX MOV   CELL [EBP] EBP LEA
   EAX EBX CMP   < IF   EAX EBX MOV   THEN
   RET   END-CODE

CODE MIN ( n1 n2 -- n3 )
   0 [EBP] EAX MOV   CELL [EBP] EBP LEA
   EAX EBX CMP   > IF   EAX EBX MOV   THEN
   RET   END-CODE

CODE UMAX ( u1 u2 -- u3 )
   0 [EBP] EAX MOV   CELL [EBP] EBP LEA
   EAX EBX CMP   U< IF   EAX EBX MOV   THEN
   RET   END-CODE

CODE UMIN ( u1 u2 -- u3 )
   0 [EBP] EAX MOV   CELL [EBP] EBP LEA
   EAX EBX CMP   U> IF   EAX EBX MOV   THEN
   RET   END-CODE

: WITHIN ( n lo hi -- flag )
   OVER - >R - R> U< ;

{ ------------------------------------------------------------------------
Unary Tests

The unary tests operate only on the top of stack.  They are
optimized for their individual requirements.

0= and NOT are identical.
------------------------------------------------------------------------ }

ICODE 0<> ( n -- flag )
   EBX NEG   EBX EBX SBB   RET   END-CODE

ICODE 0= ( n -- flag )
   EBX NEG   EBX EBX SBB   EBX NOT   RET   END-CODE

ICODE NOT ( n -- flag )
   EBX NEG   EBX EBX SBB   EBX NOT   RET   END-CODE

ICODE 0< ( n -- flag )
   EBX EAX MOV   CDQ   EDX EBX MOV   RET   END-CODE

ICODE 0> ( n -- flag )
   0 # EBX CMP   0 # EBX MOV   > IF  EBX DEC  THEN
   RET   END-CODE

{ ------------------------------------------------------------------------
Math Optimizations

The words  1+ ,  2+ , etc. provide more than convenience: they are
both shorter and faster than  1 + , etc., and thus should always be
preferred.

Also note that  2* ,  2/  ,  4*  ,  4/  use arithmetic shifts, and are
thus signed operations.  2/  and  4/  differ from 2 /  and  4 /  in
terms of rounding behavior.  2/  and  4/ floor whereas  2 /  and  4 /
round towards zero.

LSHIFT  logically shift x left n places.
RSHIFT  logically shift x right n places.
------------------------------------------------------------------------ }

ICODE 1+ ( n -- n+1 )
   EBX INC
   RET   END-CODE

ICODE 1- ( n -- n-1 )
   EBX DEC
   RET   END-CODE

ICODE 2+ ( n -- n+2 )
   2 # EBX ADD
   RET   END-CODE

ICODE 2- ( n -- n-2 )
   2 # EBX SUB
   RET   END-CODE

ICODE 2/ ( n -- n/2 )
   EBX SAR
   RET   END-CODE

ICODE 2* ( n -- n*2 )
   EBX SHL
   RET   END-CODE

ICODE U2/ ( u -- u/2 )
   EBX SHR
   RET   END-CODE

ICODE 4+ ( n -- n+2 )
   4 # EBX ADD
   RET   END-CODE

ICODE 4- ( n -- n-2 )
   4 # EBX SUB
   RET   END-CODE

ICODE RSHIFTA ( x1 n -- x2 )
   EBX ECX MOV
   POP(EBX)
   EBX CL SAR
   RET   END-CODE

ICODE RSHIFT ( x1 n -- x2 )
   EBX ECX MOV
   POP(EBX)
   EBX CL SHR
   RET   END-CODE

ICODE LSHIFT ( x1 n -- x2 )
   EBX ECX MOV
   POP(EBX)
   EBX CL SHL
   RET   END-CODE

{ ------------------------------------------------------------------------
Cell Operations

These words help the programmer to factor the size of the machine
word out of the code.  Using these operators instead of the
traditional "4+" or "2+" makes programs more portable.

CELL+  increments an address past a cell.
CELL-  decrements an address one cell.
CELLS  returns the number of bytes in the given number of cells.
CELL/  returns the number of cells in the given number of bytes.
CELL   returns the number of bytes in one cell.
-CELL  returns the negated number of bytes in one cell.

CHAR+ advances to the next character in memory
CHARS returns the number of bytes in a given number of characters.
ALIGNED aligns an address. Noop.
ALIGN aligns the dictionary. Noop.
------------------------------------------------------------------------ }

ICODE CELL  ( -- n )
   PUSH(EBX)   CELL # EBX MOV   RET   END-CODE

ICODE -CELL ( -- n )
   PUSH(EBX)   -CELL # EBX MOV   RET   END-CODE

ICODE CELL+ ( n -- n )
   CELL [EBX] EBX LEA   RET   END-CODE

ICODE CELL- ( n -- n )
   -CELL [EBX] EBX LEA   RET   END-CODE

ICODE CELLS ( n -- n )
   EBX 2 # SHL   RET   END-CODE

ICODE CELL/ ( n -- n )
   EBX 2 # SHR   RET   END-CODE

ICODE CHAR+ ( n -- n )
   EBX INC   RET   END-CODE

ICODE CHARS ( n -- n )
   RET   END-CODE

ICODE ALIGNED ( addr -- addr )
   RET   END-CODE

ICODE ALIGN ( -- )
   RET   END-CODE

{ ----------------------------------------------------------------------
Loop functions

I and I' return inner loop index and limit, respectively.
J and J' return next outer loop index and limit.
---------------------------------------------------------------------- }

ICODE UNLOOP ( -- )
   8 # ESP ADD   RET   END-CODE

ICODE I ( -- n )
   PUSH(EBX)   0 [ESP] EBX MOV   CELL [ESP] EBX ADD
   RET   END-CODE

ICODE I' ( -- n )
   PUSH(EBX)   CELL [ESP] EBX MOV   $80000000 # EBX SUB
   RET   END-CODE

ICODE J ( -- n )
   PUSH(EBX)   2 CELLS [ESP] EBX MOV   3 CELLS [ESP] EBX ADD
   RET   END-CODE

ICODE J' ( -- n )
   PUSH(EBX)   2 CELLS [ESP] EBX MOV   $80000000 # EBX SUB
   RET   END-CODE

{ --------------------------------------------------------------------
32 to 16 bit conversions

LOWORD returns the low 16 bits of n
HIWORD returns the high 16 bits of n shifted right 16 bits

HILO returns the high and low halves of a word, low in tos.
LOHI returns the low and high halves of a word, high in tos.
-------------------------------------------------------------------- }

ICODE LOWORD ( n -- n )
   $FFFF # EBX AND   RET   END-CODE

ICODE HIWORD ( n -- n )
   EBX 16 # SHR   RET   END-CODE

CODE HILO ( n -- hi lo )
   -CELL [EBP] EBP LEA   EBX EAX MOV
   EAX 16 # SHR   EAX 0 [EBP] MOV
   $FFFF # EBX AND   RET   END-CODE

CODE LOHI ( n -- lo hi )
   -CELL [EBP] EBP LEA   EBX EAX MOV
   $FFFF # EAX AND   EAX 0 [EBP] MOV
   EBX 16 # SHR   RET   END-CODE

CODE LOHI-PACK ( lo hi -- x )
   EBX 16 # SHL   0 [EBP] EAX MOV
   $FFFF # EAX AND   EAX EBX OR
   CELL [EBP] EBP LEA   RET   END-CODE

{ ------------------------------------------------------------------------
Stack stuff

DEPTH returns the number of items on the data stack.
PICK copies the Nth item to the top of the data stack.
ROLL does an N-sized rotate of the top stack items.
-ROLL rotates nos to ith position on stack.
------------------------------------------------------------------------ }

CODE DEPTH ( -- n )
   PUSH(EBX)   S0 [ESI] EBX MOV   EBP EBX SUB
   EBX SAR   EBX SAR   RET   END-CODE

ICODE PICK ( ... n -- xn )
   0 [EBP] [EBX*4] EBX MOV   RET   END-CODE

CODE ROLL ( xu xu-1 ... x0 u -- xu-1 ... x0 xu)
   0 [EBP] [EBX*4] EAX MOV   EBX EBX TEST   0= NOT IF   \ Get xu, skip if u = 0
      BEGIN   -CELL [EBP] [EBX*4] ECX MOV               \ Move stack cells
      ECX 0 [EBP] [EBX*4] MOV
   EBX DEC   0= UNTIL   THEN
   CELL [EBP] EBP LEA   EAX EBX MOV                     \ Return xu as tos
   RET   END-CODE

CODE -ROLL ( i*x n i -- n i*x )
   EBX ECX MOV   0 [EBP] EDX LEA                        \ Cell count, address of xu
   ECX ECX OR  0<> IF   0 [EDX] EBX MOV                 \ Skip if i = 0
      BEGIN   CELL [EDX] EAX MOV   EAX 0 [EDX] MOV
      CELL # EDX ADD   LOOP   EBX 0 [EDX] MOV
   THEN   POP(EBX)   RET   END-CODE
