{ ====================================================================
User variable and system data

Copyright 2001 FORTH, Inc.

User variables and global system data are defined here.
==================================================================== }

{ --------------------------------------------------------------------
USER Variables

A few notes:

DPL holds the digits to the right of the last punctiation character in
the string.  Any negative value indicates that there was no punctuation.

>IN is the standard forth input position pointer.  IN> is the actual
offset in the input stream where the parsed string began, and IN>
CELL+ is the length of the string parsed.
-------------------------------------------------------------------- }

0        8  +USER STATUS                \  0 Multi-tasking hooks
      CELL  +USER SAVEIP                \  8 "
      CELL  +USER TCB                   \  C "
      CELL  +USER S0                    \ 10 initial stack pointer
      CELL  +USER R0                    \ 14 initial return stack pointer
      CELL  +USER 'N                    \    numeric stack pointer
      CELL  +USER 'N0                   \ initial numeric stack pointer
      CELL  +USER CATCHER               \ exception stack frame pointer
      CELL  +USER HLD                   \ char pointer for numeric output
      CELL  +USER DPL                   \ decimal on numeric input
      CELL  +USER NH                    \ high half of a double number on input

      CELL  +USER HLIM                  \ dictionary limit
      CELL  +USER H                     \ dictionary pointer
      CELL  +USER FENCE                 \ limit for PRUNE
      CELL  +USER 'EMPTY                \ limit for EMPTY

      CELL  +USER #TIB                  \ number of chars in TIB
      CELL  +USER 'TIB                  \ pointer to tib (must follow #TIB - they're a pair)
      CELL  +                           \ initial TIB
      CELL  +USER BASE                  \ numeric conversion base

      CELL  +USER #ORDER                \ number of WID entries in context stack
      CELL  +USER CONTEXT               \ the top of the context stack
=VOCS CELLS +                           \ the stack itself
      CELL  +USER CURRENT               \ the wid which is compiled into

      CELL  +USER 'PERSONALITY          \ address of personality array

      CELL  +USER 'LF                   \ frame pointer for local variables
      CELL  +USER 'OB                   \ frame pointer for objects
      CELL  +USER 'THIS                 \ handle of current object class
      CELL  +USER 'SELF                 \ address of current object instance
      CELL  +USER 'METHOD               \ access method for TO in SWOOP

      CELL  +USER 'SOURCE-ID            \ input source, 0(keyboard) +(file) -(blk)
    2 CELLS +USER >MAPPED               \ addr,len of remaining mapped file being included (len first for 2@)
      CELL  +USER BLK                   \ block number for LOAD
      CELL  +USER >IN                   \ offset into TIB

      CELL  +USER 'CFA                  \ current definition's code field address
      CELL  +USER STATE                 \ compiler state
      CELL  +USER >AS                   \ pointer for remembering an input word
    2 CELLS +USER IN>                   \ offset and length of text parsed by WORD
      CELL  +USER 'LOCATE               \ last locate field found by SEARCH-THREAD

( #bytes of user variables) VALUE #USER

{ ------------------------------------------------------------------------
System constants and variables

We define constants so we don't have to use "magic numbers" in our
source code.

We also define system-wide global variables to control the behaviour
of Forth here.

BAL        keeps up with how many structures are being compiled.
BAL CELL+  is a list of LEAVE-s to resolve.

#STRANDS   the default number of strands in a WORDLIST
VLINK      a linked list of vocabularies
WIDS       a linked list of WORDLISTs
UP0        the initial user area pointer, for Warm start
DP0        the initial dictionary pointer, for Warm start

Note: CAPS is headless as it should not be accessed directly.
Set/clear with CASE-SENSITIVE and CASE-INSENSITIVE.
------------------------------------------------------------------------ }

32 CONSTANT BL                          \ Character code for space
 0 CONSTANT FALSE                       \ false
-1 CONSTANT TRUE                        \ true

|FPSTACK| CONSTANT |FPSTACK|            \ Make config values available in dictionary
|USER| CONSTANT |USER|
|TIB| CONSTANT |TIB|

=VOCS CONSTANT #VOCS                    \ max number of wids in CONTEXT

VARIABLE #STRANDS   31 #STRANDS !       \ default strands while running

VARIABLE VLINK                          \ the newest vocabulary
VARIABLE WIDS                           \ the newest wordlist

CREATE BAL   0 , 0 ,                    \ structure balancing, used in IF, DO, :

CREATE LAST                             \ track last thing defined
   0 ,                                  \ address of last name
   0 ,                                  \ address of last INLINE byte
   0 ,                                  \ last code address
   0 ,                                  \ address of last thread linked into
   0 ,                                  \ a dummy spot

VARIABLE UP0                            \ initial operator user pointer
VARIABLE DP0                            \ initial dictionary pointer

VARIABLE LINE                           \ line number for INCLUDE
CREATE F#                               \ file numbering for INCLUDE
   1 ,                                  \ current file number
   1 ,                                  \ total files used
   0 ,                                  \ temporary space to slip around restore-input

CREATE POCKET   256 ALLOT               \ temporary string storage

0 VALUE HCON                            \ a global copy of hwnd

VARIABLE OBSCURED                       \ total of hidden words

| CREATE CAPS  -1 ,                     \ 0=case sensitive; -1=insensitive (as if all in CAPS)
