{ ====================================================================
Defining words

Copyright 2001  FORTH, Inc.
Rick VanNorman, Leon Wagner

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Later, we will need to clear the local object and local variable
compilers so they are usable in class definitions.
-------------------------------------------------------------------- }

XDEFER /LOCALS   0 ,

{ --------------------------------------------------------------------
Cache line skip

We skip to the next mod-64 address boundary if we're making a new
colon or CODE definition if it might start in the same mod-64 region
as the PFA of a CREATE word.
-------------------------------------------------------------------- }

VARIABLE >CREATE

: SKIP-CL ( -- )
   LAST @ >CREATE @ = IF  HERE 1- DUP 15 + XOR -64 AND
   0= IF  HERE 15 + NEGATE 63 AND ALLOT  THEN THEN ;

{ --------------------------------------------------------------------
(;) ends a colon definition.  Checks that BAL is zero, sets STATE to
INTERPRET -ing (0).  Skips balance check if ?BAL is false.  Clears
HERE so ABORT" won't display garbage.

+BALANCE turns on balance checking in  (;).

-BALANCE turns it off.  The default is on; turn if off only if you
need to.

(:) begins a colon definition. Creates a smudged header, clears BAL,
and sets state to compiling (-1).

; compiles a return instruction, executes the '(;) vector, and -
SMUDGE -s the most recent header.

: calls (:) , creates a header, and smudges it.

:NONAME compiles an unnamed colon definition, and returns its xt.

CREATE makes a definition that will return the address of its body
when executed.

VARIABLE   creates a named storage location 1 cell long.
2VARIABLE  creates a named storage location 2 cells long.
CONSTANT   creates a named constant.
2CONSTANT  creates a named double constant.
-------------------------------------------------------------------- }

: (:) ( -- )   0 LAST !  /BAL /LOCALS ] ;

: : ( -- )   SKIP-CL  (:)  HEADER +SMUDGE ;

| CREATE ?BAL  -1 ,
: +BALANCE ( -- )   -1 ?BAL ! ;
: -BALANCE ( -- )   0 ?BAL ! ;

: (;) ( -- )   0 HERE C!
   BAL 2@ OR  ?BAL @ AND ABORT" Unbalanced control structure"
   /BAL /LOCALS  POSTPONE [  ;

: ; ( -- )   POSTPONE EXIT  (;)  -SMUDGE ;  IMMEDIATE

: :NONAME ( -- xt )
   $FF C, 0 C,  HERE DUP LAST CELL+ CELL+ !  CODE>  (:) ;

: CREATE ( -- )
   BL WORD COUNT GET-CURRENT (WID-CREATE)
   LAST @ >CREATE ! ;

: VARIABLE ( -- )   CREATE  0 , ;
: 2VARIABLE (  -- )   VARIABLE  0 , ;
: BUFFER: ( n -- )   CREATE ALLOT ;

: USER ( n -- )   HEADER  POSTPONE (USER)  , ;

: CONSTANT ( x -- )   HEADER  POSTPONE (CONSTANT)  , ;
: 2CONSTANT ( d -- )   HEADER  POSTPONE (2CONSTANT)  , , ;

: ENUM ( n -- n+1 )   DUP CONSTANT 1+ ;
: ENUM4 ( n -- n+4 )   DUP CONSTANT 4+ ;

{ --------------------------------------------------------------------
Create wordlists

A wordlist looks like:
   |widlink|strands|link|link|link|...|link|

STRANDS    creates a new wordlist with N strands.
WORDLIST   creates a new wordlist with #STRANDS @ strands.
#STRANDS @ is the system default number of strands for a wordlist.
-------------------------------------------------------------------- }

: STRANDS ( n -- wid )
   HERE WIDS @REL ,REL  WIDS !REL
   HERE -ORIGIN SWAP
   DUP ( #STRANDS) ,
   ( #STRANDS) 0 DO 0 , LOOP ;

: WORDLIST ( -- wid )
   #STRANDS @ STRANDS ;

{ --------------------------------------------------------------------
PARENT dereferences the given xt returning the xt that defined it.

?PARENT prints the error message from ADDR LEN if the parent of
   xt does not match xt2
-------------------------------------------------------------------- }

: PARENT ( xt -- xt | 0 )
   >CODE COUNT $E8 <> IF  DROP 0        \ Return 0 if no call to parent
   ELSE  @+ + -ORIGIN  THEN ;           \ Otherwise, return xt

: ?PARENT ( xt xt2 addr len -- )
   2>R  SWAP PARENT <>  2R> ?ERR ;

| : ?VALUE ( flag -- )   0= ABORT" Not a VALUE or 2VALUE" ;

| : ?DEFER ( xt -- )   ['] (DEFER) S" Not a DEFER" ?PARENT ;

{ --------------------------------------------------------------------
Values and deferred words

VALUE creates a named changeable constant.

DEFER creates a named execution vector.
IS sets an execution vector.
TO sets a value.
&OF returns the address of a VALUE or DEFER word's PFA.
-------------------------------------------------------------------- }

| : (TO) ( xt -- )
   STATE @ IF  POSTPONE LITERAL  POSTPONE >BODY!  EXIT  THEN   >BODY! ;

| : (2TO) ( xt -- )
   STATE @ IF  POSTPONE LITERAL  POSTPONE >BODY2!  EXIT  THEN   >BODY2! ;

| : (+TO) ( xt -- )
   STATE @ IF  POSTPONE LITERAL POSTPONE >BODY+!  EXIT  THEN   >BODY+! ;

: TO-VALUE ( -- flag )   >IN @ >R  '
   DUP PARENT ['] (VALUE) = IF  (TO)  R> DROP  -1  EXIT  THEN
   DUP PARENT ['] (2VALUE) = IF  (2TO)  R> DROP  -1  EXIT  THEN
   R> >IN !  DROP 0 ;

: +TO-VALUE ( -- flag )   >IN @ >R  '
   DUP PARENT ['] (VALUE) = IF  (+TO)  R> DROP  -1
   ELSE  DROP  R> >IN !  0   THEN ;

: VALUE ( x -- )   HEADER  POSTPONE (VALUE)  , ;
: 2VALUE ( x1 x2 -- )   HEADER  POSTPONE (2VALUE)  , , ;

: TO ( n -- )   TO-VALUE ?VALUE ; IMMEDIATE
: +TO ( n -- )  +TO-VALUE ?VALUE ; IMMEDIATE

| : (ADDR-OF) ( xt -- addr )
   STATE @ IF  POSTPONE LITERAL POSTPONE >BODY  EXIT  THEN   >BODY ;

: &OF-VALUE ( -- addr -1 | 0 )   >IN @ >R  '
   DUP PARENT ['] (VALUE) = IF  (ADDR-OF)  R> DROP  -1
   ELSE  DROP  R> >IN !  0  THEN ;

: &OF ( -- addr )   &OF-VALUE ?VALUE ; IMMEDIATE

| : CRASH ( -- )
   S" Uninitialized DEFER" ERR ;

: DEFER ( -- )   HEADER  POSTPONE (DEFER)  ['] CRASH , ;
: IS ( xt -- )   '  DUP ?DEFER  (TO) ; IMMEDIATE
: DEFER! ( xt2 xt1 -- )   DUP ?DEFER  >BODY! ;
: DEFER@ ( xt1 -- xt2 )   DUP ?DEFER  >BODY@ ;

: ACTION-OF ( "<spaces>name" -- xt )
   ' DUP ?DEFER  STATE @ IF  POSTPONE LITERAL  POSTPONE >BODY@
   ELSE  >BODY@  THEN ;  IMMEDIATE

{ --------------------------------------------------------------------
Chains

CHAINS points to the base of the list of linked lists.  IMPORTS is the
one chain predefined in the kernel.

CHAIN creates a new list linked to CHAINS.
-------------------------------------------------------------------- }

CREATE CHAINS   IMPORTS CELL+ ,REL(T)

: CHAIN ( -- )   CREATE  0 ,  CHAINS >LINK ;

{ --------------------------------------------------------------------
ANS provides no way for a created word to know its own xt -- which is
needed for portability in our object package. CREATE-XT provides a
means to CREATE an entity in the dictionary that knows its own xt.
-------------------------------------------------------------------- }

{ --------------------------------------------------------------------
This is the ans version. We wrote it; we can do it directly.

: CREATE-XT ( -- xt )   >IN @  CREATE >IN !
   BL WORD COUNT GET-CURRENT SEARCH-WORDLIST 0= THROW ;
-------------------------------------------------------------------- }

: CREATE-XT ( -- xt )   CREATE  LAST 2 CELLS + @ CODE> ;
