{ ====================================================================
Relative dictionary pointers

Copyright 2001  FORTH, Inc.
Rick VanNorman

SwiftForth dictionary structure:

  +-------------------------------------------------------------------------------
  | locate field
  |               +---------------------------------------------------------------
  |               | link field, contents relative or zero
  |               |               +-----------------------------------------------
  |               |               | sentinel, always $ff
  |               |               |   +-------------------------------------------
  |               |               |   | flags, 80=smudge, 40=immediate
  |               |               |   |   +---------------------------------------
  |               |               |   |   | length of name, range 0-254
  |               |               |   |   |   +-----------------------------------
  |               |               |   |   |   | name, stored as ascii
  |               |               |   |   |   |         +-------------------------
  |               |               |   |   |   |         | inline byte, 0=call only
  |               |               |   |   |   |         |   +---------------------
  |               |               |   |   |   |         |   |
  |   .   .   .   |   .   .   .   | x | f | # | n a m e | i | cfa
  ---------------------------------------------------------------

==================================================================== }

TARGET

{ --------------------------------------------------------------------
All links in SwiftForth are position relative. Zero is not relocatable.
These words support relative addressing.

@REL reads a relative value,
!REL writes one.
@+REL reads a relative value from a table.
,REL puts a relative value in the dictionary.
-------------------------------------------------------------------- }

ICODE @REL ( addr -- n )
   0 [EBX] ECX MOV   ECXNZ IF   EBX ECX ADD   THEN
   ECX EBX MOV   RET END-CODE

ICODE !REL ( n addr -- )
   0 [EBP] ECX MOV   ECXNZ IF   EBX ECX SUB  THEN
   ECX 0 [EBX] MOV   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   RET END-CODE

: @+REL ( addr1 -- addr2 n )
   DUP CELL+ SWAP @REL ;

: ,REL ( n -- )   H @  CELL H +!  !REL ;

{ --------------------------------------------------------------------
Linked lists

LINKS traverses the linked list until it finds the last entry in the
list (the one with a 0 link).

>LINK adds the top of the dictionary to the head of the linked list.

<LINK  adds the top of the dictionary to the end of the linked list.

@LINK is the same as @REL.
-------------------------------------------------------------------- }

: LINKS ( addr1 -- addr2 )
   BEGIN  DUP @REL ?DUP WHILE  NIP  REPEAT ;

: >LINK ( addr -- )
   H @   OVER @REL ,REL  SWAP !REL ;

: <LINK ( addr -- )   LINKS  >LINK ;

: @LINK ( addr1 -- addr2 )   @REL ;

{ ------------------------------------------------------------------------
XT address translation

The Forth Standard provides only one retrievable item of information
about a definition via ' (tick) -- the execution token (xt).

To be practical, we provide a means of translating from the various
header fields to and from the xt value.

Note that WID is the same as xt in Standard Forth, ie not a directly
addressable value.  WID> is provided to move from a WID to an address.
------------------------------------------------------------------------ }

ICODE >BODY ( xt -- addr )
   5 [EDI] [EBX] EBX LEA   RET   END-CODE

: >BODY! ( x xt -- )   >BODY ! ;
: >BODY2! ( x1 x2 xt -- )   >BODY 2! ;
: >BODY+! ( x xt -- )   >BODY +! ;
: >BODY@ ( xt -- x )   >BODY @ ;

ICODE BODY> ( addr -- xt )
   5 # EBX SUB   EDI EBX SUB   RET   END-CODE

ICODE >CODE ( xt -- addr )
   EDI EBX ADD   RET   END-CODE

ICODE CODE> ( addr -- xt )
   EDI EBX SUB   RET   END-CODE

ICODE WID> ( wid -- addr )
   EDI EBX ADD   RET   END-CODE

ICODE >WID ( addr -- wid )
   EDI EBX SUB   RET   END-CODE

: N>CODE ( nfa -- code )   COUNT + 1+ ;

: C>NAME ( code -- nfa )
   2 - DUP C@  $FF = IF  DROP 0  EXIT  THEN
   BEGIN 1- DUP C@ $FF = UNTIL 2+ ;

: N>LINK ( nfa -- lfa )   DUP IF  6 -  THEN ;
: L>NAME ( lfa -- nfa )   6 + ;

: L>VIEW ( lfa -- vfa )   DUP IF  4-  THEN ;
: V>LINK ( vfa -- lfa )   4+ ;

: N>VIEW ( nfa -- vfa )   N>LINK L>VIEW ;
: V>NAME ( vfa -- nfa )   V>LINK L>NAME ;

: >NAME ( xt -- nfa )   >CODE C>NAME ;
: NAME> ( nfa -- xt )   COUNT + 1+ -ORIGIN ;

: XT>LINK ( xt -- lfa )   >NAME N>LINK ;
: >VIEW ( xt -- vfa )   XT>LINK DUP IF 4 - THEN ;
: VIEW> ( vfa -- xt )   V>NAME NAME> ;
: LINK> ( link -- xt )   L>NAME NAME> ;

: >INLINE ( xt -- addr )   >CODE 1- ;

: N>BODY ( nfa -- body )   NAME> >BODY ;

: NFA-FLAGS ( xt -- n )   >NAME 1- C@ ;

{ --------------------------------------------------------------------
Dictionary limits

MEMTOP is the current end of available dictionary space.
UNUSED is the amount (in bytes) of available dictionary space.
?ENUF aborts if there isn't enough dictionary for n bytes.
-------------------------------------------------------------------- }

: MEMTOP ( -- addr )   HLIM @ ;

: UNUSED ( -- n )   MEMTOP H @ - ;

: ?ENUF ( n -- n )
   MEMTOP -EXIT  DUP 384 + UNUSED > -80 ?THROW ;
