{ ====================================================================
Double Operators

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Double Stack Operations

These routines manage double numbers on the data stack. They behave
like their cell-sized counterparts.

S>D sign extends a single precision number into a double number.
-------------------------------------------------------------------- }

ICODE 2DROP ( x1 x2 -- )
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE 2DUP ( x1 x2 -- x1 x2 x1 x2 )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   2 CELLS [EBP] EAX MOV   EAX 0 [EBP] MOV
   RET   END-CODE

ICODE 2SWAP ( x1 x2 x3 x4 -- x3 x4 x1 x2 )
   CELL [EBP] EAX MOV   EBX CELL [EBP] MOV
   EAX EBX MOV   0 [EBP] EAX MOV
   2 CELLS [EBP] ECX MOV   EAX 2 CELLS [EBP] MOV
   ECX 0 [EBP] MOV   RET   END-CODE

ICODE 2OVER ( x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2 )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   4 CELLS [EBP] EAX MOV   EAX 0 [EBP] MOV
   3 CELLS [EBP] EBX MOV   RET   END-CODE

ICODE 2NIP ( x1 x2 x3 x4 -- x3 x4 )
   0 [EBP] EAX MOV   EAX 2 CELLS [EBP] MOV
   2 CELLS [EBP] EBP LEA   RET   END-CODE

: 2ROT ( x1 x2 x3 x4 x5 x6 -- x3 x4 x5 x6 x1 x2 )
   2>R 2SWAP 2R> 2SWAP ;

ICODE S>D ( n -- d )
   PUSH(EBX)   EBX EBX ADD   EBX EBX SBB
   RET   END-CODE

: D>S ( d -- n )   DROP ;

{ --------------------------------------------------------------------
Double number arithmetic

M+ adds signed single number n to d1, yielding d2.
-------------------------------------------------------------------- }

ICODE D+ ( d1 d2 -- d3 )
   0 [EBP] EAX MOV   2 CELLS [EBP] EAX ADD
   CELL [EBP] EBX ADC   EAX 2 CELLS [EBP] MOV
   2 CELLS [EBP] EBP LEA   RET   END-CODE


ICODE DNEGATE ( d1 -- d2 )
   0 [EBP] NOT                      \ invert d1-lo
   EBX NOT                          \ invert d1-hi
   1 # 0 [EBP] ADD                  \ make two's compliment
   0 # EBX ADC                      \ from invert + 1
   RET END-CODE

: DABS ( d1 -- +d2 )
   DUP 0< IF  DNEGATE  THEN ;

ICODE D2/ ( d1 -- d2 )
   EBX SAR   0 [EBP] RCR   RET   END-CODE

ICODE D2* ( d1 -- d2 )
   0 [EBP] SHL   EBX RCL   RET   END-CODE

: D- ( d1|ud1 d2|ud2 -- d3|ud3 )
   DNEGATE D+ ;

ICODE M+ ( d1 n -- d2 )
   EBX EAX MOV   CDQ
   CELL [EBP] EAX ADD   0 [EBP] EDX ADC
   EAX CELL [EBP] MOV   EDX EBX MOV
   CELL [EBP] EBP LEA   RET   END-CODE

{ --------------------------------------------------------------------
Double comparisons
-------------------------------------------------------------------- }

: D0= ( d -- flag )   OR 0= ;
: D0< ( d -- flag )   NIP 0< ;
: D=  ( d1 d2 -- flag )   D- OR 0= ;

: D< ( d1 d2 -- flag )
   ROT  2DUP = IF  2DROP U<  EXIT THEN  2NIP > ;

: DU< ( ud1 ud2 -- flag )
   ROT  2DUP = IF  2DROP U<  EXIT THEN  2NIP SWAP U< ;

: DMAX ( d1 d2 -- d3 )
   2OVER 2OVER D< IF  2SWAP  THEN  2DROP ;

: DMIN ( d d2 -- d3 )
   2OVER 2OVER D< NOT IF  2SWAP  THEN  2DROP ;
