{ ====================================================================
File operations

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
File position, append

REWIND-FILE moves the logical file position to the start of the file.
END-OF-FILE positions the file at its end.

APPEND-FILE opens or creates the specified file.  If the file exists,
it is opened with the file position set to the end, ready to append
data. If the file does not exist, it is created.
-------------------------------------------------------------------- }

: REWIND-FILE ( fid -- ior )
   0 0 ROT REPOSITION-FILE ;

: END-OF-FILE ( fid -- ior )
   DUP FILE-SIZE DROP ROT REPOSITION-FILE ;

: APPEND-FILE ( addr u -- fid ior )
   2DUP FILE-STATUS NIP IF ( not exist)
   R/W CREATE-FILE EXIT ( return)  THEN
   R/W OPEN-FILE DUP ?EXIT DROP
   DUP END-OF-FILE ;

{ --------------------------------------------------------------------
File number, errors

We keep up with file numbers in the double variable F#.  The scheme
is that each word has a file number in its locate field; the number
is 16 bits, and the file which it represents is found by searching
the wordlist FILES-WORDLIST for a match.  The cell at F# is the current
filenumber; the second cell is the largest used so far. This is
required to deal properly with nesting of file includes, since each
new file must get a larger number than any previously assigned.

'FNAME is a pointer to the filename in a temporary buffer allocated on
the return stack in INCLUDED.

INCLUDING returns the address and count of the filename currently
being included.
-------------------------------------------------------------------- }

VARIABLE 'FNAME

: INCLUDING ( -- c-addr u )   'FNAME @ ?DUP IF  ZCOUNT
   ELSE  HERE 0  THEN ;

: FNUM++ ( -- )
   F# CELL+ @ 1+  DUP F# 2! ;

: ?CR ( -- )   GET-XY DROP IF CR THEN ;

: .FERROR ( ior1 -- ior2 )   ?CR BOLD
   LINE @  'FNAME @  ?DUP IF  VIEWED @ 0= IF  OVER VIEWED !
      DUP ZCOUNT VIEWED CELL+ PLACE  THEN
   ZCOUNT TYPE  ELSE  ." INCLUDE-FILE" THEN  NORMAL
   BASE @ DECIMAL  CR  SWAP 5 U.R ." : "  BASE !
   DUP 1+ IF  .ERRLINE  ." >>> " .CATCH  -1
   ELSE  #TIB 2@ TYPE SPACE  THEN ;

{ --------------------------------------------------------------------
Include file

INCLUDE/ warns if any words were obscured during the include.

=FILENAME adds the file to FILES-WORDLIST with file# n (if n is not
already in the list).

(NEXTLINE) is the inner working of REFILL-NEXTLINE.  It takes the
buffer and its total length addr1 u1, finds the next line (if any),
returning the line addr (same addr1) and its length u2 and the
remaining buffer address and length addr2 u3.

REFILL-NEXTLINE gets the next line from the file in the virtual memory
buffer whose remaining length and address are in the pair of cells
pointed to by >MAPPED.  There is no input line length restriction.

MONITOR is a stack-passive vector which can be used to watch the file
being included. It is called once for each line read before the line is
interpreted.

(INCLUDES) skips the first line if this is a Unix script file, then
interprets one "line" at a time from the memory-mapped source file
buffer until the end is reached.

INCLUDE-FILE calls MAP-FILE, which allocates a virtual memory buffer
for file fid and reads the file into the buffer.  It saves the total
address and length in a pair of cells in >MAPPED.  The file
is then processed with (INCLUDES).  When (INCLUDES) finishes (or
throws an error), the virtual memory is released back to the OS and
the source file is closed.

INCLUDE-FILE begins at the current file position and interprets text
thru the end of file. At eof, the file is closed.

/INCLUDE clears the OBSCURED count if the INCLUDE operation was invoked
from the command line.

INCLUDED opens a file and includes it from the beginning.  An ANS-
Forth transparent side effect here is that LOCATE information is kept
up with, a filename is added to the FILES-WORDLIST wordlist _AFTER_ the
file is included, regardless of errors during compilation.

INCLUDE parses a filename, adds the default extension, and includes it.
-------------------------------------------------------------------- }

: /INCLUDE ( -- )
   SOURCE-ID ?EXIT  OBSCURED OFF  VIEWED OFF ;

: INCLUDE/ ( -- )
   SOURCE-ID ?EXIT  OBSCURED @ -EXIT  $10 WARNS -EXIT
   CR BRIGHT OBSCURED @ DUP . 1 = IF
         ." definition was hidden"
   ELSE  ." definitions were hidden"  THEN  NORMAL CR ;

VARIABLE LASTFILE       \ Pointer to name field of last file name

: =FILENAME ( n zstr -- )
   R-BUF  ZCOUNT R@ ZPLACE  >R
   FILES-WORDLIST WID> CELL+ BEGIN ( link)
      @REL DUP WHILE  DUP LINK> >BODY @ R@ = IF ( match)
   DROP 2R> 2DROP EXIT  THEN  REPEAT DROP
   R> R> -ROOT FILES-WORDLIST -? (WID-CREATE) ,
   LAST @ LASTFILE ! ;

| CODE (NEXTLINE) ( addr1 u1 -- addr1 u2 addr2 u3 )
   0 [EBP] ECX MOV                              \ ECX EBX: buf addr len
   ECX EDX MOV   EAX EAX XOR                    \ EDX EAX: line addr len
   BEGIN   EBX EBX OR   0> WHILE                \ len=0: skip to end
      $0A # 0 [ECX] BYTE CMP 0= NOT IF          \ =LF: skip to end
      $0D # 0 [ECX] BYTE CMP   0= NOT IF        \ =CR: skip to end
      $20 # 0 [ECX] BYTE CMP   U< IF            \ other ctrl char?
      $20 # 0 [ECX] BYTE MOV   THEN             \ substitute space
      EBX DEC   ECX INC   EAX INC               \ advance
   ROT AGAIN                                    \ back to BEGIN
   THEN ( cr)   EBX DEC   ECX INC               \ resolve fwd branch for =CR
   THEN ( lf)   EBX DEC   ECX INC               \ resolve fwd branch for =LF
   THEN ( len=0)   -2 CELLS [EBP] EBP LEA
   EDX 2 CELLS [EBP] MOV   EAX CELL [EBP] MOV
   ECX 0 [EBP] MOV   RET   END-CODE

: REFILL-NEXTLINE ( -- flag )
   >MAPPED 2@ DUP IF  (NEXTLINE) >MAPPED 2!
      #TIB 2!  0 >IN !  LINE ++  -1 EXIT
   THEN  NIP  #TIB OFF ;

' REFILL-NEXTLINE ' REFILL-LINE >BODY !

XDEFER MONITOR ( -- ) 0 ,

| : (INCLUDES) ( -- )
   >MAPPED @ IF
      [DEFINED] ?SCRIPT [IF]  >MAPPED CELL+ @
      ?SCRIPT IF  REFILL-NEXTLINE DROP  THEN  [THEN]
      BEGIN  REFILL-NEXTLINE  WHILE  MONITOR
   INTERPRET  REPEAT  THEN ;

: INCLUDE-FILE ( fid -- )
   SAVE-INPUT N>R  >IN OFF  LINE OFF
   DUP 'SOURCE-ID !  MAP-FILE ?DUP ( *) 0= IF
      2DUP 2>R  >MAPPED 2!  ['] (INCLUDES) CATCH ( *)
   DUP IF  .FERROR  THEN  2R> UNMAP-FILE DROP  THEN
   SOURCE-ID CLOSE-FILE  DROP  NR> RESTORE-INPUT DROP
   ( *) THROW ;

: INCLUDED ( c-addr u -- )
   /INCLUDE  R-BUF  R@ FULLNAME         \ qualify the name
   R@ ZCOUNT R/O OPEN-FILE THROW        \ open the file
   R> 'FNAME @ >R 'FNAME !              \ point 'fname to our name
   F# @ >R  FNUM++                      \ save starting F#, make us new
   ['] INCLUDE-FILE CATCH ( *)
   DUP IF  NIP  THEN                    \ ior, discarding remainder
   HERE FENCE @REL = >R                 \ was the last thing a GILD?
   F# @ 'FNAME @  =FILENAME             \ keep filename for locate
   R> IF  HERE FENCE !REL  THEN         \ reset FENCE to preserve GILDing
   R> F# !  R> 'FNAME !                 \ restore f# and fname
   INCLUDE/  ( *) THROW ;               \ restore 'fname, etc

: FILENAME ( -- addr len )
   >IN @ >R  BL WORD COUNT  OVER C@ [CHAR] " = IF
   2DROP  R@ >IN !  [CHAR] " WORD COUNT  THEN
   R> DROP  FILENAME-FIXUP 0= IF  2DROP HERE 0  THEN ;

: INCLUDE ( -- )   FILENAME DUP 0= -38 ?THROW INCLUDED ;

{ --------------------------------------------------------------------
Require file

?INCLUDED returns true if the file name in the buffer addr u has
already been included (i.e., it is found in FILES-WORDLIST).

REQUIRED skips the file if ?INCLUDED returns true or includes if not.

REQUIRE parses the file name and passes it to REQUIRED.
-------------------------------------------------------------------- }

: ?INCLUDED ( addr u -- flag )
   FILENAME-FIXUP 0= -38 AND THROW
   R-BUF  R@ FULLNAME
   FILES-WORDLIST WID> CELL+ BEGIN ( link)
      @REL DUP WHILE
      DUP LINK> >NAME COUNT +ROOT
   R@ ZCOUNT COMPARE(NC) 0= UNTIL THEN
   R> DROP  0<> ;

: REQUIRED ( i * x c-addr u -- i * x )
   2DUP ?INCLUDED IF  2DROP EXIT  THEN  INCLUDED ;

: REQUIRE ( i * x "name" -- i * x )
   FILENAME REQUIRED ;

{ --------------------------------------------------------------------
Read and write lines

EOL-SCANNER provides the internal working for READ-LINE. It looks for
any valid line terminator (CRLF, CR, or LF).

READ-LINE gets a line of data from the file. The buffer at addr must be
larger by 2 characters than the maximum size you want read. There are
three states of the file for input:

  1) the line was less than #ask bytes; the entire line is read into
  memory, and the file pointer is set past the terminating line end

  2) the line was exactly #ask bytes; the entire line is read into
  memory, but the file pointer is pointing to the line end

  3) the line was larger than #ask bytes; #ask bytes are read, and the
  file pointer is left pointing to the next unread byte.

An ior is not returned; any error is interpreted as end of file.
Otherwise, the number of bytes read and true is returned.

WRITE-LINE writes the text to a file, followed by the implementation-
defined <EOL> string.
-------------------------------------------------------------------- }

$0D EQU <CR>   $0A EQU <LF>     \ ASCII CR, LF character codes

CODE EOL-SCANNER ( addr len -- got #adv )
   <CR> # AL MOV                \ CR, LF characters for fast compare
   <LF> # AH MOV
   0 [EBP] EDX MOV              \ EDX = addr1
   EBX ECX MOV   ECX ECX TEST   \ ECX=len, test, set flags
   BEGIN   0> WHILE             \ loop while len>0
      0 [EDX] AL CMP   0= IF    \ CR?
         EDX EBX MOV            \ EBX = addr2
         0 [EBP] EBX SUB        \ EBX = (addr2-addr1)
         EBX 0 [EBP] MOV        \ return got
         EBX INC                \ TOS = #adv
         1 [EDX] AH CMP         \ Check for LF after CR
         0= IF  EBX INC  THEN   \ skip LF
      RET   THEN
      0 [EDX] AH CMP            \ LF?
      0= IF   EDX EBX MOV       \ EBX = addr2
         0 [EBP] EBX SUB        \ EBX = (addr2-addr1)
         EBX 0 [EBP] MOV        \ return got
         EBX INC                \ TOS = #adv
      RET   THEN
      EDX INC   ECX DEC         \ addr+1, len-1
   REPEAT
   EBX 0 [EBP] MOV               \ return len len
   RET   END-CODE

: READ-LINE ( addr ask fid -- got flag ior )
   DUP >R FILE-POSITION DROP 2SWAP      \ ud addr ask      . fid
   2DUP 1+ R@ READ-FILE 0= AND          \ ud addr ask read . fid
   DUP IF                               \ ud addr ask read . fid
      MIN EOL-SCANNER                   \ ud got #adv      . fid
      2SWAP ROT M+
      R> REPOSITION-FILE DROP
      TRUE 0
   ELSE  R> 6 DISCARD 0 0 0  THEN ;

: WRITE-LINE ( c-addr u fileid -- ior )
   DUP>R  WRITE-FILE DROP  <EOL> COUNT R> WRITE-FILE ;
