{ ====================================================================
Dictionary search

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
The dictionary search is done by searching only the appropriate strand
of a wordlist.  Which strand is decided by a HASH function, depending
on the number of strands in the wordlist and the text to look for.

All hashes are case insensitive.  Searches may or may not be case
sensitive.

SEARCH-STRAND searches only the specified wordlist strand for the given
string.  if the word is found the following flags are returned:

      1: IMMEDIATE
     -1: NON-IMMEDIATE

(+/-CAPS) modifies the code to perform either a case-sensitive or
   case-in-sensitive search.
-------------------------------------------------------------------- }

CODE SEARCH-STRAND ( c-addr u strand -- 0 | xt flag )
   EDI EDX MOV                  \ copy of origin pointer
   0 [EBP] EAX MOV              \ get count
   CELL [EBP] EBP LEA           \ and discard its cell
   ESI PUSH   EDI PUSH          \ save user and origin pointers
   BEGIN BEGIN                  \ outer loop
      EBX ESI MOV               \ link address
      BEGIN
         0 [ESI] EDI MOV        \ next link
         EDI EDI OR             \ check for end
         0= IF                  \ quit on end of thread
            CELL [EBP] EBP LEA  \ discard stack space reserved
            EBX EBX SUB         \ return 0
            EDI POP   ESI POP   \ restore pointers
            RET
         THEN
         EDI ESI ADD            \ unrelative the non-zero link
         AL 6 [ESI] CMP         \ check count
         0= UNTIL               \ go to next header if unequal
      ESI EBX MOV               \ save a copy of lfa in EBX
      7 # ESI ADD               \ point to name
      EAX ECX MOV               \ count
      0 [EBP] EDI MOV           \ c-addr for compare
      0 # CAPS [EDX] CMP
      0= NOT IF
         REPE-CMPSB(UC) CALL    \ compare strings case-insensitive
      ELSE
         REPE CMPSB             \ compare strings
      THEN
      0= UNTIL                  \ repeat if no match
      $80 # 5 [EBX] BYTE TEST   \ check for smudge
   0= UNTIL                     \ not zero, was smudged, strand
   ESI INC                      \ point to code address
   $20 # 5 [EBX] BYTE TEST      \ check for alias
   0= NOT IF   0 [ESI] ESI ADD   THEN
   EDX ESI SUB                  \ now is xt
   ESI 0 [EBP] MOV              \ write onto stack
   EDI POP   ESI POP            \ restore pointers
   -CELL [EBX] EAX MOV          \ locate field
   EAX 'LOCATE [ESI] MOV        \ save in case needed
   $40 # 5 [EBX] BYTE TEST      \ check immediate bit
   0= IF   -1 # EBX MOV   RET   THEN  \ -1 means not immediate
   1 # EBX MOV                  \ 1 means non-immediate
   RET   END-CODE

{ ------------------------------------------------------------------------
Thread hash

The dictionary strands are hashed to improve search speed.

HASH takes a string ( addr len ) and a number of strands n and
computes which strand the word will be in.  The routine works best for
prime N . Returns hash index in cells.

: HASH ( c-addr u n -- hash )
   >R 0 -ROT BOUNDS DO
      4 LSHIFT  I C@ $20 or +
      DUP $F0000000 AND ?DUP IF  24 RSHIFT XOR  $0FFFFFFF AND THEN
   LOOP  R> ( #STRANDS)  MOD CELLS ;

HASHED takes a string and a WID, and returns the address of the strand
that the string belongs in.
------------------------------------------------------------------------ }

CODE HASH ( c-addr u n -- hash )
   ESI PUSH   EBX PUSH
   0 [EBP] EDX MOV              \ and edx is the loop counter
   EBX EBX XOR                  \ ebx is the accumulator, init = 0
   EAX EAX XOR                  \ eax is the character load & add register
   CELL [EBP] ESI MOV
   BEGIN
      4 # CL MOV                \ shift for accumulator
      EBX CL SHL                \ shift left
      LODSB                     \ one char
      $20 # AL OR               \ mask to "lower case"
      EAX EBX ADD               \ into the accumulator
      EBX EAX MOV               \ copy for destructive test
      $F0000000 # EAX AND       \ mask for high 4 bits
      0= NOT IF                 \ skip if no high bits set
         EAX EBX XOR            \ clear high bits
         24 # CL MOV            \ shift 24
         EAX CL SHR             \ to the right
         EAX EBX XOR            \ and xor back into the accumulator
      THEN
      EDX DEC                   \ decrement loop counter
   0= UNTIL                     \ and continue while non zero
   EBX EAX MOV                  \ into eax for divide, edx is zero from dec/jnz
   EBX POP
   EBX DIV                      \ do division
   EDX EBX MOV                  \ return remainder
   EBX 2 # SHL                  \ as a cell quantity
   ESI POP
   2 CELLS [EBP] EBP LEA
   RET   END-CODE

: HASHED ( c-addr n wid -- strand )
   +ORIGIN @+ SWAP >R HASH R> + ;

{ ------------------------------------------------------------------------
Cross reference

Built-in methods for keeping cross reference data until the real
cross reference utility is installed
------------------------------------------------------------------------ }

2VARIABLE 'REF   | VARIABLE 'TREF

| : (SREF) ( xt -- )
   LINE @ F# @ 16 LSHIFT OR  SWAP  'REF @ 2!  8 'REF +! ;

| : /SREF ( addr -- )
   $100000 + DUP 'REF 2!                \ Set ref list to 1M after start of dictionary
   'TREF @REL  ?DUP IF  DUP @+ >R       \ If 'TREF points to a kernel ref list
      'REF @ R@ CMOVE                   \ ...import it
      DUP @ CELL+ ERASE                 \ Then clear the space it occupied
   R> 'REF +!  0 'TREF !  THEN ;        \ Clear 'TREF so we don't try to import it again

XDEFER >XREF   ' (SREF) ,  ( xt -- )

{ --------------------------------------------------------------------
SEARCH-WORDLIST conforms to the ans definition, except that it will
adjust the search mechanism based on the current state of the CAPS
variable. It is the most primitive of the context-based search words.
It takes a counted string and a WID, and returns either the address of
the string and zero if not found; or the xt of the word and a flag
containing the word's immediate status.

 -1: NON-IMMEDIATE
  0: NOT FOUND
  1: IMMEDIATE

FIND searches for the word only if it has a non-zero count.

DEFINED parses the next word and returns flags about found or not.

EXISTS returns true if the word following it exists in the search order.

ABSENT returns true if the word does not exist.
-------------------------------------------------------------------- }

: SEARCH-WORDLIST ( c-addr u wid -- 0 | xt flag )
   OVER IF
      THIRD THIRD ROT HASHED   SEARCH-STRAND DUP IF OVER >XREF THEN
      EXIT
   THEN  3DROP 0 ;

: (FIND) ( c-addr len -- 0 | xt flag )
   CONTEXT #ORDER @ CELLS OVER + SWAP ?DO
      2DUP I @ SEARCH-WORDLIST ?DUP IF  2NIP UNLOOP EXIT  THEN
   CELL +LOOP 2DROP 0 ;

: FIND ( c-addr -- c-addr 0 | xt flag )
   DUP COUNT DUP IF ( a a n)
      (FIND) DUP IF ( a xt t) ROT DROP THEN EXIT
   THEN 2DROP 0 ;

: DEFINED ( -- c-addr 0 | xt -1 | -- xt 1)
   BL WORD FIND ;

: EXISTS ( -- flag )   DEFINED 0<> NIP ;
: ABSENT ( -- flag )   EXISTS NOT ;
