{ ====================================================================
Header creation

Copyright 2001  FORTH, Inc.
Rick VanNorman

+-------------------------------------------------------------------------------
| locate field
|               +---------------------------------------------------------------
|               | link field, contents relative or zero
|               |               +-----------------------------------------------
|               |               | sentinel, always $FF
|               |               |   +-------------------------------------------
|               |               |   | flags: $80=smudge, $40=immediate, $20=alias
|               |               |   |   +---------------------------------------
|               |               |   |   | length of name, range 0-254
|               |               |   |   |   +-----------------------------------
|               |               |   |   |   | name, stored as ASCII
|               |               |   |   |   |         +-------------------------
|               |               |   |   |   |         | inline byte, 0=call only
|               |               |   |   |   |         |   +---------------------
|               |               |   |   |   |         |   |
|   .   .   .   |   .   .   .   | x | f | # | n a m e | i | cfa
---------------------------------------------------------------

==================================================================== }

{ ------------------------------------------------------------------------
Building headers

IMMEDIATE  sets the immediate bit in a header.

-SMUDGE   guarantees that the smudge bit is turned off.
+SMUDGE   guarantees that the smudge bit is turned on.

?UNIQUE checks the WID for a string matching ADDR N and warns the user.
------------------------------------------------------------------------ }

| : 'LAST ( -- addr )   LAST @ DUP IF 1- EXIT THEN  DROP  HERE ;

| : SET-LAST ( or and -- )
   'LAST C@  AND OR  'LAST C! ;

HEX

: +SMUDGE ( -- )   80 FF SET-LAST ;
: -SMUDGE ( -- )   00 7F SET-LAST ;
: IMMEDIATE ( -- )   40 FF SET-LAST ;
: +ALIAS ( -- )  20 FF SET-LAST ;

DECIMAL

| : .XTNAME ( xt -- )   CR >NAME COUNT TYPE ;

| : OBSCURES ( xt -- )
   .XTNAME  ."  isn't unique. "  OBSCURED ++ ;

| : AMBIGUOUS ( xt -- )
   .XTNAME  ."  is case-ambiguous. " ;

| : ?UNIQUE ( addr n wid -- addr n wid )
   $10 WARNS IF
      3DUP SEARCH-WORDLIST IF  OBSCURES  ELSE
         $20 WARNS IF
            CAPS @ 0= IF
               CAPS ON
               3DUP SEARCH-WORDLIST IF AMBIGUOUS THEN
               CAPS OFF
            THEN
         THEN
      THEN
   THEN /WARNING ;

{ ------------------------------------------------------------------------
Header Creation

LOCATION returns the 32-bit locate field.  High word is F#, low word
is LINE.
------------------------------------------------------------------------ }

: LOCATION ( -- n )
   SOURCE=FILE IF  LINE @ $FFFF MIN  F# @ 16 LSHIFT OR
   ELSE  0  THEN ;

: LINKED ( addr n wid -- )
   HASHED ( strand)  DUP LAST 3 CELLS + !
   DUP @REL ,REL  HERE 4 - SWAP !REL ;

: EMPLACE ( addr u -- )
   HERE LAST !  HERE  OVER 1+ ALLOT  PLACE ;

{ --------------------------------------------------------------------
Since the code field of a definition is always at
here + cell + cell + cell + len
or here + 12 + len (where len occupies one byte)
if we want the code field to always be at a paragraph boundary
for here = 0
len = 1 pad = 3
len = 2 pad = 2
len = 3 pad = 1
len = 4 pad = 0
len = 5 pad = 15

16 here 12 + len + 15 and -

: z ( n -- )   >r  cr
   here h.
   here 12 + r@ + 15 and  16 swap -  here + dup h.
   12 + r> + h. ;

HEADER creates a non-executable header for the text following.
(WID-CREATE) makes an executable header from the text following.
-------------------------------------------------------------------- }

: WID-HEADER ( c-addr u wid -- )
   OVER 0= -16 ?THROW  ?UNIQUE
   -ROT 254 MIN R-BUF R@ PLACE
   HERE 13 + R@ C@ + 15 AND  16 SWAP - ALLOT
   LOCATION ,  R@ COUNT ROT LINKED              \ link into wid
   $FF C, 0 C,                                  \ sentinel and flags
   R> COUNT EMPLACE                             \ name
   HERE 1+ HERE LAST CELL+ 2!  0 C, ;           \ remember inline and code

: HEADER ( -- )
   BL WORD COUNT GET-CURRENT WID-HEADER ;

: (WID-CREATE) ( addr n wid -- )
   WID-HEADER  POSTPONE (CREATE) ;

: (.ID) ( nfa -- )
   ?DUP IF COUNT ?TYPE SPACE ELSE ." [noname]" THEN ;

XDEFER .ID ( nfa -- )   ' (.ID) ,

{ --------------------------------------------------------------------
Case sensitivity
-------------------------------------------------------------------- }

: CASE-SENSITIVE ( -- )   CAPS OFF ;
: CASE-INSENSITIVE ( -- )   CAPS ON ;
