{ ====================================================================
Import library functions

Copyright 2009  FORTH, Inc.

The calling of external library functions is implemented here.

======================================================================

When we call an external library function, we move the parameters from
the Forth data stack (pointed to by EPB) to the system stack (pointed
to by ESP).  We call all external library functions that we have
addresses of via the EXTERN-CALL routine, which handles the details of
building the stack frame for external procedure call.

LIBRARY parameter field:  | handle | flag |
IMPORT: parameter field:  | proc addr | #params | ^lib |
GLOBAL: parameter field:  |  data addr | 0 | ^lib |
==================================================================== }

TARGET

[UNDEFINED] EXTERN-CALL [IF]    \ macOS kernel supplies its own

CODE EXTERN-CALL ( x1 x2 .. xn n a -- x )
   EDI PUSH   ESI PUSH          \ save Forth VM context
   0 [EBP] ECX MOV              \ n args
   ESP EDX MOV                  \ save ESP before building stack frame
   ECXNZ IF                     \ skip building stack frame if no args
      BEGIN
         CELL [EBP] PUSH        \ move args from data stack to system stack
         CELL [EBP] EBP LEA
   LOOP THEN
   EDX 0 [EBP] MOV              \ preserve ESP for return
   EBX CALL                     \ call extern function
   EAX EBX MOV                  \ return value (LSB)
   0 [EBP] ESP MOV              \ restore ESP
   CELL [EBP] EBP LEA
   ESI POP   EDI POP            \ restore Forth VM
   RET   END-CODE

[THEN]

{ --------------------------------------------------------------------
External libraries

LIBRARY defines an external library whose dictionary header is linked
to the LIBS wordlist.
-------------------------------------------------------------------- }

HOST

VARIABLE 'LIB

INTERPRETER

: LIBRARY ( -- )
   HERE(T)  =LIBS @REL(T) ,REL(T)  =LIBS !REL(T)
   MARKER,  FLAGS,  BL WORD COUNT NAME,  0 C,
   POSTPONE(T) (CREATE)  HERE 'LIB !(H)  0 ,  -1 , ;

TARGET

{ --------------------------------------------------------------------
IMPORTS is the head of a list, maintained by the interpreter.

FUNCTION: is a compiler extension to lay down a library call.
GLOBAL: imports a library's global data address.

'EXTCALL is used after the decompiler is loaded in the IDE.
-------------------------------------------------------------------- }

CREATE IMPORTS  0 , 0 ,

HOST

: AS-NAME ( -- addr | 0 )
   >AS @ DUP IF  +SMUDGE(T)  >IN @ >R  >IN !
   'CFAX @  CREATE(T)  R> >IN !  0 >AS !  THEN ;

INTERPRETER

: EXTCALL ( -- )   DOES> ( i*x -- x )
   BEGIN  2@ EXTERN-CALL ;

: FUNCTION: ( -- )
   PARAMETERS()  CREATE  0 , ,  'LIB @(H) ,REL(T)
   HERE(T)  IMPORTS @REL(T) ,REL(T)  'LAST @(H) ,REL(T)  IMPORTS !REL(T)
   EXTCALL  AS-NAME ?DUP IF   $E9 C,  HERE 4 + - L,  THEN ;

: GLOBAL: ( -- )
   CREATE  0 , 0 ,  'LIB @(H) ,REL(T)
   HERE(T)  IMPORTS @REL(T) ,REL(T)  'LAST @(H) ,REL(T)  IMPORTS !REL(T)
   AS-NAME ?DUP IF   $E9 C,  HERE 4 + - L,  THEN
   DOES> ( -- x )   @ ;

TARGET

( addr) 7 - CONSTANT 'EXTCALL   \ xt of code in DOES> in EXTCALL above
