{ ====================================================================
File operations

Copyright 2007  FORTH, Inc.
Leon Wagner

This file supplies the Linux-specific ANS Forth File Access and File-
access Extension words.  Generic file operators are in the common
kernel source.

==================================================================== }

{ --------------------------------------------------------------------
ANS File Access Methods

The ANS file interface requires these methods for opening and creating
files.  The BIN method does nothing, as all access is binary.  The
non- ANS fam modifier EXE is supplied to set the "execute" bits in the
upper half of the fam.
-------------------------------------------------------------------- }

OCTAL

   0 EQU O_RDONLY     \ Flags for open
   1 EQU O_WRONLY
   2 EQU O_RDWR
 100 EQU O_CREAT
1000 EQU O_TRUNC

DECIMAL

O_RDONLY CONSTANT R/O
O_WRONLY CONSTANT W/O
O_RDWR   CONSTANT R/W

: BIN ( fam --  fam ) ;  IMMEDIATE
: EXE ( fam1 -- fam2)   [ &111 16 LSHIFT ] LITERAL OR ;

{ --------------------------------------------------------------------
ANS file access

The fileid tokens used here are Linux file descriptors.

CREATE-FILE makes a new file and opens it for I/O. If the file already
exists, it is truncated.  The newfile is created with '0666'
permissions, modified by the caller's umask.

OPEN-FILE opens an existing file for I/O based on fam.
CLOSE-FILE closes an open file.
DELETE-FILE removes the file from the filesystem.
RENAME-FILE changes the name of a file in the filesystem.
-------------------------------------------------------------------- }

: CREATE-FILE ( c-addr1 u fam -- fileid ior )
   DUP 3 AND [ O_CREAT O_TRUNC OR ] LITERAL OR  SWAP 16 RSHIFT &666 OR
   2SWAP R-BUF R@ ZPLACE
   R> -ROT open  DUP 0< -198 AND ;

: OPEN-FILE ( c-addr u fam -- fileid ior )
   3 AND 0 2SWAP  R-BUF R@ ZPLACE
   R> -ROT open  DUP 0< -199 AND ;

: CLOSE-FILE ( filed -- ior )   close 0<> -197 AND ;

: DELETE-FILE ( c-addr u -- ior )
   R-BUF R@ ZPLACE  R> sys_unlink  0<> -191 AND ;

: RENAME-FILE ( caddr1 u1 caddr2 u2 -- ior )
   R-BUF R@ ZPLACE R> -ROT  R-BUF R@ ZPLACE R> SWAP rename  0<> -192 AND ;

{ --------------------------------------------------------------------
ANS file read/write

READ-FILE attempts to read u1 bytes from the file into memory at the
address.  The actual number of bytes read is returned as u2.

WRITE-FILE writes u bytes to the file.

FLUSH-FILE asks the OS to write all cached data to disk.
-------------------------------------------------------------------- }

: READ-FILE ( c-addr u1 fileid -- u2 ior)
   -ROT read  DUP 0 MAX  SWAP 0< -195 AND ;

: WRITE-FILE ( c-addr u fileid -- ior)
   -ROT write  0< -196 AND ;

: FLUSH-FILE ( fileid -- ior)
   fsync 0< -194 AND ;

{ --------------------------------------------------------------------
ANS file size and position

FILE-SIZE returns the size in bytes of a file.
FILE-POSITION returns the linear position of the file.
REPOSITION-FILE forces the file to a linear position from its start.
RESIZE-FILE sets a new size for the file.
FILE-STATUS returns zero if the file exists, or an ior if not.

lseek(int filed, off_t offset, int whence);

The SEEK_xxx equates are for the <whence> parameter in lseek() which
describes the use of offset.
-------------------------------------------------------------------- }

0 EQU SEEK_SET          \ From beginning of file
1 EQU SEEK_CUR          \ From current position
2 EQU SEEK_END          \ From end of file

: FILE-SIZE ( fileid -- ud ior )
   DUP 0 SEEK_CUR lseek >R      \ current position
   DUP 0 SEEK_END lseek         \ set to end (returns size)
   SWAP R> SEEK_SET lseek DROP  \ restore current position
   0 OVER -1 = -37 AND ;        \ return ud, ior

: FILE-POSITION ( fileid -- ud ior )
   0 SEEK_CUR lseek  0 OVER 0< -36 AND ;

: REPOSITION-FILE ( ud fileid -- ior )
   SWAP IF  2DROP -36 EXIT  THEN
   SWAP SEEK_SET lseek  0< -36 AND ;

: RESIZE-FILE ( ud fileid -- ior )
   -ROT ftruncate64 0<> -193 AND ;

: FILE-STATUS ( c-addr u -- x ior )
   R-BUF R@ ZPLACE  R@ DUP stat  R> W@  SWAP 0< -37 AND ;

{ ------------------------------------------------------------------ }

CREATE <EOL>   1 C, $0A C, 0 C, 0 C,   \ Linux line-end sequence

{ --------------------------------------------------------------------
Memory mapped files

These non-ANS file functions are in support of INCLUDE-FILE.

MAP-FILE takes the fileid of a file that is open for read access
(presumably a source file) and uses syscall mmap() to map the file
into memory.  Returns address, length, ior.  If FILE-STATUS returns
error (file not found) returns error ior and bails out.  Otherise
continues on and makes calls to mmap().

UNMAP-FILE takes the buffer address and total size and frees the
mapped memory.

?SCRIPT returns true if the first line of the memory-mapped source
file pointed to by SOURCE-ID contains the magic "#!" script
identifier.
-------------------------------------------------------------------- }

1 EQU MAP_SHARED
2 EQU MAP_PRIVATE
1 EQU PROT_READ
2 EQU PROT_WRITE
4 EQU PROT_EXEC
3 EQU PROT_RW

: MAP-FILE ( fileid -- addr u ior )
   DUP FILE-SIZE  ?DUP IF  >R  DROP 2DROP  0 0 R> EXIT  THEN DROP       \ fileid, length
   SWAP >R  0 OVER PROT_RW MAP_PRIVATE R> 0 mmap                        \ size, result
   SWAP OVER -4096 0 WITHIN -195 AND ;                                  \ convert result to ior

: UNMAP-FILE ( addr u -- ior )   munmap  0<> -102 AND ;

| : ?SCRIPT ( addr -- flag )   W@  $2123 ( '#!') = ;
