{ ====================================================================
Linux file and path support

Copyright 2007  FORTH, Inc.

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Path support

-NAME strips the name and trailing / from the end of a full pathname.

-PATH strips the path from the beginning of a full pathname.

-TAIL strips the last node off the end of a path in a z-string by
writing a null.
-------------------------------------------------------------------- }

| : ENDING ( addr len char -- addr len )
   >R BEGIN  2DUP R@ SCAN
      ?DUP WHILE  2SWAP 2DROP  1 /STRING
   REPEAT  R> 2DROP ;

: -NAME ( a n -- a n )
   2DUP  [CHAR] / ENDING  NIP -  1-  0 MAX ;

| : -TAIL ( z-addr -- )
   DUP ZCOUNT + BEGIN  1- 2DUP <> WHILE
      DUP 1- C@ [CHAR] / = UNTIL  0 OVER C!
   THEN 2DROP ;

: -PATH ( a n -- a n )
   [CHAR] / ENDING  0 MAX ;

{ --------------------------------------------------------------------
Root path

We do most file operations relative to the "root" directory in which
SwiftForth is installed. This is so the base directory in which it is
installed doesn't matter, and all the files whose names are relative
to this path can be found even if installed on another drive or path.
For all this to work, SwiftForth must be launched either directly or
from a shortcut to the SwiftForth/bin/linux directory.

ROOTPATH holds the root of our system.  This is the "parent" of
the SwiftForth directory.

ROOTED? is true if the string begins with the rootpath string.

-ROOT removes the rootpath from the zstring.

+ROOT adds the rootpath back to the string.
-------------------------------------------------------------------- }

CREATE ROOTPATH   256 ALLOT

: ROOTED? ( addr n -- flag )
   DROP ROOTPATH COUNT SWAP OVER COMPARE 0= ;

: -ROOT ( z-addr -- addr len )
   DUP ZCOUNT ROOTED? IF
      DUP ZCOUNT  ROOTPATH C@ /STRING  ( z a n)
      THIRD S" %" THIRD ZPLACE  ZAPPEND
   THEN ZCOUNT ;

: +ROOT ( addr n -- addr n )
   OVER C@  [CHAR] % = IF
      ROOTPATH COUNT POCKET PLACE
      1 /STRING POCKET APPEND
   POCKET COUNT  THEN ;

{ --------------------------------------------------------------------
File name

FULLNAME places the fully qualified path name into the zbuffer. This
has a limit of 255 characters.

DEFAULT.EXT appends the default extension from DEFEXT if needed.
-------------------------------------------------------------------- }

: FULLNAME ( addr n zbuf -- )   >R
   2DUP + 1- C@ [CHAR] / = IF  2DROP  0 R> C!  EXIT  THEN       \ Ends with '/' - no filename!
   OVER C@ [CHAR] / = IF  R> ZPLACE EXIT  THEN                  \ Starts with '/' - already a full path
   R@ MAX_PATH getcwd DROP  S" /" R@ ZAPPEND                    \ Prime zbuf with current working directory
   BEGIN  2DUP  OVER >R
      S" ./" R@ OVER COMPARE 0= IF  2 /STRING  THEN             \ Eat ./ - silly, but why not?
      S" ../" R> OVER COMPARE 0= IF  R@ -TAIL  3 /STRING  THEN  \ Lop off last dir in cwd for each leading '../'
   2SWAP 2OVER D= UNTIL  R> ZAPPEND ;

CREATE DEFEXT  2 C,  CHAR . C,  CHAR f C,  253 ALLOT

: DEFAULT.EXT ( addr1 n1 -- addr2 n2 )
   2DUP -PATH  [CHAR] . SCAN NIP  ?EXIT
   POCKET PLACE  DEFEXT COUNT POCKET APPEND
   POCKET COUNT ;

| : ROOTED ( addr len -- addr len )   DUP IF
      OVER C@ [CHAR] % = IF +ROOT  HERE PLACE  HERE COUNT
   THEN  THEN ;

: FILENAME-FIXUP ( addr len -- addr len flag )
   -TRAILING DUP IF
      ROOTED  DEFAULT.EXT  2DUP FILE-STATUS NIP 0=
   ELSE  0  THEN ;

{ --------------------------------------------------------------------
Establish paths

THIS-EXE-NAME returns the full path to the current executable. Uses
POCKET.  Length is 0 if both calls to readlink fail.  The first is
for the common Linux path; the second is the FreeBSD version.

/ROOTPATH takes places the root path into ROOTPATH for subsequent use
by the +ROOT and -ROOT operations.  The root path is derived from the
path to the current executable.
-------------------------------------------------------------------- }

: THIS-EXE-NAME ( -- addr u )
   Z" /proc/self/exe" POCKET 256 readlink DUP 0< IF
   DROP Z" /proc/curproc/file" POCKET 256 readlink  THEN
   POCKET SWAP 0 MAX ;

| : /ROOTPATH ( -- )
   THIS-EXE-NAME -NAME -NAME -NAME -NAME  ROOTPATH PLACE
   S" /" ROOTPATH APPEND ;
