{ ====================================================================
Trap handler

Copyright 2008  FORTH, Inc.

Memory structures documented at the end of this file.
====================================================================

Top of machine stack on entry to <EXCEPTION>

 +0  Return addr
 +4  SIG#
 +8  Pointer to siginfo_t structure
+12  Pointer to ucontext_t structure (see below)

==================================================================== }

TARGET

LABEL <EXCEPTION>
   4 [ESP] EAX MOV                      \ EAX = EXCEPTION#
   EBX PUSH   ESI PUSH   EDI PUSH       \ Save registers for return to OS
   24 [ESP] EBX MOV   36 [EBX] EDX MOV  \ EBX = context pointer, EDX = ORIGIN addr
   BEGIN ( *)  0 # EDX MOV              \ Data space origin (updated by /EXCEPTION)
   EAX EXCEPTION# [EDX] MOV             \ Save exception#
   TRACEBACK [EDX] EDI LEA              \ EDI = Address of TRACEBACK buffer in our data space
   20 [EBX] ESI LEA                     \ ESI = Address of first register in machine context
   #TRACE_REGS # ECX MOV   REP MOVSD    \ Copy registers to traceback
   48 [EBX] ESI MOV                     \ ESI = ESP from uc_mcontext
   #TRACE_RSTACK # ECX MOV   REP MOVSD  \ Copy return stack to traceback
   44 [EBX] ESI MOV                     \ EDI = EBP from uc_mcontext
   #TRACE_DSTACK # ECX MOV   REP MOVSD  \ Copy data stack to traceback
   'CATCHING [EDX] EAX LEA   ESP EDI MOV
   $8000 # ECX MOV   REPNE SCASD        \ Find CATCH stack frame
   0= IF   8 [EDI] EAX MOV              \ Get pushed ESI (user pointer)
   EAX 40 [EBX] MOV   THEN              \ Replace ESI in context record so THROW will work correctly
   THROWBACK [EDX] EAX LEA              \ Code address of THROWBACK
   EAX 76 [EBX] MOV                     \ Stuff THROWBACK addr in place of EIP in context record
   EDI POP   ESI POP   EBX POP          \ Restore registers
   EAX EAX XOR   RET   END-CODE         \ Return 0 result

( *) 1+ EQU 'ORIGIN

4 EQU SA_SIGINFO

| : /EXCEPTION ( -- )
   ORIGIN 'ORIGIN +ORIGIN !             \ Stuff ORIGIN address into literal
   R-BUF  R@ 140 ERASE                  \ This is struct sigaction
   <EXCEPTION> +ORIGIN R@ !             \ sa_siginfo handler
   SA_SIGINFO R@ 132 + !                \ sa_flags (skip 128-bit sa_mask field)
   4 R@ 0 sigaction DROP                \ SIGILL - Illegal instruction
   8 R@ 0 sigaction DROP                \ SIGFPE - Arithmetic exception
   11 R@ 0 sigaction DROP               \ SIGSEGV - Seg fault
   2 R@ 0 sigaction DROP                \ SIGINT - User interrupt (Ctrl-C)
   R> DROP ;


\\ =====================================================================

If SA_SIGINFO is specified in sa_flags, then sa_sigaction (instead of
sa_handler) specifies the  signal-handling function  for  signum.
This function receives the signal number as its first argument, a
pointer to a siginfo_t as its second argument and a pointer to a
ucontext_t (cast to void *) as its third argument.

/* Userlevel context.  */
typedef struct ucontext
  {
    unsigned long int uc_flags;         [4]
    struct ucontext *uc_link;           [4]
    stack_t uc_stack;                   [12]
    mcontext_t uc_mcontext;             [88]
    __sigset_t uc_sigmask;              [128]
    struct _libc_fpstate __fpregs_mem;  [112]
  } ucontext_t;                                 [348]

typedef struct sigaltstack
  {
    void *ss_sp;                        [4]
    int ss_flags;                       [4]
    size_t ss_size;                     [4]
  } stack_t;                                    [12]


/* Context to describe whole processor state.  */
typedef struct
  {
    gregset_t gregs;                    [76]
    fpregset_t fpregs;                  [4]
    unsigned long int oldmask;          [4]
    unsigned long int cr2;              [4]
  } mcontext_t;                                 [88]


/* Type for general register.  */
typedef int greg_t;

/* Number of general registers.  */
#define NGREG   19

/* Container for all general registers.  */
typedef greg_t gregset_t[NGREG];

/* Number of each register in the `gregset_t' array.  */
enum
{
  REG_GS = 0,   // +20 (offset into ucontext above)
  REG_FS,       // +24
  REG_ES,       // +28
  REG_DS,       // +32
  REG_EDI,      // +36
  REG_ESI,      // +40
  REG_EBP,      // +44
  REG_ESP,      // +48
  REG_EBX,      // +52
  REG_EDX,      // +56
  REG_ECX,      // +60
  REG_EAX,      // +64
  REG_TRAPNO,   // +68
  REG_ERR,      // +72
  REG_EIP,      // +76
  REG_CS,       // +80
  REG_EFL,      // +84
  REG_UESP,     // +88
  REG_SS        // +92
};
#endif
