{ ====================================================================
Command-line args and environment

Copyright 2007  FORTH, Inc.
Leon Wagner, Ron Oliver

A pointer to the arg and environment list (passed in on the stack in
POWER-UP) is saved in 'ARGS.  This is an array of cells:

argc            [dword] argument counter (integer, at least 1 as
                argv[0] is always present)
argv[0]         [dword] program name (pointer, always present)
argv[1]         [dword] program name (pointer, optional)
...
argv[argc-1]    [dword] program args (pointers, optional)
NULL            [dword] end of args (integer, always present)
envp[0]         [dword] environment variables (pointers, optional,
                variable length list)
envp[1]
...
envp[n]
NULL            [dword] end of environment (integer, always present)

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Command-line args

'ARGS is set to point to the program entry stack frame arg and
environment arrays (see above) in the POWER-UP code.

ARGC returns the arg count (includes arg0, the program name itself).

ARGV returns arg i (counting from 0) as a string.  If i is an invalid arg
index (i.e. not less then argc), returned address is undefined and
length is 0.

CMDLINE concatenates the command-line ARGV strings into a single
command line buffer and returns its address and count. Note: Arguments
are concatenated to a counted string buffer, so the max length is 255.
-------------------------------------------------------------------- }

| VARIABLE 'ARGS

: ARGC ( -- u )   'ARGS @ @ ;

: ARGV ( i -- addr len )   DUP ARGC U< IF
      1+ CELLS 'ARGS @ + @ ZCOUNT
   ELSE  0  THEN ;

| CREATE CMDBUF  256 ALLOT

: CMDLINE ( -- addr len )
   0 CMDBUF C!  ARGC DUP 1 > IF
      DUP 1 DO  I ARGV CMDBUF APPEND  S"  "
      CMDBUF APPEND  LOOP
   THEN DROP  CMDBUF COUNT -TRAILING ;

{ --------------------------------------------------------------------
Environment query

FIND-ENV searches for the string addr1 len1 in the environment and
returns its value and true if found.  Returns the original string
parameters and false if not found.  For example, if the environment
contains the entry "SHELL=/bin/ksh" then the query  S" SHELL" FIND-ENV
returns the address and length of the string "/bin/ksh".  Do not
include the '=' in the search string; FIND-ENV appends it to its
search buffer.

'ENV returns the address of the environment pointers.
.ENV prints the environment for diagnostic purposes.
"printenv" is the usual Unix command to print the environment.
-------------------------------------------------------------------- }

: FIND-ENV ( addr1 len1 -- addr2 len2 flag )
   R-BUF  2DUP R@ PLACE  S" =" R@ APPEND
   'ARGS @ @+ CELLS +  BEGIN  CELL+ DUP @  ?DUP WHILE
      R@ COUNT ROT OVER COMPARE 0= UNTIL
      @ ZCOUNT R> C@ /STRING  2SWAP 2DROP
   -1  EXIT  THEN  R> 2DROP  0 ;

: 'ENV ( -- addr )
   'ARGS @ @+ 1+ CELLS + ;

: .ENV ( -- )   'ARGS @ @+ CELLS +
   BEGIN  CELL+ DUP @  ?DUP WHILE  CR  ZCOUNT TYPE  REPEAT DROP ;

: PRINTENV ( -- )   .ENV ;

{ --------------------------------------------------------------------
Arg conversion support

These words are useful for parsing command lines, which may contain
words, numbers, or single character options.

ATOI converts a counted string to a single integer.  Returns 0 if it
fails.  0x preceeds hex numbers; all others are decimal.

GETCH removes the first character from the string and returns it.
-------------------------------------------------------------------- }

: ATOI ( addr len -- n )   BASE @ >R
   OVER C@ [CHAR] - = DUP >R IF 1 /STRING THEN
   OVER W@ $2020 OR $7830 = IF ( 0x format use hex) HEX 2 /STRING THEN
   0 0 2SWAP >NUMBER 2DROP DROP R> IF NEGATE THEN
   R> BASE ! ;

: GETCH ( addr1 len1 -- addr2 len2 char )
   DUP DUP IF DROP OVER C@ >R 1 /STRING R> THEN ;
