{ ====================================================================
Start SwiftForth

Copyright 2007  FORTH, Inc.

Main entry point and initialization of SwiftForth.
==================================================================== }

TARGET

{ --------------------------------------------------------------------
Command line parsing

PARSE-ARGS evaluates the command line. If the command line contains
only a filename, that file is included.  Otherwise, the string is
evaluated. If an error is caught, the user has the option of pressing
X to terminate the program.

For example, if SwiftForth is started by
   sf foo.f
...the file foo.f will be included if it is found.

If it is started by
   sf here 100 dump
...memory from HERE will be displayed.
-------------------------------------------------------------------- }

| : PARSE-ARGS ( -- )   ARGC 1 > IF             \ Bail if no args following prog name
      1 ARGV FILE-STATUS NIP IF                 \ Check 1st arg as file name
         CMDLINE ['] EVALUATE                   \ Not a file, try to interpret command line tail
      ELSE  1 ARGV ['] INCLUDED  THEN           \ If file found, include it
      CATCH ?DUP IF  DUP .CATCH
         ?TTY @ 0= IF  EXITSTATUS !  BYE  THEN
         BAILOUT?  DROP 2DROP  THEN
   ."  ok" CR QUIT  THEN ;

' PARSE-ARGS ' /CMDLINE >BODY !

{ --------------------------------------------------------------------
START is not a user-accessible function. It simply is enough to get
things going and run the 'MAIN routine.

The behavior differs if this is run from the target-compiled kernel or
from a saved running system.  This difference is detected by the state
of the global variable UP0, which contains the relative address of the
CONSOLE user area after the system gets running. If zero, then we are
running from a system that has never been run before (i.e., just
compiled via the target compiler) and the system memory allocations
are made here. If we have been running, the system allocations are
restored.  Then, the user area is initialized and 'MAIN is executed.
-------------------------------------------------------------------- }

| : START ( -- )
   UP0 @ IF  UP0 @REL /KERNEL DROP      \ If UP0 points to memory, use it
   ELSE  DP0 @REL /KERNEL  DP0 !REL  THEN
   RP@ R0 !  SP@ S0 !  0 CATCHER !      \ Stack is where we find it
   DP0 @REL H !                         \ Start of free memory
   |RAM| +ORIGIN HLIM !                 \ Top of dictionary
   'THIS OFF  'SELF OFF                 \ No object or class
   /FORTH  DECIMAL                      \ Search order, base
   STATE OFF   0 'METHOD !              \ Interpreting
   WARNING ON  $11 WARNING CELL+ !      \ Limited warnings in command window
   /IMPORTS  /SYSTEM  /EXCEPTION        \ Init libraries, system chain, exception handler
   /ROOTPATH  /VERSION  /NDP            \ Set rootpath, version, fp, xref
   'MAIN @EXECUTE  BYE ;                \ And run the specified entry point

{ --------------------------------------------------------------------
Loader entry

ESP is passed in on launch.  Move it to EBP for Forth data stack and
allocate the return stack below it.  Inial ESP points to args list, so
we save it in 'ARGS for arg access.  The loader passes the SwiftForth
ORIGIN address in EDI.
-------------------------------------------------------------------- }

LABEL POWER-UP
   ESP EBP MOV   0 # EBX MOV            \ EBP --> data stack
   EBP 'ARGS [EDI] MOV                  \ Save pointer to arg list
   $100 # EBP SUB   $8100 # ESP SUB     \ Move down to make room
   ' START JMP   END-CODE               \ Transfer to START
