{ ====================================================================
Math operators

Copyright 2001  FORTH, Inc.
Author: Rick VanNorman

Revised 2020 with contributions from Dr. Heinrich Hohl

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Multiply and divide

Multiply and divide are very simple, and make no checks for validity
of data.  The actual divide code implemented here uses SM/REM for its
primitive. The floored primitive is also available.

U*/  is an unsigned  */  with the twist of rounding up.  It adds
one less than the divisor to the dividend before dividing.
-------------------------------------------------------------------- }

ICODE UM* ( u1 u2 -- ud )
   0 [EBP] EAX MOV   EBX MUL
   EAX 0 [EBP] MOV   EDX EBX MOV
   RET   END-CODE

ICODE M* ( n1 n2 -- d )
   0 [EBP] EAX MOV   EBX IMUL
   EAX 0 [EBP] MOV   EDX EBX MOV
   RET   END-CODE

ICODE * ( n|u1 n|u2 -- n|u3 )
   0 [EBP] EAX MOV   EBX MUL
   EAX EBX MOV   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE M/ ( d n -- n )
   0 [EBP] EDX MOV   CELL [EBP] EAX MOV
   EBX IDIV   EAX EBX MOV
   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE UM/MOD ( ud u -- ur uq )
   0 [EBP] EDX MOV   CELL [EBP] EAX MOV
   EBX DIV   CELL [EBP] EBP LEA
   EDX 0 [EBP] MOV   EAX EBX MOV
   RET   END-CODE

ICODE /MOD ( n1 n2 -- r q )
   0 [EBP] EAX MOV   CDQ   EBX IDIV
   EAX EBX MOV   EDX 0 [EBP] MOV
   RET   END-CODE

ICODE / ( n1 n2 -- n3 )
   0 [EBP] EAX MOV   CDQ   EBX IDIV
   CELL [EBP] EBP LEA   EAX EBX MOV
   RET   END-CODE

ICODE MOD ( n1 n2 -- n3 )
   0 [EBP] EAX MOV   CDQ   EBX IDIV
   CELL [EBP] EBP LEA   EDX EBX MOV
   RET   END-CODE

ICODE */MOD ( n1 n2 n3 -- r q )
   0 [EBP] ECX MOV   CELL [EBP] EAX MOV
   ECX IMUL   CELL [EBP] EBP LEA
   EBX IDIV   EDX 0 [EBP] MOV   EAX EBX MOV
   RET   END-CODE

ICODE */ ( n n n -- n )
   0 [EBP] ECX MOV   CELL [EBP] EAX MOV
   ECX IMUL   2 CELLS [EBP] EBP LEA
   EBX IDIV   EAX EBX MOV
   RET   END-CODE

CODE U*/ ( u1 u2 u3 - u)
   0 [EBP] EDX MOV   CELL [EBP] EAX MOV
   EDX MUL   EBX DEC   EBX EAX ADD
   0 # EDX ADC   EBX INC   EBX DIV
   EAX EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE SM/REM ( d n -- r q )
   0 [EBP] EDX MOV   CELL [EBP] EAX MOV
   EBX IDIV   CELL [EBP] EBP LEA
   EDX 0 [EBP] MOV   EAX EBX MOV
   RET   END-CODE

ICODE FM/MOD ( d n -- r q)
   0 [EBP] EDX MOV   CELL [EBP] EAX MOV
   EBX ECX MOV   EDX ECX XOR   EBX IDIV
   ECX ECX OR   0< IF
      EDX EDX OR   0= NOT IF
         EBX EDX ADD   EAX DEC
   THEN THEN
   CELL [EBP] EBP LEA
   EDX 0 [EBP] MOV   EAX EBX MOV
   RET   END-CODE

{ ------------------------------------------------------------------------
Triple Numbers

TNEGATE returns the twos compement of the triple number.

T* is a signed double by single multiply to triple result.

TU/ is a signed triple by unsigned single symmetric divide to double result.

M*/ multiplies d by n to triple result; divide by +n2 to double result.
------------------------------------------------------------------------ }

: TNEGATE ( t -- t2 )
   >R  2DUP OR DUP IF  DROP  DNEGATE 1  THEN
   R> + NEGATE ;

: T* ( d n -- t)
   2DUP XOR >R  ABS >R  DABS  R>
   2>R  R@ UM*  0  2R> UM*  D+ ( ie UT*)
   R> 0< IF  TNEGATE  THEN ;

: TU/ ( t u -- d)
   OVER >R >R  DUP 0< IF  TNEGATE  THEN
   R@ UM/MOD  -ROT  R> UM/MOD  NIP SWAP ( ie TU/)
   R> 0< IF  DNEGATE  THEN ;

: M*/ ( d n +n2 -- d2)
   >R T*  R> TU/ ;

{ --------------------------------------------------------------------
Stack triplets
-------------------------------------------------------------------- }

ICODE 3DROP ( x1 x2 x3 -- )
   2 CELLS [EBP] EBX MOV   3 CELLS [EBP] EBP LEA
   RET   END-CODE

ICODE 3DUP ( x1 x2 x3 -- x1 x2 x3 x1 x2 x3 )
   -3 CELLS [EBP] EBP LEA   EBX 2 CELLS [EBP] MOV
   3 CELLS [EBP] EAX MOV   EAX 0 [EBP] MOV
   4 CELLS [EBP] EAX MOV   EAX CELL [EBP] MOV
   RET   END-CODE
