{ ====================================================================
Input number conversion

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

{ --------------------------------------------------------------------
Vectored literal compiler

LASTLIT has the last two literals; kept for the optimizer.
LITERAL compiles code to push a literal onto the data stack.  It
   has _intimate_ knowledge of the routine (LITERAL) .
2LITERAL compiles two sequential literals comprising a double number.
-------------------------------------------------------------------- }

2VARIABLE LASTLIT

: LITERAL ( n -- )   LASTLIT @ OVER LASTLIT 2!
   ['] (LITERAL) COMPILE, -4 ALLOT ,  ;  IMMEDIATE

: 2LITERAL ( d -- )
   SWAP  POSTPONE LITERAL  POSTPONE LITERAL ; IMMEDIATE

{ --------------------------------------------------------------------
The rules governing double numeric input are discussed at length in
Forth Programmer's Handbook and other documentation.

>NUMBER supplies the basic ANS text to number conversion primitive.

CONVERT supplies the old ANS number conversion primitive. This word is
obsolescent and is included in support of legacy applications. Its
function is superseded by >NUMBER.

NUMBER? takes a string address and length and attempts to convert it
to a number.  Returns 0 to indicate that the conversion failed, a single
number and 1, or a double number (if it contains puctuation characters)
and 2.  Also leaves DPL and NH set as described above.

ANY-NUMBER? takes a string address and length and converts it to a
number using NUMBER? possibly using the first character of the string
to indicate a temporary input base override.

-------------------------------------------------------------------- }

: >NUMBER ( ud1 c-addr1 u1 -- ud2 c-addr2 u2)
   DUP IF  BEGIN  1 DPL +!  OVER C@ UPPER
      [CHAR] 0 - DUP 10 17 WITHIN OR DUP
      16 > 7 AND -  DUP 0 BASE @ WITHIN WHILE
      SWAP >R  2SWAP  BASE @ *  SWAP BASE @ UM* D+ ROT 1+
   R> 1-  DUP 0= UNTIL  EXIT  THEN DROP  THEN ;

: CONVERT ( ud1 c-addr1 -- ud2 c-addr2 )
   1+ -1 >NUMBER DROP ;

: NUMBER? ( c-addr u -- d 2 | n 1 | 0)
   -1024 DPL !  OVER C@ [CHAR] - =  DUP NH ! IF  1 /STRING  THEN
   DUP 1 < IF  2DROP 0  EXIT  THEN  0 0 2SWAP
   BEGIN  >NUMBER  DUP WHILE
      OVER C@  DUP [CHAR] + [CHAR] 0 WITHIN
      SWAP [CHAR] : = OR WHILE  1 /STRING  0 DPL !  REPEAT
   2DROP 2DROP 0  EXIT  THEN
   2DROP  NH @ IF DNEGATE THEN  2
   DPL @ 0< IF  SWAP NH ! 1-  THEN ;

| : TEMP-BASE ( char -- 0|1 )   1 SWAP
   DUP [CHAR] $ = IF  DROP  HEX     EXIT THEN
   DUP [CHAR] # = IF  DROP  DECIMAL EXIT THEN
   DUP [CHAR] & = IF  DROP  OCTAL   EXIT THEN
   DUP [CHAR] % = IF  DROP  BINARY  EXIT THEN
                      DROP  1- ( 0) ;

: ANY-NUMBER? ( c-addr u -- d 2 | n 1 | 0)
   BASE @ >R  OVER C@ TEMP-BASE /STRING  NUMBER?  R> BASE ! ;
