{ ====================================================================
User I/O interface

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Output

TYPES is a slower version of TYPE uses EMIT (which interprets control
characters) to output the string. TYPE does not interpret control
characters, but is faster.

SPACE outputs a single space.
SPACES outputs multiple spaces.
-------------------------------------------------------------------- }

: TYPES ( a n -- )
   0 ?DO COUNT EMIT LOOP DROP ;

: SPACE  ( -- )   BL EMIT  ;

: SPACES ( +n -- )
   BEGIN   DUP 0> WHILE
      S"         " TUCK FOURTH MIN TYPE -
   REPEAT DROP ;

{ ------------------------------------------------------------------------
Numeric conversion bases

HEX DECIMAL OCTAL BINARY control the numeric conversion base.
------------------------------------------------------------------------ }

: HEX     ( -- )   $10 BASE ! ;
: DECIMAL ( -- )   $0A BASE ! ;
: OCTAL   ( -- )   $08 BASE ! ;
: BINARY  ( -- )   $02 BASE ! ;

{ ------------------------------------------------------------------------
Low-level Pictured Numeric Output

<#    initializes output below the user's PAD .
#>    finishes conversion, leaving an address and count.
HOLD  inserts one character into the pictured output.
HOLDS inserts a string into the pictured output.
DIGIT converts a single digit of a number into ascii.
(#)    converts a single digit of a double number to a character.
#     converts one digit of the double number to ascii based on
        the current base.
#S    converts the double number one digit at a time until the
        double number is reduced to zero.
SIGN  tests the top of stack and inserts a minus sign if it is negative.
------------------------------------------------------------------------ }

: <# ( -- )
   PAD HLD ! ;

: #> ( xd -- c-addr u )
   2DROP HLD @ PAD OVER - ;

: HOLD ( c -- )
   HLD @ HERE = -17 ?THROW
   -1 HLD +!  HLD @ C! ;

: HOLDS ( addr u -- )
   BEGIN  DUP WHILE  1- 2DUP + C@ HOLD  REPEAT 2DROP ;

: DIGIT ( n -- n )
   DUP 9 > IF 7 + THEN [CHAR] 0 + ;

: (#) ( ud1 -- ud2 char )
   BASE @ >R  0 R@ UM/MOD  R> SWAP >R  UM/MOD  R>  ROT DIGIT ;

: # ( ud1 -- ud2 )   (#) HOLD ;

: #S ( ud1 -- ud2 )
   BEGIN # 2DUP OR 0= UNTIL ;

: SIGN ( n -- )
   0< IF [CHAR] - HOLD THEN ;

{ --------------------------------------------------------------------
(D.) (DU.) (.) Z(.) (H.0) convert various numbers into
   printable strings.
-------------------------------------------------------------------- }

: (D.) ( d -- addr # )
   SWAP OVER DUP 0< IF DNEGATE THEN
   <#  #S ROT SIGN  #> ;

: (DU.) ( d -- addr # )
   <# #S #> ;

: (.) ( n -- addr n )
   DUP 0< (D.) ;

: Z(.) ( n -- zaddr )
   DUP 0< (D.)  OVER +  0 SWAP C! ;

: (.0) ( u n -- addr n )
   <# 0 TUCK ?DO # LOOP #>  ;

: (B.0) ( n f -- addr len )
   BASE @ >R BINARY (.0) R> BASE ! ;

: (H.0) ( u n -- addr n )
   BASE @ >R HEX  (.0)  R> BASE ! ;

: (H.8) ( u -- addr n )   8 (H.0) ;

: (U.R) ( u n -- addr len )
   >R 0 (D.) R@ OVER - 0 MAX 0 ?DO BL HOLD LOOP #> R> MIN ;

{ ------------------------------------------------------------------------
Formatted Numeric Output

D.  Display d in free field format; followed by space.
DU. Display du in free field format; followed by space.
D.R Display d right-justified in a field of width n.
U.  Display u in free field format; followed by space.
U.R Display u right-justified in a field of width n.
.   Display n in free field format.
.R  Display n right-justified in a field of width n2.
?   Display memory contents.
------------------------------------------------------------------------ }

: D. ( d -- )   (D.) TYPE SPACE ;
: DU. ( d -- )   (DU.) TYPE SPACE ;

: D.R ( d n -- )
   >R  (D.)  R> OVER - 0 MAX SPACES  TYPE ;

: U. ( u -- )   0 D. ;
: U.R ( u n -- )   0 SWAP D.R ;

: . ( n -- )   DUP 0< D. ;
: .R ( n n2 -- )   SWAP DUP 0< ROT D.R ;

: ? ( a -- )    @ . ;

{ --------------------------------------------------------------------
H.  Display u in hex in free field format; followed by space.
H.R Display u in hex in a right-justified field of width n.
H.0 Display u in hex in a right-justified field of n zeros.
-------------------------------------------------------------------- }

: H. ( u -- )   BASE @ HEX SWAP U. BASE ! ;
: H.R ( u n -- )   BASE @ -ROT HEX U.R  BASE ! ;
: H.0 ( u n -- )   (H.0) TYPE ;
: H.8 ( u -- )   (H.8) TYPE ;

: B.0 ( n f -- )   (B.0) TYPE SPACE ;

{ --------------------------------------------------------------------
.S Displays the stack contents non-destructively.
H.S displays the stack in hex.
-------------------------------------------------------------------- }

: .S ( ? -- ? )
   CR DEPTH 0> IF DEPTH 0 DO S0 @ I 1+ CELLS - @ . LOOP THEN ." <-Top "
   DEPTH 0< ABORT" Underflow" ;

: H.S ( ? -- ? )
   CR DEPTH 0 MAX 0 ?DO
      S0 @ I 1+ CELLS - @ H.8 SPACE
   LOOP ." <-Top " ;
