{ ====================================================================
Optimizer patterns

Copyright 2001  FORTH, Inc.
==================================================================== }

HOST

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

   '(T) (VALUE)    TO '(VALUE)
   '(T) (USER)     TO '(USER)
   '(T) (CREATE)   TO '(CREATE)
   '(T) (CONSTANT) TO '(CONSTANT)

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: NIP-LITERAL ( -- )   [+ASSEMBLER]
      EBX 0 [EBP] MOV
      LASTLIT @ # EBX MOV
   [PREVIOUS] ;

OPTIMIZE NIP (LITERAL) WITH NIP-LITERAL

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-COMPARE ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX CMP   0 # EBX MOV
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE (LITERAL) <     WITH LIT-COMPARE   ASSEMBLE JGE
OPTIMIZE (LITERAL) >     WITH LIT-COMPARE   ASSEMBLE JLE
OPTIMIZE (LITERAL) <=    WITH LIT-COMPARE   ASSEMBLE JG
OPTIMIZE (LITERAL) >=    WITH LIT-COMPARE   ASSEMBLE JL
OPTIMIZE (LITERAL) =     WITH LIT-COMPARE   ASSEMBLE JNZ
OPTIMIZE (LITERAL) <>    WITH LIT-COMPARE   ASSEMBLE JZ
OPTIMIZE (LITERAL) U<    WITH LIT-COMPARE   ASSEMBLE JAE
OPTIMIZE (LITERAL) U>    WITH LIT-COMPARE   ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R@-LIT-COMP ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTLIT @ # 0 [ESP] CMP
   0 # EBX MOV   HERE(T) 3 + 1 RULEX
   EBX DEC   [PREVIOUS] ;

OPTIMIZE R@ LIT-COMPARE  WITH R@-LIT-COMP

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R@-LIT-COMP-IF ( -- )
   [+ASSEMBLER]
   LASTLIT @ # 0 [ESP] CMP
   HERE(T) $400 + 2 RULEX
   [PREVIOUS] ;

OPTIMIZE R@-LIT-COMP (IF) WITH R@-LIT-COMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: STACK-COMP ( n -- )
   [+ASSEMBLER]
   ( n) [EBP] EBX CMP   0 # EBX MOV
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

: OVER-COMP 0 STACK-COMP ;
: THIRD-COMP  4 STACK-COMP ;

OPTIMIZE OVER <      WITH OVER-COMP  ASSEMBLE JGE
OPTIMIZE OVER >      WITH OVER-COMP  ASSEMBLE JLE
OPTIMIZE OVER <=     WITH OVER-COMP  ASSEMBLE JG
OPTIMIZE OVER >=     WITH OVER-COMP  ASSEMBLE JL
OPTIMIZE OVER =      WITH OVER-COMP  ASSEMBLE JNZ
OPTIMIZE OVER <>     WITH OVER-COMP  ASSEMBLE JZ
OPTIMIZE OVER U<     WITH OVER-COMP  ASSEMBLE JAE
OPTIMIZE OVER U>     WITH OVER-COMP  ASSEMBLE JBE

OPTIMIZE THIRD <     WITH THIRD-COMP  ASSEMBLE JGE
OPTIMIZE THIRD >     WITH THIRD-COMP  ASSEMBLE JLE
OPTIMIZE THIRD <=    WITH THIRD-COMP  ASSEMBLE JG
OPTIMIZE THIRD >=    WITH THIRD-COMP  ASSEMBLE JL
OPTIMIZE THIRD =     WITH THIRD-COMP  ASSEMBLE JNZ
OPTIMIZE THIRD <>    WITH THIRD-COMP  ASSEMBLE JZ
OPTIMIZE THIRD U<    WITH THIRD-COMP  ASSEMBLE JAE
OPTIMIZE THIRD U>    WITH THIRD-COMP  ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: OVER-COMP-IF ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX CMP   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE OVER-COMP (IF) WITH OVER-COMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: SWAP-COMP ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX CMP   CELL [EBP] EBP LEA
   0 # EBX MOV   HERE(T) 3 + 0 RULEX
   EBX DEC   [PREVIOUS] ;

OPTIMIZE SWAP <      WITH SWAP-COMP  ASSEMBLE JGE
OPTIMIZE SWAP >      WITH SWAP-COMP  ASSEMBLE JLE
OPTIMIZE SWAP <=     WITH SWAP-COMP  ASSEMBLE JG
OPTIMIZE SWAP >=     WITH SWAP-COMP  ASSEMBLE JL
OPTIMIZE SWAP =      WITH SWAP-COMP  ASSEMBLE JNZ
OPTIMIZE SWAP <>     WITH SWAP-COMP  ASSEMBLE JZ
OPTIMIZE SWAP U<     WITH SWAP-COMP  ASSEMBLE JAE
OPTIMIZE SWAP U>     WITH SWAP-COMP  ASSEMBLE JBE

: SWAP-COMP-IF ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX CMP   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE SWAP-COMP (IF) WITH SWAP-COMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-ALU ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE (LITERAL) -    WITH LIT-ALU  ASSEMBLE SUB
OPTIMIZE (LITERAL) AND  WITH LIT-ALU  ASSEMBLE AND
OPTIMIZE (LITERAL) OR   WITH LIT-ALU  ASSEMBLE OR
OPTIMIZE (LITERAL) XOR  WITH LIT-ALU  ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-PLUS ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX ADD
   [PREVIOUS] ;

OPTIMIZE (LITERAL) +    WITH LIT-PLUS

: LIT+LIT+ ( -- )
   [+ASSEMBLER]
   LASTLIT 2@ + LASTLIT !   LASTLIT @ # EBX ADD
   [PREVIOUS]  ['] LIT-PLUS XTHIST ! ;

OPTIMIZE LIT-PLUS LIT-PLUS WITH LIT+LIT+

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R>-ALU ( -- )
   [+ASSEMBLER]
   EAX POP   EAX EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE R> +   WITH R>-ALU ASSEMBLE ADD
OPTIMIZE R> -   WITH R>-ALU ASSEMBLE SUB
OPTIMIZE R> AND WITH R>-ALU ASSEMBLE AND
OPTIMIZE R> OR  WITH R>-ALU ASSEMBLE OR
OPTIMIZE R> XOR WITH R>-ALU ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R@-ALU ( -- )
   [+ASSEMBLER]
   0 [ESP] EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE R@ +   WITH R@-ALU ASSEMBLE ADD
OPTIMIZE R@ -   WITH R@-ALU ASSEMBLE SUB
OPTIMIZE R@ AND WITH R@-ALU ASSEMBLE AND
OPTIMIZE R@ OR  WITH R@-ALU ASSEMBLE OR
OPTIMIZE R@ XOR WITH R@-ALU ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: ALU-DUP ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX 0 RULEX   EBX 0 [EBP] MOV
   [PREVIOUS] ;

OPTIMIZE +   DUP WITH ALU-DUP ASSEMBLE ADD
OPTIMIZE AND DUP WITH ALU-DUP ASSEMBLE AND
OPTIMIZE OR  DUP WITH ALU-DUP ASSEMBLE OR
OPTIMIZE XOR DUP WITH ALU-DUP ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: STACK-ALU ( n -- )
   [+ASSEMBLER]
   ( n) [EBP] EBX 0 RULEX
   [PREVIOUS] ;

: OVER-ALU   0 STACK-ALU ;
: THIRD-ALU   4 STACK-ALU ;


OPTIMIZE OVER +    WITH  OVER-ALU ASSEMBLE ADD
OPTIMIZE OVER -    WITH  OVER-ALU ASSEMBLE SUB
OPTIMIZE OVER AND  WITH  OVER-ALU ASSEMBLE AND
OPTIMIZE OVER OR   WITH  OVER-ALU ASSEMBLE OR
OPTIMIZE OVER XOR  WITH  OVER-ALU ASSEMBLE XOR

OPTIMIZE THIRD +   WITH  THIRD-ALU ASSEMBLE ADD
OPTIMIZE THIRD -   WITH  THIRD-ALU ASSEMBLE SUB
OPTIMIZE THIRD AND WITH  THIRD-ALU ASSEMBLE AND
OPTIMIZE THIRD OR  WITH  THIRD-ALU ASSEMBLE OR
OPTIMIZE THIRD XOR WITH  THIRD-ALU ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: ALU->R ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX 0 RULEX   EBX PUSH
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   [PREVIOUS] ;

OPTIMIZE +   >R  WITH  ALU->R ASSEMBLE ADD
OPTIMIZE AND >R  WITH  ALU->R ASSEMBLE AND
OPTIMIZE OR  >R  WITH  ALU->R ASSEMBLE OR
OPTIMIZE XOR >R  WITH  ALU->R ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R@-COMP ( -- )
   [+ASSEMBLER]
   0 [ESP] EBX CMP   0 # EBX MOV
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE R@ <      WITH R@-COMP  ASSEMBLE JGE
OPTIMIZE R@ >      WITH R@-COMP  ASSEMBLE JLE
OPTIMIZE R@ <=     WITH R@-COMP  ASSEMBLE JG
OPTIMIZE R@ >=     WITH R@-COMP  ASSEMBLE JL
OPTIMIZE R@ =      WITH R@-COMP  ASSEMBLE JNZ
OPTIMIZE R@ <>     WITH R@-COMP  ASSEMBLE JZ
OPTIMIZE R@ U<     WITH R@-COMP  ASSEMBLE JAE
OPTIMIZE R@ U>     WITH R@-COMP  ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R>-COMP ( -- )
   [+ASSEMBLER]
   EAX POP   EAX EBX CMP   0 # EBX MOV
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE R> <      WITH R>-COMP  ASSEMBLE JGE
OPTIMIZE R> >      WITH R>-COMP  ASSEMBLE JLE
OPTIMIZE R> <=     WITH R>-COMP  ASSEMBLE JG
OPTIMIZE R> >=     WITH R>-COMP  ASSEMBLE JL
OPTIMIZE R> =      WITH R>-COMP  ASSEMBLE JNZ
OPTIMIZE R> <>     WITH R>-COMP  ASSEMBLE JZ
OPTIMIZE R> U<     WITH R>-COMP  ASSEMBLE JAE
OPTIMIZE R> U>     WITH R>-COMP  ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: R@-CMP-IF ( -- )
   [+ASSEMBLER]
   0 [ESP] EBX CMP   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE R@-COMP (IF) WITH R@-CMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: #-CMP-IF ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX CMP   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE LIT-COMPARE (IF) WITH #-CMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: OVER-C@ ( -- )
   [+ASSEMBLER]
   -CELL [EBP] EBP LEA   EBX 0 [EBP] MOV
   CELL [EBP] EAX MOV   EBX EBX XOR
   0 [EAX] BL MOV
   [PREVIOUS] ;

OPTIMIZE OVER C@ WITH OVER-C@

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: OVER-C@-#-CMP-IF ( -- )
   [+ASSEMBLER]
   0 [EBP] EAX MOV   0 [EAX] AL MOV
   LASTLIT @ # AL CMP   HERE(T) $400 + 2 RULEX
   [PREVIOUS] ;

OPTIMIZE OVER-C@ #-CMP-IF WITH OVER-C@-#-CMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: DUP-#-CMP-IF ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX CMP   HERE(T) $400 + 2 RULEX
   [PREVIOUS] ;

OPTIMIZE DUP #-CMP-IF WITH DUP-#-CMP-IF

: OVER-#-CMP-IF ( -- )
   [+ASSEMBLER]
   LASTLIT @ # 0 [EBP] CMP   HERE(T) $400 + 2 RULEX
   [PREVIOUS] ;

OPTIMIZE OVER #-CMP-IF WITH OVER-#-CMP-IF

: THIRD-#-CMP-IF ( -- )
   [+ASSEMBLER]
   LASTLIT @ # CELL [EBP] CMP
   HERE(T) $400 + 2 RULEX
   [PREVIOUS] ;

OPTIMIZE THIRD #-CMP-IF WITH THIRD-#-CMP-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: ALU-IF ( -- )
   [+ASSEMBLER]
   0 [EBP] EBX 0 RULEX   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   HERE(T) $400 + JZ
   [PREVIOUS] ;

OPTIMIZE +   (IF) WITH ALU-IF  ASSEMBLE ADD
OPTIMIZE -   (IF) WITH ALU-IF  ASSEMBLE SUB
OPTIMIZE AND (IF) WITH ALU-IF  ASSEMBLE AND
OPTIMIZE OR  (IF) WITH ALU-IF  ASSEMBLE OR
OPTIMIZE XOR (IF) WITH ALU-IF  ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: STACK-ALU-IF ( n -- )
   [+ASSEMBLER]
   ( n) [EBP] EBX 1 RULEX   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE(T) $400 + JZ
   [PREVIOUS] ;

: OVER-ALU-IF   0 STACK-ALU-IF ;
: THIRD-ALU-IF   4 STACK-ALU-IF ;

OPTIMIZE OVER  ALU-IF WITH OVER-ALU-IF
OPTIMIZE THIRD ALU-IF WITH THIRD-ALU-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: DUP-UNARY ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   EBX EBX OR   0 # EBX MOV
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE DUP 0=  WITH DUP-UNARY ASSEMBLE JNZ
OPTIMIZE DUP 0<> WITH DUP-UNARY ASSEMBLE JZ
OPTIMIZE DUP 0<  WITH DUP-UNARY ASSEMBLE JNS
OPTIMIZE DUP 0>  WITH DUP-UNARY ASSEMBLE JLE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: OVER-UNARY ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   EBX EBX XOR   0 # CELL [EBP] CMP
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE OVER 0=  WITH OVER-UNARY ASSEMBLE JNZ
OPTIMIZE OVER 0<> WITH OVER-UNARY ASSEMBLE JZ
OPTIMIZE OVER 0<  WITH OVER-UNARY ASSEMBLE JNS
OPTIMIZE OVER 0>  WITH OVER-UNARY ASSEMBLE JLE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: THIRD-UNARY ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   EBX EBX XOR
   0 # 2 CELLS [EBP] CMP
   HERE(T) 3 + 0 RULEX   EBX DEC
   [PREVIOUS] ;

OPTIMIZE THIRD 0=  WITH THIRD-UNARY ASSEMBLE JNZ
OPTIMIZE THIRD 0<> WITH THIRD-UNARY ASSEMBLE JZ
OPTIMIZE THIRD 0<  WITH THIRD-UNARY ASSEMBLE JNS
OPTIMIZE THIRD 0>  WITH THIRD-UNARY ASSEMBLE JLE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: UNARY-IF ( -- )
   [+ASSEMBLER]
   EBX EBX OR   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE(T) $400 + 0 RULEX
   [PREVIOUS] ;

OPTIMIZE 0=  (IF) WITH UNARY-IF ASSEMBLE JNZ
OPTIMIZE 0<> (IF) WITH UNARY-IF ASSEMBLE JZ
OPTIMIZE 0<  (IF) WITH UNARY-IF ASSEMBLE JNS
OPTIMIZE 0>  (IF) WITH UNARY-IF ASSEMBLE JLE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: CMP-IF ( -- )
   [+ASSEMBLER]
   EBX 0 [EBP] CMP   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA   HERE(T) $400 + 0 RULEX
   [PREVIOUS] ;

OPTIMIZE <  (IF)    WITH CMP-IF   ASSEMBLE JGE
OPTIMIZE >  (IF)    WITH CMP-IF   ASSEMBLE JLE
OPTIMIZE <= (IF)    WITH CMP-IF   ASSEMBLE JG
OPTIMIZE >= (IF)    WITH CMP-IF   ASSEMBLE JL
OPTIMIZE =  (IF)    WITH CMP-IF   ASSEMBLE JNZ
OPTIMIZE <> (IF)    WITH CMP-IF   ASSEMBLE JZ
OPTIMIZE U< (IF)    WITH CMP-IF   ASSEMBLE JAE
OPTIMIZE U> (IF)    WITH CMP-IF   ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: CMP-?EXIT ( -- )
   [+ASSEMBLER]
   EBX 0 [EBP] CMP   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA
   HERE(T) 3 + 0 RULEX   RET
   [PREVIOUS] ;

OPTIMIZE <  ?EXIT    WITH CMP-?EXIT   ASSEMBLE JGE
OPTIMIZE >  ?EXIT    WITH CMP-?EXIT   ASSEMBLE JLE
OPTIMIZE <= ?EXIT    WITH CMP-?EXIT   ASSEMBLE JG
OPTIMIZE >= ?EXIT    WITH CMP-?EXIT   ASSEMBLE JL
OPTIMIZE =  ?EXIT    WITH CMP-?EXIT   ASSEMBLE JNZ
OPTIMIZE <> ?EXIT    WITH CMP-?EXIT   ASSEMBLE JZ
OPTIMIZE U< ?EXIT    WITH CMP-?EXIT   ASSEMBLE JAE
OPTIMIZE U> ?EXIT    WITH CMP-?EXIT   ASSEMBLE JBE

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: DUP-UNARY-IF ( -- )
   [+ASSEMBLER]
   EBX EBX OR   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE DUP-UNARY (IF) WITH DUP-UNARY-IF

: OVER-UNARY-IF ( -- )
   [+ASSEMBLER]
   0 # 0 [EBP] CMP   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE OVER-UNARY (IF) WITH OVER-UNARY-IF

: THIRD-UNARY-IF ( -- )
   [+ASSEMBLER]
   0 # CELL [EBP] CMP   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE THIRD-UNARY (IF) WITH THIRD-UNARY-IF

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: PLUS-! ( -- )
   [+ASSEMBLER]
   0 [EBP] EAX MOV   CELL [EBP] ECX MOV
   ECX 0 [EBX] [EAX] MOV   2 CELLS [EBP] EBX MOV
   3 CELLS [EBP] EBP LEA
   [PREVIOUS] ;

OPTIMIZE + ! WITH PLUS-!

: LIT-ALU-! ( -- )
   [+ASSEMBLER]
   0 [EBP] EAX MOV   LASTLIT @ # EBX 1 RULEX
   EAX 0 [EBX] MOV   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA
   [PREVIOUS] ;

OPTIMIZE LIT-ALU ! WITH LIT-ALU-!

: OVER-LIT-ALU-! ( -- )
   [+ASSEMBLER]
   CELL [EBP] EAX MOV   LASTLIT @ # EAX 2 RULEX
   EBX 0 [EAX] MOV   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE OVER LIT-ALU-! WITH OVER-LIT-ALU-!

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-PICK ( -- )
   [+ASSEMBLER]
   PUSH(EBX)
   LASTLIT @ CELLS [EBP] EBX MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) PICK WITH LIT-PICK

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: DROP-LIT ( -- )
   [+ASSEMBLER]
   LASTLIT @ # EBX MOV
   [PREVIOUS] ;

OPTIMIZE DROP (LITERAL) WITH DROP-LIT

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT->R ( -- )
   [+ASSEMBLER]
   LASTLIT @ # PUSH
   [PREVIOUS] ;

OPTIMIZE (LITERAL) >R WITH LIT->R

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: >R-LIT ( -- )
   [+ASSEMBLER]
   EBX PUSH   LASTLIT @ # EBX MOV
   [PREVIOUS] ;

OPTIMIZE >R (LITERAL) WITH >R-LIT

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-CELLS ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTLIT @ CELLS # EBX MOV
   LASTLIT @ CELLS LASTLIT !
   [PREVIOUS]
   ['] (LITERAL) XTHIST ! ;

OPTIMIZE (LITERAL) CELLS WITH LIT-CELLS

: LIT-CELLS+ ( -- )
   [+ASSEMBLER]
   LASTLIT @ CELLS # EBX ADD
   [PREVIOUS] ;

OPTIMIZE LIT-CELLS + WITH LIT-CELLS+

: CELLS+@ ( -- )
   [+ASSEMBLER]
   LASTLIT @ CELLS [EBX] EBX MOV
   [PREVIOUS] ;

OPTIMIZE LIT-CELLS+ @ WITH CELLS+@

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-SWAP ( -- )
   [+ASSEMBLER]
   -CELL [EBP] EBP LEA   LASTLIT @ # 0 [EBP] MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) SWAP WITH LIT-SWAP

: LIT-SWAP-! ( -- )
   [+ASSEMBLER]
   LASTLIT @ # 0 [EBX] MOV   POP(EBX)
   [PREVIOUS] ;

: LIT-SWAP-C! ( -- )
   [+ASSEMBLER]
   LASTLIT @ # 0 [EBX] BYTE MOV
   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE LIT-SWAP !  WITH LIT-SWAP-!
OPTIMIZE LIT-SWAP C! WITH LIT-SWAP-C!

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-SHIFT ( -- )
   [+ASSEMBLER]
   EBX LASTLIT @ # 0 RULEX
   [PREVIOUS] ;

OPTIMIZE (LITERAL) LSHIFT WITH LIT-SHIFT ASSEMBLE SHL
OPTIMIZE (LITERAL) RSHIFT WITH LIT-SHIFT ASSEMBLE SHR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: I-ALU ( -- )
   [+ASSEMBLER]
   0 [ESP] EAX MOV   CELL [ESP] EAX ADD
   EAX EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE I +     WITH I-ALU ASSEMBLE ADD
OPTIMIZE I -     WITH I-ALU ASSEMBLE SUB
OPTIMIZE I AND   WITH I-ALU ASSEMBLE AND
OPTIMIZE I OR    WITH I-ALU ASSEMBLE OR
OPTIMIZE I XOR   WITH I-ALU ASSEMBLE XOR

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: TWO-LITERALS ( -- )
   [+ASSEMBLER]
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   LASTLIT CELL+ @ # 0 [EBP] MOV   LASTLIT @ # EBX MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) (LITERAL) WITH TWO-LITERALS

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: 2LIT-ALU ( -- )
   LASTLIT 2@ 0 RULEX POSTPONE LITERAL ;

OPTIMIZE TWO-LITERALS +      WITH 2LIT-ALU WITH +
OPTIMIZE TWO-LITERALS -      WITH 2LIT-ALU WITH -
OPTIMIZE TWO-LITERALS AND    WITH 2LIT-ALU WITH AND
OPTIMIZE TWO-LITERALS OR     WITH 2LIT-ALU WITH OR
OPTIMIZE TWO-LITERALS XOR    WITH 2LIT-ALU WITH XOR
OPTIMIZE TWO-LITERALS LSHIFT WITH 2LIT-ALU WITH LSHIFT
OPTIMIZE TWO-LITERALS RSHIFT WITH 2LIT-ALU WITH RSHIFT

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: LIT-DUP ( -- )
   [+ASSEMBLER]
   -2 CELLS [EBP] EBP LEA
   EBX CELL [EBP] MOV
   LASTLIT @ # EBX MOV
   EBX 0 [EBP] MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) DUP WITH LIT-DUP

{ --------------------------------------------------------------------
-------------------------------------------------------------------- }

: DOES>-ALU ( -- )
   [+ASSEMBLER]
   EAX POP
   EAX EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE (DOES>) +   WITH DOES>-ALU ASSEMBLE ADD
OPTIMIZE (DOES>) AND WITH DOES>-ALU ASSEMBLE AND
OPTIMIZE (DOES>) OR  WITH DOES>-ALU ASSEMBLE OR
OPTIMIZE (DOES>) XOR WITH DOES>-ALU ASSEMBLE XOR

{ --------------------------------------------------------------------
Constants
-------------------------------------------------------------------- }

: CONST->LITERAL ( -- )
   LASTLIT @ LASTCHILD @ 5 + @(T) LASTLIT 2!
   ['] (LITERAL) XTHIST ! ;

: NEW-CONST ( -- )
   [+ASSEMBLER]
   CONST->LITERAL
   PUSH(EBX)   LASTLIT @ # EBX MOV
   [PREVIOUS] ;

OPTIMIZE ANY (CONSTANT) WITH NEW-CONST

: LIT-CONST ( -- )
   LASTLIT @ LASTCHILD @ 5 + @(T) LASTLIT 2!
   TWO-LITERALS  ['] TWO-LITERALS XTHIST ! ;

OPTIMIZE (LITERAL) (CONSTANT) WITH LIT-CONST

{ --------------------------------------------------------------------
User variables
-------------------------------------------------------------------- }

: NEW-USER ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTCHILD @ 5 + @(T) [ESI] EBX LEA
   [PREVIOUS] ;

OPTIMIZE ANY (USER) WITH NEW-USER

: USER-@ ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTCHILD @ 5 + @(T) [ESI] EBX MOV
   [PREVIOUS] ;

OPTIMIZE NEW-USER @ WITH USER-@

: USER-! ( -- )
   [+ASSEMBLER]
   EBX LASTCHILD @ 5 + @(T) [ESI] MOV   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE NEW-USER ! WITH USER-!

: LIT-USER-! ( -- )
   [+ASSEMBLER]
   LASTLIT @ # LASTCHILD @ 5 + @(T) [ESI] MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) USER-! WITH LIT-USER-!

: DUP-USER-! ( -- )
   [+ASSEMBLER]
   EBX LASTCHILD @ 5 + @(T) [ESI] MOV
   [PREVIOUS] ;

OPTIMIZE DUP USER-! WITH DUP-USER-!

: USER-@-ALU ( -- )
   [+ASSEMBLER]
   LASTCHILD @ 5 + @(T) [ESI] EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE USER-@ +   WITH USER-@-ALU ASSEMBLE ADD
OPTIMIZE USER-@ AND WITH USER-@-ALU ASSEMBLE AND
OPTIMIZE USER-@ OR  WITH USER-@-ALU ASSEMBLE OR
OPTIMIZE USER-@ XOR WITH USER-@-ALU ASSEMBLE XOR

: USER-@->R ( -- )
   [+ASSEMBLER]
   LASTCHILD @ 5 + @(T) [ESI] PUSH
   [PREVIOUS] ;

: R>-USER-! ( -- )
   [+ASSEMBLER]
   LASTCHILD @ 5 + @(T) [ESI] POP
   [PREVIOUS] ;

OPTIMIZE USER-@ >R WITH USER-@->R
OPTIMIZE R> USER-! WITH R>-USER-!

: USER-@-IF ( -- )
   [+ASSEMBLER]
   0 # LASTCHILD @ 5 + @(T) [ESI] CMP
   HERE(T) $400 + JZ
   [PREVIOUS] ;

OPTIMIZE USER-@ (IF) WITH USER-@-IF

: USER-@-UNARY-IF ( -- )
   [+ASSEMBLER]
   0 # LASTCHILD @ 5 + @(T) [ESI] CMP
   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE USER-@ UNARY-IF WITH USER-@-UNARY-IF

{ --------------------------------------------------------------------
Variables
-------------------------------------------------------------------- }

: NEW-VAR ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTCHILD CELL+ @ [EDI] EBX LEA
   [PREVIOUS] ;

OPTIMIZE ANY (CREATE) WITH NEW-VAR

: VAR-@ ( -- )
   [+ASSEMBLER]
   PUSH(EBX)   LASTCHILD CELL+ @ [EDI] EBX MOV
   [PREVIOUS] ;

OPTIMIZE NEW-VAR @ WITH VAR-@

: VAR-! ( -- )
   [+ASSEMBLER]
   EBX LASTCHILD CELL+ @ [EDI] MOV   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE NEW-VAR ! WITH VAR-!

: VAR-PLUS ( -- )
   [+ASSEMBLER]
   LASTCHILD CELL+ @ [EDI] [EBX] EBX LEA
   [PREVIOUS] ;

OPTIMIZE NEW-VAR +    WITH VAR-PLUS

: DUP-VAR-! ( -- )
   [+ASSEMBLER]
   EBX LASTCHILD CELL+ @ [EDI] MOV
   [PREVIOUS] ;

OPTIMIZE DUP VAR-! WITH DUP-VAR-!

: VAR-ON ( -- )
   [+ASSEMBLER]
   -1 # LASTCHILD CELL+ @ [EDI] MOV
   [PREVIOUS] ;

: VAR-OFF ( -- )
   [+ASSEMBLER]
   EAX EAX XOR   EAX LASTCHILD CELL+ @ [EDI] MOV
   [PREVIOUS] ;

OPTIMIZE NEW-VAR ON WITH VAR-ON
OPTIMIZE NEW-VAR OFF WITH VAR-OFF

: LIT-VAR-! ( -- )
   [+ASSEMBLER]
   LASTLIT @ # LASTCHILD CELL+ @ [EDI] MOV
   [PREVIOUS] ;

OPTIMIZE (LITERAL) VAR-! WITH LIT-VAR-!

: VAR-+! ( -- )
   [+ASSEMBLER]
   EBX LASTCHILD CELL+ @ [EDI] ADD   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE NEW-VAR +! WITH VAR-+!

: LIT-VAR-+! ( -- )
   [+ASSEMBLER]
   LASTLIT @ # LASTCHILD CELL+ @ [EDI] ADD
   [PREVIOUS] ;

OPTIMIZE (LITERAL) VAR-+! WITH LIT-VAR-+!

{ --------------------------------------------------------------------
Values
-------------------------------------------------------------------- }

OPTIMIZE ANY (VALUE) WITH VAR-@

: VAR-@-ALU ( -- )
   [+ASSEMBLER]
   LASTCHILD CELL+ @ [EDI] EBX 0 RULEX
   [PREVIOUS] ;

OPTIMIZE VAR-@ +   WITH VAR-@-ALU ASSEMBLE ADD
OPTIMIZE VAR-@ -   WITH VAR-@-ALU ASSEMBLE SUB
OPTIMIZE VAR-@ AND WITH VAR-@-ALU ASSEMBLE AND
OPTIMIZE VAR-@ OR  WITH VAR-@-ALU ASSEMBLE OR
OPTIMIZE VAR-@ XOR WITH VAR-@-ALU ASSEMBLE XOR

: VAR-@-IF ( -- )
   [+ASSEMBLER]
   0 # LASTCHILD CELL+ @ [EDI] CMP   HERE(T) $400 + JZ
   [PREVIOUS] ;

OPTIMIZE VAR-@ (IF) WITH VAR-@-IF

: VAR-@-UNARY-IF ( -- )
   [+ASSEMBLER]
   0 # LASTCHILD CELL+ @ [EDI] CMP   HERE(T) $400 + 1 RULEX
   [PREVIOUS] ;

OPTIMIZE VAR-@ UNARY-IF WITH VAR-@-UNARY-IF

{ --------------------------------------------------------------------
Loops
-------------------------------------------------------------------- }

: LIT-DO ( -- )
   [+ASSEMBLER]
   $80000000 # EBX ADD   EBX PUSH
   EBX NEG   LASTLIT @ # EBX ADD
   EBX PUSH   POP(EBX)
   [PREVIOUS] ;

OPTIMIZE (LITERAL) (DO) WITH LIT-DO

: 2LIT-DO ( -- )
   [+ASSEMBLER]
   LASTLIT CELL+ @ $80000000 + # PUSH
   LASTLIT @ LASTLIT CELL+ @ $80000000 + - # PUSH
   [PREVIOUS] ;

OPTIMIZE TWO-LITERALS (DO) WITH 2LIT-DO

: LIT+LOOP ( -- )
   [+ASSEMBLER]
   LASTLIT @ # 0 [ESP] ADD   HERE(T) $400 + JNO
   [PREVIOUS] ;

OPTIMIZE (LITERAL) (+LOOP) WITH LIT+LOOP
