{ ====================================================================
Personalities

Copyright 2001  FORTH, Inc.
Rick VanNorman

The terminal I/O personality table is defined here.
==================================================================== }

TARGET

{ --------------------------------------------------------------------
Table format

A personality consists of variable length data followed by a variable
number of vectors.

The minimum data size is 16 bytes (4 cells) but can be larger.

The minimum vector size is 3 cells, as the first three vectors are
absolutely required to be present and must not throw under any
circumstance.

Personality table layout:

DATA
   datasize                             bytes in data section
   maxvector                            number of vectors
   handle                               handle for input/output
   previous personality                 address of previous personality
VECTORS
   INVOKE    ( -- )                     open the personality
   REVOKE    ( -- )                     close the personality
   /INPUT    ( -- )                     reset the input stream
   EMIT      ( char -- )                print a character
   TYPE      ( addr len -- )            print a string
   ?TYPE     ( addr len -- )            print a string, respecting margins
   CR        ( -- )                     print crlf
   PAGE      ( -- )                     clear the output device
   ATTRIBUTE ( n -- )                   set attribute for type and emit
   KEY       ( -- char )                "raw" wait for a character
   KEY?      ( -- flag )                "raw" true if a character is waiting
   EKEY      ( -- echar )               "raw" wait for an echaracter
   EKEY?     ( -- flag )                "raw" true if an echaracter is waiting
   AKEY      ( -- char )                "cooked" key variation, used by ACCEPT
   PUSHTEXT  ( addr len -- )            push text into the input stream
   AT-XY     ( x y -- )                 set the output cursor position
   GET-XY    ( -- x y )                 current cursor position
   GET-SIZE  ( -- x y )                 size in characters of the display
   ACCEPT    ( addr u1 -- u2)           accept a line into a buffer

[PERSONALITY] defines named vectors in the personality table.  It
makes two definitions.  The first executes the vector; the second
returns the address of the vector in the table (used reassign that
vector's behavior).

(PERSONALITY) is the runtime code for the vectored behavior.  The
little code stub defined by [PERSONALITY] passes its vector offset in
register EAX.
-------------------------------------------------------------------- }

LABEL (PERSONALITY)                     \ EAX = index
   'PERSONALITY [ESI] EDX MOV           \ EDX --> personality
   'SELF [ESI] PUSH                     \ Preserve 'SELF
   EDX 'SELF [ESI] MOV                  \ Set 'SELF to personality (for object-based personality)
   CELL [EDX] ECX MOV                   \ ECX is max vector
   0 [EDX] EDX ADD                      \ EDX is base of vectors
   ECX EAX CMP  < IF
      0 [EDX] [EAX*4] EAX MOV           \ EAX = xt
      EDI EAX ADD   EAX CALL            \ Call address(xt)
      'SELF [ESI] POP
   RET  THEN
   'SELF [ESI] POP
   -1 # EBX MOV  ' THROW JMP
   END-CODE

INTERPRETER

: [PERSONALITY] ( n1 -- n2 )
   CODE  [+ASSEMBLER]  DUP # EAX MOV   (PERSONALITY) JMP   END-CODE  [PREVIOUS]
   CREATE DUP CELLS , 1+  DOES> ( -- addr )   @ 'PERSONALITY @ DUP @ + + ;

TARGET

0  [PERSONALITY]  INVOKE     'INVOKE     ( -- )
   [PERSONALITY]  REVOKE     'REVOKE     ( -- )
   [PERSONALITY]  /INPUT     '/INPUT     ( -- )
   [PERSONALITY]  EMIT       'EMIT       ( char -- )
   [PERSONALITY]  TYPE       'TYPE       ( addr len -- )
   [PERSONALITY]  ?TYPE      '?TYPE      ( addr len -- )
   [PERSONALITY]  CR         'CR         ( -- )
   [PERSONALITY]  PAGE       'PAGE       ( -- )
   [PERSONALITY]  ATTRIBUTE  'ATTRIBUTE  ( n -- )
   [PERSONALITY]  KEY        'KEY        ( -- char )
   [PERSONALITY]  KEY?       'KEY?       ( -- flag )
   [PERSONALITY]  EKEY       'EKEY       ( -- echar )
   [PERSONALITY]  EKEY?      'EKEY?      ( -- flag )
   [PERSONALITY]  AKEY       'AKEY       ( -- echar )
   [PERSONALITY]  PUSHTEXT   'PUSHTEXT   ( addr len -- )
   [PERSONALITY]  AT-XY      'AT-XY      ( x y -- )
   [PERSONALITY]  GET-XY     'GET-XY     ( -- x y )
   [PERSONALITY]  GET-SIZE   'GET-SIZE   ( -- x y )
   [PERSONALITY]  ACCEPT     'ACCEPT     ( addr u1 -- u2)
DROP

: NORMAL    ( -- )   0 ATTRIBUTE ;
: INVERSE   ( -- )   1 ATTRIBUTE ;
: BRIGHT    ( -- )   2 ATTRIBUTE ;
: BOLD      ( -- )   3 ATTRIBUTE ;
: INVISIBLE ( -- )  31 ATTRIBUTE ;

{ --------------------------------------------------------------------
Personality handles

PERSONALITY-HANDLE returns the address of the personality's handle.
PREVIOUS-PERSONALITY returns the address of the "save" location.

PHANDLE returns the personality handle.

OPEN-PERSONALITY sets up a new personality; the previous one is saved
and the new one is opened via its INVOKE vector.

CLOSE-PERSONALITY terminates the active personality via its REVOKE
vector and the previous one is restored.
-------------------------------------------------------------------- }

: PERSONALITY-HANDLE ( -- a )   'PERSONALITY @ CELL+ CELL+ ;
: PREVIOUS-PERSONALITY ( -- a )   'PERSONALITY @ 3 CELLS + ;

: PHANDLE ( -- h )   PERSONALITY-HANDLE @ ;

: OPEN-PERSONALITY ( 'p -- )
   'PERSONALITY @  SWAP 'PERSONALITY !
   PREVIOUS-PERSONALITY !  INVOKE ;

: CLOSE-PERSONALITY ( -- )
   PREVIOUS-PERSONALITY @  0 PREVIOUS-PERSONALITY !
   REVOKE  'PERSONALITY ! ;
