{ ====================================================================
Common SwiftForth startup

Copyright 2008   FORTH, Inc.

This file supplies the common SwiftForth startup code.
==================================================================== }

{ --------------------------------------------------------------------
Version information

VERSION is the actual version string. It may be over-written before an
image is saved.

/VERSION creates the entire version control string.
.VERSION displays VERSION.

SwiftForth is an alias for .VERSION and also allows package managers
to test if this is a SwiftForth system.
-------------------------------------------------------------------- }

CREATE VERSION  0 C,  256 ALLOT

| : /VERSION ( -- )   VERSION C@ IF EXIT THEN
   S" SwiftForth i386-" VERSION ZPLACE
   SYSTEM-NAME VERSION ZAPPEND
   RELEASE-VERSION ZCOUNT VERSION ZAPPEND
   RELEASE-DATE ZCOUNT VERSION ZAPPEND ;

: .VERSION ( -- )
   VERSION ZCOUNT -TRAILING TYPE ;

: SwiftForth ( -- )   .VERSION ;

{ --------------------------------------------------------------------
Load system extensions

HI loads the system extensions from hi.f.  First checks if there's a
hi.f in the current working directory.  If not, uses the default hi.f
specified by HI-PATH (from the kernel make file).
-------------------------------------------------------------------- }

: HI ( -- )
   S" hi.f" 2DUP FILE-STATUS NIP IF  2DROP
   HI-PATH +ROOT  THEN  INCLUDED ;

{ ------------------------------------------------------------------------
Non-interactive quit

N-QUIT is called in place of QUIT if we start up with no TTY.  There
is no "ok" prompt and any THROW back to here terminates the program
with the throw code as the exit status.
------------------------------------------------------------------------ }

| : (N-QUIT) ( -- )
   BEGIN  REFILL DROP  INTERPRET AGAIN ;

: N-QUIT ( -- )   BEGIN  R0 @ RP!  /INTERPRETER  ['] (N-QUIT) CATCH
      ?DUP IF  DUP CR .CATCH EXITSTATUS !  BYE  THEN
   S0 @ SP!  /NDP  AGAIN ;

{ ------------------------------------------------------------------------
Cold and warm start

/KERNEL initializes the system variables for the position independent
kernel.  It assumes that the address given is where to begin the
allocations, and returns the address following the allocations.

The map it builds from the address is

   addr user area       |user|
        tib             |tib|
        qbuf            |qbuf|
        fpstack         |fpstack| + |fpitem|

QBUF is a temporary pad for strings, fixed at 1k bytes.

/DLLS and /SYSTEM are defined in the full system; the startup needs to
know about them so they are deferred as nothings here.

/CMDLINE does the platform-specific command line processing.

EDIT-START is set to the G function (launch editor at last error
location) when the system extensions are loaded.

START-CONSOLE opens the simplest possible console for interactive use.

'MAIN is the first user accessable hook for an application.  If the
user replaces this, he is responsible for making everything except the
dlls load!
------------------------------------------------------------------------ }

| : /KERNEL ( addr1 -- addr2 )
   DUP /SREF                            \ Initialize xref before stomping initial kernel table
   DUP UP0 !REL  DUP UP!  |USER| +
   DUP DUP 'TIB 2!  |TIB| +
   DUP QBUF !  |QBUF| +
   |FPSTACK| + DUP 'N0 ! DUP 'N !  |FPITEM| + ;

XDEFER /IMPORTS   ' /KERNEL-IMPORTS ,
XDEFER /SYSTEM  0 ,
XDEFER /CMDLINE  0 ,
XDEFER EDIT-START  0 ,

| : ?ED ( -- flag )   ['] EDIT-START >BODY @ 0<> ;

| : DEFAULT-BAILOUT? ( -- )
   BEGIN  CR ." Exit? (Yes/No"  ?ED IF  ." /Edit"  THEN ." ) "
      KEY DUP EMIT  $20 OR
      DUP [CHAR] y = IF  DROP  EXITSTATUS !  BYE  THEN
      DUP [CHAR] e = ?ED AND IF  EDIT-START  DROP [CHAR] n  THEN
   [CHAR] n = UNTIL ;

XDEFER BAILOUT?  ' DEFAULT-BAILOUT? ,

| CREATE ?HI  -1 ,      \ Flag to greet "hi" the first time
| CREATE ?UPLINK 0 ,    \ Flag, true=uplink program attached via pipe

: --uplink ( -- )
   0 ?HI !  -1 ?UPLINK ! ;

: START-CONSOLE ( -- )
   /CONSOLE  /CMDLINE
   ?TTY @ ?UPLINK @ OR IF
      .VERSION  CR
      ?HI @ IF  ." hi "  0 ?HI !  THEN
   QUIT  THEN  N-QUIT ;

CREATE 'MAIN   ' START-CONSOLE ,  -1 ,  \ Entry point, flag (0=gui, -1=console)

: IS-CONSOLE? ( -- flag )   'MAIN CELL+ @ 0<> ;
