{ ====================================================================
String manipulation

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

{ --------------------------------------------------------------------
String Literal Compilers

STRING,  compiles counted string.  Note that an uncounted null is
always appended to the string.

STRING parses and compiles a string.
SLITERAL compiles a string literal.

," compiles a string from text which ends with a quote mark.

(RSTR) is used internally to move a string to the 4k buffer on the
caller's return stack.  Used only in Z, and U, below.

Z, compiles a zero-terminated string.
U, compiles a unicode zero-terminated string.
-------------------------------------------------------------------- }

: STRING, ( a n -- )   R-BUF  R@ PLACE
   R> HERE OVER C@ 1+  DUP /ALLOT  CMOVE 0 C, ;

: STRING  ( char -- )   PARSE STRING, ;

: SLITERAL ( c-addr1 u -- )   ( -- c-addr1 u )
   POSTPONE (S")  STRING, ; IMMEDIATE

: ," ( -- )   [CHAR] " STRING ;

| : (RSTR) ( addr len dest -- )
   SWAP  0 MAX 4096 MIN  TUCK !+  SWAP CMOVE ;

: Z, ( addr n -- )   4096 CELL+ R-ALLOC DUP >R (RSTR)
   R> @+ 0 ?DO  COUNT C,  LOOP DROP  0 C, ;

: U, ( addr n -- )   4096 CELL+ R-ALLOC DUP >R (RSTR)
   R> @+ 0 ?DO  COUNT C, 0 C,  LOOP DROP  0 C, 0 C, ;

{ --------------------------------------------------------------------
String buffer

We allocate and use a buffer in memory for the strings.  A constant
size is used to simplify the calculation of the correct offset, which
should properly be a _mod_ operation.  Note that a mask of $380
corresponds to a buffer size of $400 (1k) and a max string length of
$80 (128).

QBUF has the address of the string buffer.
QHEAD has the offset of the most recent string.
QPAD is the address of the current buffer segment.
>QPAD puts the string into the buffer and returns the address.
-------------------------------------------------------------------- }

VARIABLE QBUF
VARIABLE QHEAD

: QPAD ( -- addr )
   QHEAD @ $380 AND QBUF @ + ;

: >QPAD ( addr n -- qaddr )
   $80 QHEAD +!  QPAD PLACE   QPAD ;

{ ------------------------------------------------------------------------
String Parsers

All the string parsers are state smart.  If compiling, they will
compile the string and the behaviour to leave the strings parameters on
the stack at execution.  If interpreting, they will put the string into
the string buffer QPAD, leaving string parameters on the stack.

S" parses a " -delimited string and returns address and count.
C" parses a " -delimited string and returns address of the count.
Z" parses a " -delimited string and returns address of the start
   of the string (which contains a null terminator).
." compiles a " -delimited string and the behaviour to print it.

ABORT" compiles an error behaviour which, if TOS is non-zero, will
print a message and abort.
------------------------------------------------------------------------ }

: S" ( "ccc"<"> -- )
   STATE @ IF  POSTPONE (S") ,"  EXIT  THEN
   [CHAR] " PARSE  >QPAD COUNT ;  IMMEDIATE

: C" ( "ccc"<"> -- )
   STATE @ IF  POSTPONE (C") ,"  EXIT  THEN
   [CHAR] " PARSE  >QPAD ;  IMMEDIATE

: Z" ( "ccc"<"> -- )
   STATE @ IF  POSTPONE (Z") ,"  EXIT  THEN
   [CHAR] " PARSE  >QPAD   0 OVER COUNT + C!  1+ ;  IMMEDIATE

: ."  \ compiling: ( "ccc<">" -- )  executing: ( -- )
   POSTPONE S"  POSTPONE TYPE ;  IMMEDIATE

: ABORT" \ comp: ( "ccc<">" -- ) exec: ( i*x x1 -- ) ( R: j*x -- )
   POSTPONE S"  POSTPONE ?ABORT ; IMMEDIATE

{ ------------------------------------------------------------------------
Character parsers

CHAR parses the next blank delimited string and return the numeric
value of its first character.

[CHAR] is the compile-mode version of CHAR .

CTRL is the control-character version of CHAR.  The values returned are
masked into the range 0-31.

[CTRL]  is the compile-mode version of CTRL .
------------------------------------------------------------------------ }

: CHAR ( "name" -- char )   BL WORD 1+ C@ ;

: [CHAR] ( -- )   CHAR  POSTPONE LITERAL ;  IMMEDIATE

: CTRL ( "name" -- char )   CHAR $1F AND ;

: [CTRL] ( -- )   CTRL  POSTPONE LITERAL ;  IMMEDIATE
