{ ====================================================================
Structured compiler

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Structure building support

>RESOLVE   resolves a forward reference.
<RESOLVE   resolves a backward reference.
-------------------------------------------------------------------- }

| : !BELOW ( n addr -- )
   DUP ORIGIN HERE 1+ WITHIN NOT -22 ?THROW
   4 - DUP @ 1018 1020 WITHIN NOT -22 ?THROW  ! ;

: -BAL   -1 BAL +! /OPT ;
: +BAL    1 BAL +! /OPT ;
: /BAL   0 0 BAL 2!  /OPT ;

: >RESOLVE ( addr -- )   HERE OVER - SWAP !BELOW  -BAL  ;
: <RESOLVE ( addr -- )   HERE - HERE !BELOW  -BAL ;

{ --------------------------------------------------------------------
Tested branch structures

Figure: Structures supported (and not an exaustive list!) are:

IF .. THEN
IF .. ELSE .. THEN

BEGIN .. UNTIL
BEGIN .. AGAIN
BEGIN .. WHILE .. REPEAT
BEGIN .. WHILE .. WHILE .. REPEAT .. THEN
BEGIN .. WHILE .. WHILE .. REPEAT .. ELSE .. THEN
BEGIN .. WHILE .. UNTIL .. THEN
BEGIN .. WHILE .. UNTIL .. ELSE .. THEN


-------------------------------------------------------------------- }

| : (BEGIN) ( -- addr)   HERE  +BAL ;

: IF ( -- addr)   POSTPONE (IF)  (BEGIN) ;  IMMEDIATE
: ELSE ( addr1 -- addr2)   POSTPONE (ELSE)  (BEGIN)  SWAP >RESOLVE ;  IMMEDIATE
: AHEAD ( -- addr)   POSTPONE (ELSE)  (BEGIN) ;  IMMEDIATE
: THEN ( addr -- )   >RESOLVE ;  IMMEDIATE

: BEGIN ( -- addr)   (BEGIN) ;  IMMEDIATE
: AGAIN ( addr -- )   POSTPONE (ELSE)  <RESOLVE ;  IMMEDIATE
: UNTIL ( addr -- )   POSTPONE (IF)  <RESOLVE ;  IMMEDIATE

: WHILE ( addr1 -- addr2 addr1)   POSTPONE (IF)  (BEGIN) SWAP  ;  IMMEDIATE
: REPEAT ( addr2 addr1 -- )   POSTPONE (ELSE)  <RESOLVE  >RESOLVE ;  IMMEDIATE

{ --------------------------------------------------------------------
Iterated structures

DO .. LOOP
DO .. +LOOP
?DO .. LOOP
?DO .. +LOOP

Notice that ?DO is contains a forward branch which LOOP and +LOOP
must resolve.

LEAVE disposes of the loop parameters and force a branch to the code
following the nearest structurally level LOOP or +LOOP. LEAVE builds a
chain in BAL CELL+ .

>LEAVES resolves all branches from instances of LEAVE through end of
loop. The address is the top of the loop, all LEAVE tags greater than
the address are resolved.

BOUNDS converts an address and length into bounding addresses.
-------------------------------------------------------------------- }

| : -?>RESOLVE ( -- )   DUP 4 - @ SWAP
   HERE OVER - SWAP 4 - !  -BAL ;

: LEAVE ( -- )   POSTPONE UNLOOP  POSTPONE (ELSE)
   (BEGIN) BAL CELL+ 2DUP @  SWAP !BELOW  ! ;  IMMEDIATE

: >LEAVES ( addr -- )   BAL CELL+ @  BEGIN  2DUP U< WHILE
      -?>RESOLVE
   REPEAT  BAL CELL+ !  DROP ;

: DO ( -- flag addr)
   POSTPONE (DO) 0  (BEGIN) ;  IMMEDIATE

: ?DO ( -- addr1 flag addr2)
   POSTPONE (?DO) (BEGIN)
   POSTPONE (DO) 1  (BEGIN) ;  IMMEDIATE

: LOOP ( flag addr | addr1 flag addr2 -- )
   POSTPONE (LOOP)  DUP <RESOLVE POSTPONE UNLOOP >LEAVES
   IF  POSTPONE THEN  THEN ;  IMMEDIATE

: +LOOP ( flag addr | addr1 flag addr2 -- )
   POSTPONE (+LOOP)  DUP <RESOLVE  POSTPONE UNLOOP >LEAVES
   IF  POSTPONE THEN  THEN ;  IMMEDIATE

ICODE BOUNDS ( a n -- a+n a )   \ : BOUNDS OVER + SWAP ;
   0 [EBP] EBX ADD
   0 [EBP] EAX MOV
   EBX 0 [EBP] MOV
   EAX EBX MOV
   RET END-CODE

{ ------------------------------------------------------------------------
The Eaker case statement

This is the ANS version with the implied DROP in ENDCASE.

OF is the selector statement.
ENDOF ends an OF statement.
ENDCASE ends the case statement
CASE begins the case statement.
------------------------------------------------------------------------ }

: OF ( -- ofsys )
   POSTPONE (OF) (BEGIN) POSTPONE 2DROP ;  IMMEDIATE

: ENDOF ( casesys ofsys -- casesys )
   POSTPONE (ELSE)  >RESOLVE  HERE 4 - !  (BEGIN) ; IMMEDIATE

: ENDCASE ( casesys -- )
   POSTPONE DROP
   BEGIN
      ?DUP WHILE
      -?>RESOLVE
   REPEAT -BAL ;  IMMEDIATE

: CASE ( -- casesys )
   +BAL 0 ; IMMEDIATE
