{ ====================================================================
Execution Switches - classic version

Copyright 2001 FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ ------------------------------------------------------------------------
Message switch data structures

   head: |link|default-xt|

   node: |link|value|xt|


All the _link_ s are relative. All the _value_ s are integers.

SWITCH searches the linked list, looking for a _value_ which
   matches the _key_ .  If one is found, the _xt_ is executed with
   the _i*x_ parameters on the stack. Otherwise, the default is
   executed with _i*x_ and _key_ on the stack.

Figure: Switch examples

[SWITCH FOO DROP
   1 RUNS WORDS
   2 RUN: HERE 100 DUMP ;
   3 RUNS ABOUT
SWITCH]

[SWITCH FOO .
   3 RUN: ." THREE" ;
   5 RUN: ." FIVE" ;
SWITCH]

------------------------------------------------------------------------ }

{ --------------------------------------------------------------------
Run-time execution

The code switch is a simple, fast execution engine for lists of keys.
SWITCH acts on the head of the list.
-------------------------------------------------------------------- }

CODE SWITCH ( i*x switch-key switch-head -- j*x )
   CELL [EBX] PUSH                      \ save switch default xt
   0 [EBP] EAX MOV                      \ eax = n
   EBX EDX MOV                          \ edx is address in links
   BEGIN                                \ ecx is link
      0 [EDX] ECX MOV                   \ ecx = [a]
      ECX ECX OR  0<> WHILE
      ECX EDX ADD                       \ make absolute
      CELL [EDX] EAX CMP                \ check n
      0= IF                             \ match!
         CELL # ESP ADD                 \ discard switch address
         CELL [EBP] EBX MOV             \ ebx = i*x(0)
         2 CELLS [EBP] EBP LEA
         2 CELLS [EDX] EAX MOV
         EDI EAX ADD   EAX JMP          \ execute xt
   THEN  REPEAT
   CELL [EBP] EBP LEA   EAX EBX MOV
   EAX POP   EDI EAX ADD   EAX JMP      \ execute xt
   END-CODE

{ --------------------------------------------------------------------
Compile SWITCH lists

SWITCHES has the head of the global list of all switches defined.
[SWITCH defines a new switch with a default behavior.
[+SWITCH adds to an existing switch.
SWITCH] validates and ends the definition of a switch.

RUNS compiles a switch entry based on a key and an XT.
RUN: compiles a switch entry based on a key and a :NONAME phrase.
DATA: compiles a switch entry based on a key and a string.

ZERO is a common default for switches; throw away the key and
return a zero.
-------------------------------------------------------------------- }

| : ?PAIRS ( n n -- )   /OPT  <> THROW ;

VARIABLE SWITCHES

: [SWITCH ( -- switch-sys addr ) \ Usage: [SWITCH <name> <default>
   CREATE ' HERE 0 ,  SWAP ( xt) , SWITCHES >LINK  7777 SWAP
   (;CODE) R> SWITCH ;  \   DOES> ( -- n a ) SWITCH ;

: [+SWITCH ( -- switch-sys addr ) \ Usage: [+SWITCH <name>
   ' >BODY  7777 SWAP ;

: SWITCH] ( switch-sys addr -- )
   DROP  7777 ?PAIRS ;

: RUNS ( switch-sys addr n -- switch-sys addr ) \ Usage: nn RUNS <name>
   THIRD 7777 ?PAIRS   ' THIRD >LINK  SWAP , , ;

: RUN: ( switch-sys addr n -- switch-sys addr )  \ Usage: <switch> <value> RUNS: <forth code> ;
   THIRD 7777 ?PAIRS  OVER >LINK ( a n) ,
   HERE 0 ,  :NONAME  SWAP ! ;

: DATA: ( switch-sys addr n -- switch-sys addr )  \ Usage: <switch> <value> DATA: data structure
   THIRD 7777 ?PAIRS
   OVER >LINK ( a n) , HERE CELL+ CODE> ,  ['] (CREATE) COMPILE, ;

: ZERO ( n -- 0 )   DROP 0 ;
