{ ====================================================================
Warnings and errors

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Warning flags

SwiftForth has multiple warning flags. The actions taken depend on the
content of the WARNING variable, which is implemented as a mask and a
set of bitmapped flags.

The first cell of warning is the current warning mask, normally only
zero or -1. The second cell is the bitmap. The logical AND of the two
cells determines the current warning behavior.

Normally, one would say only WARNING ON or WARNING OFF . This would
set or clear the mask completely.  To control the actual actions, one
would edit the value in the second cell of WARNING .

Bitmapped values of WARNING
  0000.0000.0001.0000  warn on redefine (not unique)
  0000.0000.0010.0000  warn on case ambiguity

Bits[12:8] are reserved for SwiftForth-Win32 protection option.

-? suppressed the "not unique" warning message by clearing the WARNING
mask temporarily. It preserves bit[31] as a flag for restore.

/WARNING restores the mask after -? based on bit[31] saved in -? .

WARNS returns true if the particular bitmask is set. For instance, you
can say $10 WARNS IF to see if you should respond to a non-unique
condition or not.
-------------------------------------------------------------------- }

$10 CONSTANT %REDEF     \ Bit mask for warn on redefine
$20 CONSTANT %CASE      \ Bit mask for warn on case ambiguity

2VARIABLE WARNING

: -?  ( -- )   WARNING @  $80000000 AND  WARNING ! ;
: /WARNING ( -- )   WARNING @ 0<  WARNING ! ;
: WARNS ( mask -- flag )   WARNING 2@ AND AND 0<> ;

{ ------------------------------------------------------------------------
?THROW will throw the ior if the flag is true; if false, discard ior.
ABORT simply throws a -1.
ERRMSG contains the error message indicated by ABORT" .
@ERRMSG returns the error message string addr u, clearing the count
byte in ERRMSG (making this a one-shot).
VIEWED contains a line number followed by a counted string filename.
ERR puts an error message into ERRMSG and does the ABORT" .
?ERR does an ABORT" if the flag is true.
(ABORT") forms an abort message from the text at HERE and the quote-
delimited string.
?ABORT is the runtime action of ABORT" .
------------------------------------------------------------------------ }

: ?THROW ( k*x flag ior -- k*x | i*x ior )
   SWAP IF  THROW EXIT  THEN  DROP ;

: ABORT ( i*x -- ) ( R: j*x -- )
   -1 THROW ;

CREATE ERRMSG   256 ALLOT
: @ERRMSG ( -- addr u )   ERRMSG COUNT  0 ERRMSG C! ;

CREATE VIEWED   0 ,  256 ALLOT

: ERR ( addr u -- )
   ERRMSG PLACE  -2 THROW ;

: ?ERR ( flag a n -- )
   ROT IF ERR THEN 2DROP ;

: (ABORT") ( addr u -- )
   H @ COUNT ERRMSG PLACE       \ Word from HERE
   S"  " ERRMSG APPEND          \ Space in between
   ERRMSG APPEND  -2 THROW ;    \ Append error string

: ?ABORT ( flag addr u -- )
   ROT IF (ABORT") THEN 2DROP ;
