{ --------------------------------------------------------------------------
Arg conversion support

These words are useful for parsing command lines, which may contain
words, numbers, or single character options.

ATOI converts a counted string to a single integer.  Returns 0 if it
fails.  0x preceeds hex numbers; all others are decimal.

GETC removes the first character from the string and returns it.
-------------------------------------------------------------------------- }

: ATOI ( addr u1 -- u2 )   BASE @ >R
   OVER C@ [CHAR] - = DUP >R IF 1 /STRING THEN
   OVER W@ $2020 OR $7830 = IF ( 0x format use hex) HEX 2 /STRING THEN
   0 0 2SWAP >NUMBER 2DROP DROP R> IF NEGATE THEN
   R> BASE ! ;

: GETCH ( addr1 u1 -- addr2 u2 char )
   DUP DUP IF  DROP  OVER C@ >R  1 /STRING  R>  THEN ;

{ --------------------------------------------------------------------------
Command-line args

CMDLINE returns a string which doesn't contain the exe name.

SKIP-EXE lets us ignore the exe name that preceeds the command line.

ARGC  returns the number of blank delimited words in the string.

ARGV  returns the Nth word (counting from zero) in the given string.
-------------------------------------------------------------------------- }

| : SKIP-EXE ( a n -- a n )
   GETCH  DUP [CHAR] " <> IF  DROP BL  THEN  SCAN  GETCH DROP  BL SKIP ;

: CMDLINE ( -- addr n )
   GetCommandLine ZCOUNT SKIP-EXE ;

| : QUOTED? ( addr len -- addr len flag )   OVER C@ [CHAR] " = ;

| : THARG ( addr len -- rest len this len )
   0 MAX  BL SKIP  2DUP QUOTED? IF
      1 /STRING  [CHAR] " SCAN  1 /STRING
   ELSE  BL SCAN  THEN
   2SWAP 2 PICK -  QUOTED? IF  1 /STRING  1-  THEN  0 MAX ;

: ARGC ( addr len -- n )
   0 BEGIN   OVER WHILE  ROT ROT THARG 2DROP ROT  1+
   REPEAT  NIP NIP ;

: ARGV ( addr len n -- addr len )
   0 MAX  BEGIN  ?DUP WHILE  OVER WHILE
      ROT ROT THARG 2DROP ROT 1-
   REPEAT 2DROP 0 ELSE THARG 2NIP THEN ;

{ --------------------------------------------------------------------
Environment query

FIND-ENV searches for the string addr1 u1 in the environment and
returns its value and true if found.  Returns the original string
parameters and false if not found.  For example, if the environment
contains the entry "SystemRoot=C:\WINDOWS" then the query
S" SystemRoot" FIND-ENV returns the address and length of the
string "C:\WINDOWS".  Do not include the '=' in the search string.

'ENV returns the address of the environment pointers.
.ENV prints the environment for diagnostic purposes.
"printenv" is the usual Unix command to print the environment.
-------------------------------------------------------------------- }

: FIND-ENV ( addr1 u1 -- addr2 u2 flag )
   R-BUF  R@ ZPLACE  R> PAD 1024 GetEnvironmentVariable
   PAD SWAP DUP 0<> ;

: 'ENV ( -- addr )
   GetEnvironmentStrings ;

: .ENV ( -- )   'ENV  BEGIN  DUP C@ WHILE
   CR ZCOUNT 2DUP TYPE + 1+  REPEAT DROP ;

: PRINTENV ( -- )   .ENV ;

