{ ====================================================================
Console window I/O for Win32

Copyright 2000  FORTH, Inc.

Details on the Win32 Console API prototypes, data structures, etc. can
be found on MSDN (msdn.microsoft.com -- search for "Character-Mode
Applications" or "Console Reference").
==================================================================== }

TARGET

{ --------------------------------------------------------------------
Initialization and exit

INH is the handle to the console input.
OUTH is the handle to the console output.
CMODE-IN and -OUT retain the console mode flags on entry.

INIT-TERM preserves the current tty attributes and initializes ?TTY.
SET-TERM configures the tty for key input.
RESET-TERM restores tty attributes.
-------------------------------------------------------------------- }

VARIABLE INH    : STDIN  ( -- h )   INH @ ;
VARIABLE OUTH   : STDOUT ( -- h )   OUTH @ ;

VARIABLE ?TTY           \ flag set if console is a tty

-HEADS

VARIABLE CMODE-IN       \ saved tty entry states
VARIABLE CMODE-OUT

-11 EQU STD_OUTPUT_HANDLE
-10 EQU STD_INPUT_HANDLE

: INIT-TERM ( -- )
   STD_OUTPUT_HANDLE GetStdHandle OUTH !
   STD_INPUT_HANDLE GetStdHandle INH !
   Z" SwiftForth Kernel" SetConsoleTitle DROP
   STDIN CMODE-IN GetConsoleMode  0<> ?TTY !
   STDOUT CMODE-OUT GetConsoleMode DROP ;

: SET-TERM ( -- )
   STDIN 0 SetConsoleMode DROP          \ Set tty input to raw
   STDOUT 1 SetConsoleMode DROP ;       \ ENABLE_PROCESSED_OUTPUT

: RESET-TERM ( -- )                     \ Reset tty back to entry state
   STDIN CMODE-IN @ SetConsoleMode DROP
   STDOUT CMODE-OUT @ SetConsoleMode DROP ;

{ --------------------------------------------------------------------
Terminate

EXITSTATUS is the exit status returned to the OS.

_BYE provides the default vectored behavior for BYE.  It terminates
SwiftForth with the code in EXITSTATUS.
-------------------------------------------------------------------- }

+HEADS

VARIABLE EXITSTATUS

: _BYE ( -- )
   CR  ?TTY @ IF  RESET-TERM  NORMAL  THEN
   EXITSTATUS @ ExitProcess ;

XDEFER BYE   ' _BYE ,

-HEADS

{ --------------------------------------------------------------------
Console I/O

(C-KEY?) tests the console input device.
(C-KEY) waits for and returns a keystroke from the console.
(C-EKEY) waits for and returns an extended keystroke from the console.
(C-TYPE) writes a string to the console.
(C-EMIT) writes a character to the console.
(C-CR) writes a CR,LF to the console.

typedef struct _INPUT_RECORD (
  WORD  EventType;
  union (
    KEY_EVENT_RECORD          KeyEvent;
    MOUSE_EVENT_RECORD        MouseEvent;
    WINDOW_BUFFER_SIZE_RECORD WindowBufferSizeEvent;
    MENU_EVENT_RECORD         MenuEvent;
    FOCUS_EVENT_RECORD        FocusEvent;
  ) Event;
)

If EventType is KEY_EVENT 0x0001, this is the KEY_EVENT_RECORD:

typedef struct _KEY_EVENT_RECORD (
  BOOL  bKeyDown;
  WORD  wRepeatCount;
  WORD  wVirtualKeyCode;
  WORD  wVirtualScanCode;
  union (
    WCHAR UnicodeChar;
    CHAR  AsciiChar;
  ) uChar;
  DWORD dwControlKeyState;
) KEY_EVENT_RECORD;

-------------------------------------------------------------------- }

: (C-TYPE) ( addr n -- )   STDOUT WRITE-FILE DROP ;

: (C-EMIT) ( char -- )   SP@ 1 (C-TYPE) DROP ;

: (C-CR) ( -- )   13 (C-EMIT) 10 (C-EMIT) ;

: (C-KEY?) ( -- flag )
   [ 20 256 * CELL + ] LITERAL R-ALLOC >R
   STDIN R@ CELL+ 256 R@ PeekConsoleInput DROP
   R> @+ 0 ?DO ( addr)  DUP W@ 1 ( KEY_EVENT) =
      OVER 4 + W@ ( bKeyDown) 0<> AND
      OVER 14 + W@ ( uChar) AND IF
   DROP -1 UNLOOP EXIT  THEN  20 +  LOOP DROP  0 ;

CREATE CIB 20 ALLOT   VARIABLE #CI

: (C-KEY) ( -- char )
   BEGIN  STDIN CIB 1 #CI ReadConsoleInput DROP
      CIB W@ 1 ( KEY_EVENT) =  CIB 4 + W@ ( bKeyDown) 0<> AND
   CIB 14 + W@ ( uChar) AND  ?DUP UNTIL ;

: (C-EKEY) ( -- echar )
   BEGIN  STDIN CIB 1 #CI ReadConsoleInput DROP
   CIB W@ 1 =  CIB CELL+ @ 0<> AND UNTIL
   CIB 14 + W@ ?DUP 0= IF  CIB 10 + W@  $10000 OR  THEN ;

{ --------------------------------------------------------------------
Console window

'CSBI holds the console screen buffer info.
[CSBI] defines word fields within 'CSBI.  Each returns its value.
/CSBI reads the current console screen buffer info.

(C-ATXY) uses SetConsoleCursorPosition to set the cursor position.

(C-GETXY) calls GetConsoleScreenBufferInfo to read the screen buffer
info into 'CSBI, then returns the X,Y position coordinates from there.

(C-ATTRIBUTE) sets attribute n.

(C-PAGE) clears the screen and sets the cursor at 0,0.

See the end of this source file for API functions and structures.
-------------------------------------------------------------------- }

CREATE 'CSBI  22 ALLOT          \ CONSOLE_SCREEN_BUFFER_INFO, see below

: /CSBI ( -- )   STDOUT 'CSBI GetConsoleScreenBufferInfo DROP ;

INTERPRETER
: [CSBI] ( n1 -- n2 )   CREATE DUP C, 2+  DOES> ( -- x )  C@ 'CSBI + W@ ;
TARGET

0  [CSBI] CS.BUFX       \ Total buffer size x
   [CSBI] CS.BUFY       \ Total buffer size y
   [CSBI] CS.CPX        \ Cursor x
   [CSBI] CS.CPY        \ Cursor y
   [CSBI] CS.ATTR       \ Attribute
   [CSBI] CS.ULX        \ Window upper left x (within buffer)
   [CSBI] CS.ULY        \ " upper left y
   [CSBI] CS.LRX        \ " lower right x
   [CSBI] CS.LRY        \ " lower right y
DROP

: (C-ATXY) ( x y -- )
   /CSBI  SWAP CS.ULX +  SWAP CS.ULY +  16 LSHIFT +
   STDOUT SWAP SetConsoleCursorPosition DROP ;

: (C-GETXY) ( -- x y )
   /CSBI  CS.CPX CS.ULX -  CS.CPY CS.ULY - ;

: (C-GETSIZE) ( -- x y )
   /CSBI  CS.LRX CS.ULX - 1+  CS.LRY CS.ULY - 1+ ;

CREATE 'ATTRIB
   $07 C,       \ Normal
   $70 C,       \ Inverse
   $0C C,       \ Bright
   $0B C,       \ Bold
     0 C,       \ Invisible

: (C-ATTRIBUTE) ( n -- )   0 MAX 4 MIN 'ATTRIB + C@
   STDOUT SWAP SetConsoleTextAttribute DROP ;

VARIABLE 'WLEN

: (C-PAGE) ( -- )
   /CSBI  CS.BUFX CS.BUFY * >R
   STDOUT BL R@ 0 'WLEN FillConsoleOutputCharacter DROP
   STDOUT 'ATTRIB C@ R> 0 'WLEN FillConsoleOutputAttribute DROP
   STDOUT 0 SetConsoleCursorPosition DROP ;

{ --------------------------------------------------------------------
Non-tty input
-------------------------------------------------------------------- }

: (N-XY) ( -- x y )   0 0 ;

: (N-ACCEPT) ( c-addr u1 -- u2 )
   OVER SWAP 0 DO  DUP 1 STDIN READ-FILE 0= AND IF
      DUP C@ $0A = IF  1- LEAVE  THEN  1+
      ELSE  I 0= IF  BYE  THEN  LEAVE
   THEN LOOP  SWAP - ;

: (N-KEY) ( -- char )
   0 SP@ 1 STDIN READ-FILE 0= AND 0= IF  BYE  THEN ;

{ --------------------------------------------------------------------
Initialization

CONSOLE-MODE is a personality vector for Windows console I/O.

/CONSOLE sets the handles and vectors for the console device.
-------------------------------------------------------------------- }

CREATE TTY-MODE
        16 ,            \ datasize
        19 ,            \ #vectors
         0 ,            \ PHANDLE
         0 ,            \ PREVIOUS
   ' NOOP ,             \ INVOKE    ( -- )
   ' NOOP ,             \ REVOKE    ( -- )
   ' NOOP ,             \ /INPUT    ( -- )
   ' (C-EMIT) ,         \ EMIT      ( char -- )
   ' (C-TYPE) ,         \ TYPE      ( addr len -- )
   ' (C-TYPE) ,         \ ?TYPE     ( addr len -- )
   ' (C-CR) ,           \ CR        ( -- )
   ' (C-PAGE) ,         \ PAGE      ( -- )
   ' (C-ATTRIBUTE) ,    \ ATTRIBUTE ( n -- )
   ' (C-KEY) ,          \ KEY       ( -- char )
   ' (C-KEY?) ,         \ KEY?      ( -- flag )
   ' (C-EKEY) ,         \ EKEY      ( -- echar )
   ' (C-KEY?) ,         \ EKEY?     ( -- flag )
   ' (C-EKEY) ,         \ AKEY      ( -- echar )
   ' 2DROP ,            \ PUSHTEXT  ( addr len -- )
   ' (C-ATXY) ,         \ AT-XY     ( x y -- )
   ' (C-GETXY) ,        \ GET-XY    ( -- x y )
   ' (C-GETSIZE) ,      \ GET-SIZE  ( -- x y )
   ' (ACCEPT) ,         \ ACCEPT    ( addr u1 -- u2 )

CREATE NOTTY-MODE
        16 ,            \ datasize
        19 ,            \ #vectors
         0 ,            \ PHANDLE
         0 ,            \ PREVIOUS
   ' NOOP ,             \ INVOKE    ( -- )
   ' NOOP ,             \ REVOKE    ( -- )
   ' NOOP ,             \ /INPUT    ( -- )
   ' (C-EMIT) ,         \ EMIT      ( char -- )
   ' (C-TYPE) ,         \ TYPE      ( addr len -- )
   ' (C-TYPE) ,         \ ?TYPE     ( addr len -- )
   ' (C-CR) ,           \ CR        ( -- )
   ' NOOP ,             \ PAGE      ( -- )
   ' DROP ,             \ ATTRIBUTE ( n -- )
   ' (N-KEY) ,          \ KEY       ( -- char )
   ' (C-KEY?) ,         \ KEY?      ( -- flag )
   ' (N-KEY) ,          \ EKEY      ( -- echar )
   ' (C-KEY?) ,         \ EKEY?     ( -- flag )
   ' (N-KEY) ,          \ AKEY      ( -- char )
   ' 2DROP ,            \ PUSHTEXT  ( addr len -- )
   ' 2DROP ,            \ AT-XY     ( x y -- )
   ' (N-XY) ,           \ GET-XY    ( -- x y )
   ' (N-XY) ,           \ GET-SIZE  ( -- x y )
   ' (N-ACCEPT) ,       \ ACCEPT    ( addr u1 -- u2 )

+HEADS

: /CONSOLE ( -- )   NOTTY-MODE  INIT-TERM
   ?TTY @ IF  DROP TTY-MODE  SET-TERM  THEN
   'PERSONALITY ! ;
