{ ====================================================================
File operations

Copyright 2001  FORTH, Inc.  

This file supplies the Windows-specific ANS Forth File Access and
File-access Extension words.  Generic file operators are in the
common kernel source.
==================================================================== }

TARGET

{ --------------------------------------------------------------------
ANS File Access Methods

The ANS file interface requires these methods for opening and creating
files.  They are inherited directly from Windows. The BIN method is a
noop, as all access is binary.

The (non-ANS) EXE method is here for compatibility with the Linux
implementation.  It's also a noop as Windows does not have
"executable" file attributes.
-------------------------------------------------------------------- }

GENERIC_READ  CONSTANT R/O
GENERIC_WRITE CONSTANT W/O
R/O W/O OR    CONSTANT R/W

: BIN ( fam -- fam ) ;  IMMEDIATE
: EXE ( fam -- fam ) ;  IMMEDIATE

{ ------------------------------------------------------------------------
ANS file access

The fileid tokens used here are Windows file handles.

CREATE-FILE makes a new file and opens it for I/O. If the file already
exists, it is deleted without warning; the new file takes its place.

OPEN-FILE opens an existing file for I/O based on fam.
CLOSE-FILE closes an open file.
DELETE-FILE removes the file from the filesystem.
RENAME-FILE changes the name of a file in the filesystem.

The EQU definitions make the Windows "constants" available to the
cross-compiler.
------------------------------------------------------------------------ }

CREATE_ALWAYS             EQU CREATE_ALWAYS
OPEN_EXISTING             EQU OPEN_EXISTING
FILE_ATTRIBUTE_NORMAL     EQU FILE_ATTRIBUTE_NORMAL
FILE_SHARE_READ           EQU FILE_SHARE_READ
FILE_FLAG_RANDOM_ACCESS   EQU FILE_FLAG_RANDOM_ACCESS
FILE_FLAG_SEQUENTIAL_SCAN EQU FILE_FLAG_SEQUENTIAL_SCAN

: CREATE-FILE ( c-addr u fam -- fileid ior )
   R-BUF -ROT R@ ZPLACE
   R> SWAP FILE_SHARE_READ 0 CREATE_ALWAYS
   FILE_ATTRIBUTE_NORMAL FILE_FLAG_SEQUENTIAL_SCAN OR
   0 CreateFile   DUP 1+ 0= -198 AND ;

: OPEN-FILE ( c-addr u fam -- fileid ior )
   R-BUF -ROT R@ ZPLACE
   R> SWAP FILE_SHARE_READ 0 OPEN_EXISTING
   FILE_ATTRIBUTE_NORMAL FILE_FLAG_SEQUENTIAL_SCAN OR
   0 CreateFile   DUP 1+ 0= -199 AND ;

: CLOSE-FILE ( fileid -- ior )
   CloseHandle 0= -197 AND ;

: DELETE-FILE ( caddr u -- ior )
   R-BUF R@ ZPLACE  R> DeleteFile 0= -191 AND ;

: RENAME-FILE ( caddr1 u1 caddr2 u2 -- ior )
   R-BUF R@ ZPLACE  R>  -ROT
   R-BUF R@ ZPLACE  R>  SWAP MoveFile 0= -192 AND ;

{ --------------------------------------------------------------------
ANS file read/write

READ-FILE attempts to read u1 bytes from the file into memory at the
address.  The actual number of bytes read is returned as u2.

WRITE-FILE writes u bytes to the file.

FLUSH-FILE asks windows to write all cached data to disk.
-------------------------------------------------------------------- }

: READ-FILE ( c-addr u1 fileid -- u2 ior )
   -ROT 0 >R RP@ 0 ReadFile  R> SWAP 0= -195 AND ;

: WRITE-FILE ( c-addr u fileid -- ior )
   -ROT 0 >R RP@ 0 WriteFile R>DROP 0= -196 AND ;

: FLUSH-FILE ( fileid -- ior )
   FlushFileBuffers 0= -194 AND ;

{ --------------------------------------------------------------------
ANS file size and position

FILE-SIZE returns the size in bytes of a file.  If low half of value
returned by GetFileSize is -1, this *might* be error.  Only in that
case do we query GetLastError.  If the -1 is not there, we must not
check GetLastError because it might be from a previous operation.

FILE-POSITION returns the linear position of the file.
REPOSITION-FILE forces the file to a linear position from its start.
RESIZE-FILE sets a new size for the file.
FILE-STATUS returns zero if the file exists, or an ior if not.
-------------------------------------------------------------------- }

: FILE-SIZE ( fileid -- ud ior )
   0 >R RP@ GetFileSize R>  OVER 1+ IF  0
   ELSE  GetLastError 0<> -37 AND  THEN ;

FILE_BEGIN    EQU FILE_BEGIN
FILE_CURRENT  EQU FILE_CURRENT
FILE_END      EQU FILE_END

: FILE-POSITION ( fileid -- ud ior )
   0  0 >R RP@  FILE_CURRENT SetFilePointer R> 0 ;

: REPOSITION-FILE ( ud fileid -- ior )
   -ROT >R RP@ FILE_BEGIN SetFilePointer R>2DROP 0 ;

: RESIZE-FILE ( ud fileid -- ior )
   DUP>R REPOSITION-FILE DROP R> SetEndOfFile 0= -193 AND ;

: FILE-STATUS ( caddr u -- x ior )
   R-BUF R@ ZPLACE  R> GetFileAttributes DUP -1 = -199 AND ;

{ -------------------------------------------------------------------- }

CREATE <EOL>   2 C, $0D C, $0A C,  0 C,         \ Windows line-end sequence

{ --------------------------------------------------------------------
Memory mapped files

These two non-ANS file functions are in support of INCLUDE-FILE.

MAP-FILE takes the fileid of a file that is open for read access
(presumably a source file) and does the following:
   1) determines its size
   2) allocates a memory buffer for it
   3) reads the file into the buffer
   4) returns the address and amount read
Returns ior.  Cleans up after itself (i.e. unallocates memory) if
there's a failure, but always leaves file associated with fileid open.

UNMAP-FILE takes the buffer address and total size and frees the
buffer.  The total size is not used in Windows, but is supplied for
compatibility with other operating systems.
-------------------------------------------------------------------- }

: MAP-FILE ( fileid -- addr u ior )
   DUP FILE-SIZE  ?DUP IF  >R  2DROP DROP  0 0 R> EXIT  THEN    \ Can't get the size, return ior and bail
   DROP DUP ALLOCATE  ?DUP IF  >R 2DROP 0 0 R> EXIT  THEN       \ Can't allocate memory, return ior and bail
   DUP 2SWAP SWAP READ-FILE ;                                   \ Return buf addr, actual read, ior

: UNMAP-FILE ( addr u -- ior )
   DROP FREE ;
