{ ====================================================================
Win32 file and path support

Copyright 2001  FORTH, Inc.

==================================================================== }

TARGET

{ ------------------------------------------------------------------------
Path support

-NAME strips the name and trailing \ from the end of a full pathname.
-EXT strips the trailing extension from the filename.
-PATH strips the path from the beginning of a full pathname.

DEFEXT is the default Forth extension .F  The buffer it occupies is
   larger than needed by ".f" so that it can be changed to anything.
------------------------------------------------------------------------ }

| : ENDING ( addr len char -- addr len )
   >R BEGIN  2DUP R@ SCAN
      ?DUP WHILE  2SWAP 2DROP  1 /STRING
   REPEAT  R> 2DROP ;

: -NAME ( addr len -- addr len )
   2DUP  [CHAR] \ ENDING  NIP -  1-  0 MAX ;

: -EXT ( addr len -- addr len )
   2DUP  [CHAR] . ENDING  NIP -  1-  0 MAX ;

: -PATH ( addr len -- addr len )
   [CHAR] \ ENDING  0 MAX ;

CREATE DEFEXT  2 C,  CHAR . C,  CHAR f C,  253 ALLOT

: THIS-EXE-NAME ( -- addr n )
   0 GetModuleHandle PAD 255 GetModuleFileName DROP
   PAD ZCOUNT POCKET PLACE  POCKET COUNT ;

{ --------------------------------------------------------------------
Root path

We do most file operations relative to the "root" directory in which
SwiftForth is installed. This is so the base directory in which it is
installed doesn't matter, and all the files whose names are relative
to this path can be found even if installed on another drive or path.
For all this to work, SwiftForth must be launched either directly or
from a shortcut to the SwiftForth\Bin directory.

ROOTPATH holds the root of our system.  This is the "parent" of
the SwiftForth directory.

/ROOTPATH takes THIS-EXE-NAME (the full path to the SwiftForth
executable, normally sf.exe in the SwiftForth\bin directory) and
places the root path (the path that leads up to SwiftForth\ ) into
ROOTPATH.

ROOTED? is true if the string begins with the rootpath string.

-ROOT removes the rootpath from the zstring.

+ROOT adds the rootpath back to the string.
-------------------------------------------------------------------- }

CREATE ROOTPATH   256 ALLOT

: /ROOTPATH ( -- )
   THIS-EXE-NAME -NAME -NAME -NAME  ROOTPATH PLACE
   S" \" ROOTPATH APPEND ;

: ROOTED? ( addr n -- flag )
   DROP ROOTPATH COUNT TUCK COMPARE(NC) 0= ;

: -ROOT ( z-addr -- addr len )
   DUP ZCOUNT ROOTED? IF
      DUP ZCOUNT  ROOTPATH C@ /STRING  ( z-addr addr len)
      THIRD S" %" THIRD ZPLACE  ZAPPEND
   THEN ZCOUNT ;

: +ROOT ( addr n -- addr n )
   OVER C@ [CHAR] % <> ?EXIT
   ROOTPATH COUNT POCKET PLACE
   1 /STRING POCKET APPEND
   POCKET COUNT ;

{ --------------------------------------------------------------------
File name

FULLNAME places the fully qualified path name into the zbuffer. This
has a limit of 255 characters!

DEFAULT.EXT appends the default extension from DEFEXT if needed.
-------------------------------------------------------------------- }

: FULLNAME ( addr n zbuf -- )
   R-BUF -ROT R@ ZPLACE
   R@ 255 ROT 0 >R RP@ GetFullPathName DROP 2R> 2DROP ;

: DEFAULT.EXT ( addr1 n1 -- addr2 n2 )
   2DUP -PATH [CHAR] . SCAN NIP ?EXIT
   POCKET PLACE  DEFEXT COUNT POCKET APPEND  POCKET COUNT ;

| : ROOTED ( addr len -- addr len )   DUP IF
      OVER C@ [CHAR] % = IF +ROOT  HERE PLACE  HERE COUNT
   THEN  THEN ;

: FILENAME-FIXUP ( addr len -- addr len flag )
   -TRAILING DUP IF
      2DUP OVER + SWAP DO  I C@ [CHAR] / = IF  [CHAR] \ I C!  THEN  LOOP
      ROOTED  DEFAULT.EXT  2DUP FILE-STATUS NIP 0=
   ELSE  0  THEN ;
