{ ====================================================================
Import library functions

Copyright 2001  FORTH, Inc.

The calling of external library functions is implemented here.

======================================================================

LIBRARY parameter field:  | handle | flag |
IMPORT: parameter field:  | proc addr | #params | ^lib |
==================================================================== }

TARGET

{ --------------------------------------------------------------------
SwiftForth's PE header includes automatic resolution for these API calls.
-------------------------------------------------------------------- }

FUNCTION: LoadLibrary ( lpLibFileName -- h )
FUNCTION: GetProcAddress ( hModule lpProcName -- addr )
FUNCTION: MessageBox ( hWnd lpText lpCaption uType -- n )
FUNCTION: ExitProcess ( uExitCode -- x )
FUNCTION: GetModuleHandle ( pModuleName -- h )

0 IMPORTS !

: HINST ( -- h )   0 GetModuleHandle ;

{ --------------------------------------------------------------------
KERNEL32 library functions required by the SwiftForth kernel.
-------------------------------------------------------------------- }

LIBRARY KERNEL32

FUNCTION: CloseHandle ( hObject -- b )
FUNCTION: CreateFile ( lpFileName dwDesiredAccess dwShareMode lpSecurityAttributes dwCreationDistribution dwFlagsAndAttributes hTemplateFile -- h )
FUNCTION: DeleteFile ( lpFileName -- b )
FUNCTION: FlushFileBuffers ( hFile -- b )
FUNCTION: MoveFile ( lpExistingFileName lpNewFileName -- b )
FUNCTION: CopyFile ( lpExistingFileName lpNewFileName bFailIfExists -- b )
FUNCTION: ReadFile ( hFile lpBuffer nNumberOfBytesToRead lpNumberOfBytesRead lpOverlapped -- b )
FUNCTION: SetCurrentDirectory ( lpPathName -- b )
FUNCTION: SetEndOfFile ( hFile -- b )
FUNCTION: SetFilePointer ( hFile lDistanceToMove lpDistanceToMoveHigh dwMoveMethod -- x )
FUNCTION: WriteFile ( hFile lpBuffer nNumberOfBytesToWrite lpNumberOfBytesWritten lpOverlapped -- b )
FUNCTION: GetFileAttributes ( pFileName -- u )
FUNCTION: GetCurrentDirectory ( nBufferLength lpBuffer -- u )
FUNCTION: GetFileSize ( hFile lpFileSizeHigh -- u )
FUNCTION: GetFullPathName ( lpFileName nBufferLength lpBuffer *lpFilePart -- u )
FUNCTION: DeleteCriticalSection ( lpCriticalSection -- x )
FUNCTION: InitializeCriticalSection  ( lpCriticalSection -- x )
FUNCTION: FatalAppExit ( uAction lpMessageText -- x )
FUNCTION: FreeLibrary ( hLibModule -- b )
FUNCTION: GetCommandLine ( -- addr )
FUNCTION: GetModuleFileName ( hModule lpFilename nSize -- u )
FUNCTION: GlobalAlloc ( uFlags dwBytes -- h )
FUNCTION: GlobalFree ( hMem -- h )
FUNCTION: GlobalSize ( hMem -- u )
FUNCTION: VirtualAlloc ( lpAddress dwSize flAllocationType flProtect -- addr )
FUNCTION: VirtualFree ( lpAddress dwSize dwFreeType -- b )
FUNCTION: GetLastError ( -- u )
FUNCTION: GetEnvironmentStrings ( -- addr )
FUNCTION: GetEnvironmentVariable ( lpName lpBuffer nSize -- u )
FUNCTION: AddVectoredExceptionHandler ( u lpHandler -- h )

\ Console window functions:

FUNCTION: GetStdHandle ( nStdHandle -- h )
FUNCTION: PeekConsoleInput ( hConsoleInput lpBuffer nLength lpNumberOfEventsRead -- b )
FUNCTION: ReadConsoleInput ( hConsoleInput lpBuffer nLength lpNumberOfEventsRead -- b )
FUNCTION: SetConsoleMode ( hConsoleHandle dwMode -- b )
FUNCTION: GetConsoleMode ( hConsoleHandle lpMode -- b )
FUNCTION: FreeConsole ( -- b )
FUNCTION: SetConsoleCursorPosition ( hConsoleOutput dwCursorPosition -- b )
FUNCTION: GetConsoleScreenBufferInfo ( hConsoleOutput lpConsoleScreenBufferInfo -- b )
FUNCTION: SetConsoleTitle ( lpConsoleTitle -- b )
FUNCTION: SetConsoleTextAttribute ( hConsoleOutput wAttributes -- b )
FUNCTION: FillConsoleOutputCharacter ( hConsoleOutput cCharacter nLength dwWriteCoord lpNumberOfCharsWritten -- b )
FUNCTION: FillConsoleOutputAttribute ( hConsoleOutput cAttribute nLength dwWriteCoord lpNumberOfAttrsWritten -- b )

{ --------------------------------------------------------------------
USER32 library functions required by the kernel.
-------------------------------------------------------------------- }

LIBRARY USER32

4 IMPORT: GetMessage      ( msg hwnd min max -- bool )
5 IMPORT: PeekMessage     ( msg hwnd min max flags -- bool )
1 IMPORT: TranslateMessage ( msg -- bool )
1 IMPORT: DispatchMessage ( msg -- bool )
0 IMPORT: WaitMessage     ( -- bool )

{ --------------------------------------------------------------------
Resolve DLL calls

/PE-IMPORTS initializes the base set of API imports from the address
at ORIGIN CELL+ (set by the PE loader code).  This short list of API
calls is resolved by the PE loader.  Each entry in the loader table is
the self-relative dword offset to the location that holds the API
resolved address.

See SwiftForth/src/loader/win/<arch> (where <arch> is i386 or x64) for
the PE loader source.

GetProcAddressX attempts to find a function in the DLL whose handle is
given.  If the initial attempt fails, it appends "A" to the name and
tries again.  This allows simple use of functions that are both
Unicode and ASCII.

RESOLVE-PROCS looks up all functions in the IMPORTS list.
-------------------------------------------------------------------- }

| : (/PE-IMPORT) ( addr1 xt -- addr2 )
   OVER DUP L@ + L@  SWAP >BODY !  4 + ;

: /PE-IMPORTS ( -- )
   ORIGIN CELL+ @
   ['] ExitProcess (/PE-IMPORT)
   ['] LoadLibrary (/PE-IMPORT)
   ['] GetProcAddress (/PE-IMPORT)
   ['] GetModuleHandle (/PE-IMPORT)
   ['] MessageBox (/PE-IMPORT)  DROP ;

| : (OPEN-LIBS) ( addr -- )
   DUP @ IF  DUP @REL RECURSE  THEN
   L>NAME  DUP N>BODY  SWAP 1+ LoadLibrary SWAP ! ;

| : OPEN-LIBS ( -- )
   LIBS WID> CELL+ @REL (OPEN-LIBS) ;

: GetProcAddressX ( handle z -- addr | 0 )
   R-BUF ZCOUNT R@ ZPLACE
   DUP R@ GetProcAddress ( h a)
   ?DUP IF  NIP R>DROP EXIT  THEN
   S" A" R@ ZAPPEND
   R> GetProcAddress ;

| : RESOLVE-PROCS ( -- )
   IMPORTS BEGIN                        \ link
      @REL ?DUP WHILE                   \ link
      DUP CELL+ @REL                    \ link nfa
      DUP N>BODY                        \ link nfa pfa
      DUP CELL+ CELL+ @REL @            \ link nfa pfa libh
      ROT 1+ GetProcAddressX            \ link pfa procaddr
      SWAP !
   REPEAT ;

| : /KERNEL-IMPORTS ( -- )
   OPEN-LIBS  RESOLVE-PROCS ;
