{ ====================================================================
Start SwiftForth

Copyright 2001  FORTH, Inc.

Main entry point and initialization of SwiftForth.
==================================================================== }

TARGET

{ ------------------------------------------------------------------------
Command line parsing

PARSE-ARGS evaluates the command line. If the command line contains
only a filename, that file is included.  Otherwise, the string is
evaluated. If an error is caught, the user has the option of pressing
X to terminate the program.

if file exists and is not directory, include it. otherwise evaluate

file-status


For example, if SwiftForth is started by
   SF FOO.F
...the file FOO.F will be included if it is found.

If it is started by
   SF HERE 100 DUMP
...memory from HERE will be displayed.

------------------------------------------------------------------------ }

: PARSE-ARGS ( -- )   CMDLINE NIP -EXIT
   CMDLINE 2DUP 0 ARGV
   FILE-STATUS 0<> SWAP 16 = OR IF ( no file, or directory)
      ['] EVALUATE
   ELSE ( normal file)
      0 ARGV ['] INCLUDED
   THEN
   CATCH ?DUP IF  DUP .CATCH
      ?TTY @ 0= IF  EXITSTATUS !  BYE  THEN
   BAILOUT?  DROP 2DROP  THEN  ."  ok" CR QUIT ;

' PARSE-ARGS ' /CMDLINE >BODY !

{ ------------------------------------------------------------------------
START is not a user accessable function. It simply is enough to get
things going and run the 'MAIN routine.

The behavior differs if this is run from the target-compiled kernel or
from a turney image.  This difference is detected by the state of the
global variable UP0, which contains the relative address of the
CONSOLE user area after the system gets running. If zero, then we are
running from a kernel image and the system memory allocations are made
here.  If non-zero, this is a turnkey image and the saved system
allocations are restored.  Then the user area is initialized and 'MAIN
is executed.
------------------------------------------------------------------------ }

: START ( -- )
   /PE-IMPORTS                          \ Import PE header functions
   UP0 @ IF  UP0 @REL /KERNEL DROP      \ If UP0 points to memory, use it
   ELSE  DP0 @REL  /KERNEL  DP0 !REL  THEN
   DP0 @REL H !                         \ Start of dictionary
   HINST $1F8 + @ +ORIGIN HLIM !        \ End of dictionary (SizeOfImage for .data section in PE header)
   RP@ R0 !  SP@ S0 !                   \ stack is where we find it
   'THIS OFF  'SELF OFF                 \ no object or class
   /FORTH  DECIMAL                      \ Search order, base
   STATE OFF   0 'METHOD !              \ interpreting
   %REDEF -1 WARNING 2!                 \ default warnings
   /IMPORTS  /SYSTEM  /SIGNAL           \ init dlls, system chain, vectored exception handler
   /ROOTPATH  /VERSION  /NDP            \ set rootpath, version, fp
   'MAIN @EXECUTE  BYE ;                \ and run the specified entry point

{ --------------------------------------------------------------------
Loader entry

The SwiftForth PE header stub code passes control to POWER-UP with the
SwiftForth virtual memory base address in EDX.

POWER-UP sets EBP to the SwiftForth data stack on the Windows return
stack and adjusts ESP below it, then passes control to START.
-------------------------------------------------------------------- }

LABEL POWER-UP
   EDX EDI MOV                  \ Virtual memory base (from PE loader).
   ESP EBP MOV                  \ EBP is Forth data stack
   $8000 # ESP SUB              \ Allocate 32k for data stack
   ' START JMP   END-CODE       \ Transfer to START

HOST
