{ ====================================================================
Vocabularies and wordlists

Copyright 2001  FORTH, Inc.
==================================================================== }

TARGET

{ --------------------------------------------------------------------
Standard wordlists

A wordlist can have multiple strands. Different wordlists may have
different numbers of strands. The initial system only knows four
wordlists. These are built here:

1) FORTH-WORDLIST contains all the primary words known to the system.
When FORTH is defined later, it will be referenced.

2) FILES-WORDLIST contains all the filenames of the sources of the
system (used for LOCATE).

3) LIBS has all the external library names (as defined by LIBRARY both
here in the kernel and later in the IDE electives), used to open the
libraries and resolve library calls.

4) ROOT-WORDLIST contains no words now, and will have a minimal set of
words when it is added to later.

The search order is a stack built at CONTEXT. Definitions are compiled
into CURRENT. No defining word changes CONTEXT.

Wordlist structure
   | wid-link | #STRANDS | strand 0 | strand 1 | ... | strand n-1 |

The WID returned for a wordlist is the address of the #STRANDS entry
of the structure relative to the kernel ORIGIN. Note that all links
stored in a wordlist must be position relative.
-------------------------------------------------------------------- }

HERE  0 ,                       \ wid link
HERE ( *)  =STRANDS ,           \ number of strands
HERE EQU =FORTH                 \ equate for strands
   =STRANDS CELLS ALLOT         \ the actual strands

( *) CONSTANT FORTH-WORDLIST    \ FORTH wid


HERE  SWAP ,REL(T)              \ wid link
HERE ( *)  1 ,                  \ single strand
HERE EQU =FILES                 \ equate for thread
   0 ,                          \ Pointer to filename list

( *) CONSTANT FILES-WORDLIST    \ Files list wid


HERE  SWAP ,REL(T)              \ wid link
HERE ( *)  1 ,                  \ single strand
HERE EQU =LIBS                  \ equate for thread
   0 ,                          \ Pointer to filename list

( *) CONSTANT LIBS              \ Libraries list wid


HERE  SWAP ,REL(T)              \ wid link
HERE ( *)  1 ,                  \ single strand
   0 ,                          \ pointer to root word list
( *) CONSTANT ROOT-WORDLIST

( Head) WIDS !REL(T)

{ --------------------------------------------------------------------
Search order

GET-ORDER returns the number of word lists n in the search order and
the word list identifiers widn ... wid1 identifying these word lists
where wid1 identifies the word list that is searched first, and widn
the word list that is searched last. The search order is unaffected.

SET-ORDER sets the search order to the word lists identified by widn
... wid1. Subsequently, word list wid1 will be searched first, and
word list widn searched last. If n is zero, the search order is
cleared. If n is -1, the search order is set to ROOT-WORDLIST.

GET-CURRENT returns wid, the identifier of the compilation word list.

SET-CURRENT sets the compilation word list to wid.

DISCARD tosses away u items from the stack.

ONLY forces the minimum search order (ROOT-WORDLIST).

ALSO duplicates the first wordlist in the search order.

PREVIOUS deletes the first wordlist in the search order. [PREVIOUS] is
the immediate version of PREVIOUS.

SEAL forces the search order to be only the first item in it.

DEFINITIONS sets the compilation wordlist to the first one in the
search order.
-------------------------------------------------------------------- }

: GET-ORDER ( -- widn ... wid1 n )
   #ORDER @ IF
      #ORDER @ 0 DO
         #ORDER @  I - 1- CELLS CONTEXT + @
      LOOP
   THEN
   #ORDER @ ;

: SET-ORDER ( widn .. wid1 n -- )
   DUP -1 = IF  DROP ROOT-WORDLIST 1  THEN
   DUP #ORDER !  0 ?DO  I CELLS CONTEXT + !  LOOP ;

: SET-CURRENT ( wid -- )   CURRENT ! ;
: GET-CURRENT ( -- wid )   CURRENT @ ;

: DISCARD ( u*x u -- )
   BEGIN ?DUP WHILE NIP 1- REPEAT ;

: ONLY ( -- )   -1 SET-ORDER ;
: ALSO     ( -- )   GET-ORDER OVER SWAP 1+ SET-ORDER ;
: PREVIOUS ( -- )   GET-ORDER NIP 1-       SET-ORDER ;
: [PREVIOUS]   PREVIOUS ; IMMEDIATE
: SEAL     ( -- )   GET-ORDER OVER 1       SET-ORDER  DISCARD ;
: DEFINITIONS ( -- )   GET-ORDER  OVER SET-CURRENT  DISCARD ;

| : /FORTH ( -- )   FORTH-WORDLIST DUP 1 SET-ORDER  SET-CURRENT ;

{ --------------------------------------------------------------------
-ORDER removes a WID from the context list and
+ORDER guarantees that the WID is in the context list a) only once,
   and b) at the top, or first, in the search order.
?ORDER is true if the WID is in the search order.
-------------------------------------------------------------------- }

: -ORDER ( wid -- )
   >R  GET-ORDER R> ( wid wid ... wid n wid)
   0 >R BEGIN
      SWAP ?DUP WHILE 1- SWAP ( wid...wid n wid)
      ROT 2DUP = IF DROP ELSE
         R> SWAP >R 1+ >R
      THEN
   REPEAT DROP
   R> DUP BEGIN ( n n)
      ?DUP WHILE 1-
      R> -ROT
   REPEAT SET-ORDER ;

: +ORDER ( wid -- )
   DUP -ORDER  >R GET-ORDER R> SWAP 1+ SET-ORDER ;

: ?ORDER ( wid -- flag )
   >R GET-ORDER ( ... n ) R> 0 ROT 0 ?DO ( ... wid flag)
      ROT ROT  TUCK = ROT OR
   LOOP NIP ;
